"""
  Package to translate UML Model into Document Representation
  This module translated the model into a simple hierarchy based on ownership.

"""

builtin_style = {

    'read-only'             : [ ( 'based-on', 'para'),
                                ( 'editable', '0' ) ],

    'changed-error'         : [ ( 'background', 'pink' ) ],
                                
    'changed-ok'            : [ ( 'background', 'yellow' ) ],
                                    'para'                  : [ ( 'left-margin', '0' ),
                                ( 'wrap-mode', 'wrap_word' ),
                                ( 'foreground', 'black' ),
                                ( 'pixels-above-lines', '3' ),
                                ( 'pixels-below-lines', '4' ),
                                ( 'pixels-inside-wrap', '1' ),
                                ( 'size-points', '14' ) ],

    'para-bold'             : [ ( 'weight', '700' ),
                                ( 'based-on', 'para' ) ],

    'para-100'              : [ ( 'based-on', 'para'),
                                ( 'left-margin', '110' ) ],

    'display-100'           : [ ( 'based-on', 'para'),
                                ( 'indent', '-100' ),
#                                ( 'tabs', display_tabs_100 ),
                                ( 'pixels-above-lines', '1' ),
                                ( 'pixels-below-lines', '1' ) ],

    'display-100-bold'      : [ ( 'based-on', 'display-100'),
                                ( 'weight', '700' ),
#                                ( 'tabs', display_tabs_100 ),
                                ( 'indent', '-100' ),
                                ( 'pixels-above-lines', '8' ),
                                ( 'pixels-below-lines', '4' ) ],

    'heading-1'             : [ ( 'based-on', 'para'),
                                ( 'weight', '700' ),
                                ( 'foreground', 'blue' ),
                                ( 'pixels-above-lines', '20' ),
                                ( 'pixels-below-lines', '10' ),
                                ( 'size-points', '24' ) ],
    
    'heading-2'             : [ ( 'based-on', 'para'),
                                ( 'weight', '700' ),
                                ( 'foreground', 'blue' ),
                                ( 'pixels-above-lines', '10' ),
                                ( 'pixels-below-lines', '5' ),
                                ( 'size-points', '20' ) ],
    
    'heading-3'             : [ ( 'based-on', 'para'),
                                ( 'weight', '700' ),
                                ( 'foreground', 'blue' ),
                                ( 'pixels-above-lines', '5' ),
                                ( 'pixels-below-lines','3' ),
                                ( 'size-points', '14' ) ] }
    

def _set_tag(tag, format, level = 0):
    """Helper function to recursively set the format for a tag"""
    for (type, value) in builtin_style[format]:
        if type == 'based-on':
            _set_tag(tag, value, level)
        else:
            if type == "indent" or type == "left-margin":
                value = "%d" % (int(value) + level * 10)
            #print "\tSet tag", type, value
            tag.set_attribute(type, value)


icon_table = {
    "Package": "palette/package.png",
    "Model": "palette/model.png",
    "Class": "palette/class.png",
    "Operation": "palette/operation-public.png",
    "Attribute": "palette/attribute-public.png",
    "Parameter": "palette/parameter-in.png",
    "Diagram": "palette/class-diagram.png" }


# --------------------------------------------------------------------------------
# Translation Classes
# --------------------------------------------------------------------------------

import Toolkit
import Reflective
import Foundation
import Diagram
import Drawing

class Field:
    def __init__(self, name, style = None, lead_text = "", post_text = "\n"):
        self.type = "Static"
        self.name = name
        self.tags = { }
        self.style = style
        self.lead_text = lead_text
        self.post_text = post_text

    def has_content(self):
        return 1
    
    def get_content(self, element):
        return (self.lead_text, self.post_text)

    def set_content(self, element, text):
        pass

    def has_children(self):
        return 0

# --------------------------------------------------------------------------------

class Field_Attribute:
    def __init__(self, name, attr_name, style = None, lead_text = "", post_text = "\n"):
        self.type = "Attribute"
        self.name = name
        self.tags = { }
        self.style = style
        self.lead_text = lead_text
        self.post_text = post_text
        self.attr_name = attr_name

    def has_content(self):
        return 1
    
    def get_content(self, element):
        text = element.get_attribute(self.attr_name)
        return self.lead_text + text + self.post_text

    def set_content(self, element, text):
        # strip lead_text and post_text
        if text.find(self.lead_text) == 0:
            text = text[len(self.lead_text):]
        n = text.rfind(self.post_text)
        if n > 0:
            text = text[:n]
        text.strip()
        element.set_attribute(self.attr_name, text)

    def has_children(self):
        return 0

# --------------------------------------------------------------------------------

class Field_TagValue:
    def __init__(self, name, tag_name, style = None, lead_text = "", post_text = "\n"):
        self.type = "Tag"
        self.name = name
        self.tags = { }
        self.style = style
        self.lead_text = lead_text
        self.post_text = post_text
        self.tag_name = tag_name

    def has_content(self):
        return 1
    
    def get_content(self, element):
        text = element.get_tagged_value(self.tag_name)
        return self.lead_text + text + self.post_text

    def set_content(self, element, text):
        # strip lead_text and post_text
        if text.find(self.lead_text) == 0:
            text = text[len(self.lead_text):]
        n = text.rfind(self.post_text)
        if n > 0:
            text = text[:n]
        text.strip()
        element.set_tagged_value(self.tag_name, text)

    def has_children(self):
        return 0

# --------------------------------------------------------------------------------

class Field_RoleList:
    def __init__(self, name, role_name, style = None):
        self.type = "Role"
        self.name = name
        self.tags = { }
        self.style = style
        self.role_name = role_name

    def has_content(self):
        return 0
    
    def has_children(self):
        return 1

    def get_children(self, element):
        return element.get_role_list(self.role_name)

    def add_child(self, element, prev_element, new_element):
        element.add_role(self.role_name, new_element)

# --------------------------------------------------------------------------------

class Field_Diagram:
    def __init__(self, name, style = None):
        self.type = "Diagram"
        self.name = name
        self.tags = { }
        self.lead_text = ""
        self.post_text = "\n"
        self.style = "read-only"

    def has_content(self):
        return 1
    
    def get_content(self, element):
        diagram = Diagram.Class(element)
        return diagram

    def set_content(self, element, text):
        pass

    def has_children(self):
        return 0

# --------------------------------------------------------------------------------

class Field_Icon:
    def __init__(self, file_name = ""):
        self.type = "Icon"
        self.name = "icon"
        self.tags = { }
        self.style = "read-only";
        self.lead_text = " "
        self.post_text = " "
        self.file_name = file_name

    def has_content(self):
        return 1
    
    def get_content(self, element):
        image = Drawing.Image(self.file_name)
        return image

    def set_content(self, element, text):
        pass

    def has_children(self):
        return 0

#================================================================================

class Block:
    def __init__(self, name, etype, stype, title, field_list):
        self.name = name
        self.type = "Block"
        self.tags = { }
        self.etype = etype
        self.stype = stype
        self.title = title
        self.field_list = field_list
        
#================================================================================

# Standard fields
field_name              = Field_Attribute('Heading_Name', 'name', 'heading-1')
field_attr_value        = Field_Attribute('Para_Doco', 'data_value', 'para')
field_doco              = Field_RoleList('Group_Doco', 'tagged_value')
#field_heading           = Field('Heading_', 'heading-2')
field_children          = Field_RoleList('Group_Elements', 'owned_element')
field_diagram           = Field_Diagram('Group_Diagram', 'read-only')

# Table to relate element types to formatting

formatting_table = [
    
    Block('Group_Class', 'Class', None, None,
          [ Field_Icon("palette/class.png"), field_name, field_doco, field_children ]),
    Block('Group_Package', 'Package', None, None,
          [ Field_Icon("palette/package.png"), field_name, field_doco, field_children ]),
    Block('Group_Model', 'Model', None, None,
          [ Field_Icon("palette/model.png"), field_name, field_doco, field_children ]),
    Block('Group_Diagram', 'Diagram', None, None,
          [ Field_Icon("palette/model.png"), field_name, field_diagram ]),

    # Components of classes
    Block('Group_Doco', 'TaggedValue', None, "doco",
          [ field_attr_value ]) ]
       


#================================================================================


class Translate:
    def __init__(self, doc):
        """Initialise the Translate object to convert the model into a document"""
        self.mContext = None
        self.mDocument = doc
        self.block_mark_serial = 0
        self.element_index = { }
        self.mark_changes = Toolkit.Mark("Temp_Change", doc.get_insert_mark())

        # Index of all tags that exist in document
        # (these are created on demand as the document is created)
        self.tag_index = { }
        
        # Create the tags used in blocks and fields
        block_index = { }
        for block in formatting_table:

            # Verify that block names are unique
            if block_index.has_key(block.name):
                print "ERROR duplicate block ", block.name
                raise ValueError, "Duplicate block"
            block_index[block.name] = block

            # Now for the fields of the block
            for field in block.field_list:
                field_index = { }
                
                # Fields within a block must be unique
                if field_index.has_key(field.name):
                    print "ERROR duplicate field ", field.name, "in block", block.name
                    raise ValueError, "Duplicate field"

                field_index[field.name] = field


        # Create some special tags
        # - this is a builtin style to highlight changes
        tag = Toolkit.Tag(doc, "Changed-OK")
        print "Create tag Changed_OK"
        _set_tag(tag, "changed-ok")
        self.tag_index["Changed-OK"] = tag
        print "Create tag Changed_Error"
        tag = Toolkit.Tag(doc, "Changed-Error")
        _set_tag(tag, "changed-error")
        self.tag_index["Changed-Error"] = tag


        # connect as a ChangeListener
        Reflective.set_change_listener(self)
        
        # connect to tree maintainer
        self.mDocument.set_maintainer(self)
        
        # add a mouse listener
        #self.mDocument.listener_add(self)

    #================================================================================
    # Content Generation
    def _lookup_block(self, element):
        etype = element.get_type()
        for block in formatting_table:
            # Element type must match
            if block.etype != etype:
                continue
            # Stereotype name if present must match
            if block.stype:
                stype = element.get_role_list('stereotype')
                if len(stype) > 0 and stype.count(block.stype) == 0:
                    continue
            # Title name if present must match
            if block.title:
                title = element.get_attribute('name')
                if title != block.title:
                    continue

            # Found a match
            return block
        print "ERROR could not locate block for", etype, stype, title
        return None

    def _generate_block_tag(self, block, level):
        name = "%s-%d" % (block.name, level)
        if self.tag_index.has_key(name):
            return self.tag_index[name]
        else:
            print "Create tag", name
            tag = Toolkit.Block(self.mDocument, name, level)
            self.tag_index[name] = tag
            return tag

    def _generate_field_tag(self, field, level):
        name = "%s-%d" % (field.name, level)
        if self.tag_index.has_key(name):
            return self.tag_index[name]
        else:
            print "Create tag", name
            tag = Toolkit.Field(self.mDocument, name, level)

            # Set according to style and nesting level
            if field.style:
                _set_tag(tag, field.style, level)
            self.tag_index[name] = tag
            return tag

    def _register_element(self, location, block_tag, block, level, element):
        # Create a mark and store it in the index
        self.block_mark_serial += 1
        mark = Toolkit.BlockMark("Block%d" % self.block_mark_serial,
                                 location, block_tag, level, element)
        if self.element_index.has_key(element):
            self.element_index[element].append((block, level, mark))
        else:
            self.element_index[element] = [(block, level, mark)]
        return mark

    def _insert_field(self, location, field, level, element):
        print "  Processing field", field.name
        if field.has_content():
            print "  - inserting content"
            content = field.get_content(element)
            field_tag = self._generate_field_tag(field, level)
            self.mDocument.insert_field_begin(location, field_tag)
            self.mDocument.insert_content(location, content)

            # Kludge to correct wrap problem
            if field.type == "Diagram":
                self.mDocument.insert_content(location, "\n")
                content.show()
            self.mDocument.insert_field_end(location, field_tag)
                
        if field.has_children():
            print "  - inserting children"
            children = field.get_children(element)
            if children and len(children):
                field_tag = self._generate_field_tag(field, level)
                self.mDocument.insert_field_begin(location, field_tag)
                level += 1
                for child in children:
                    self._recursive_build(level, location, child)
                level -= 1
                self.mDocument.insert_field_end(location, field_tag)

    def _recursive_build(self, level, location, element):
        block = self._lookup_block(element)
        if block:
            print "Processing block", block.name
            block_tag = self._generate_block_tag(block, level)
            mark = self._register_element(location, block_tag, block, level, element)
            self.mDocument.insert_block_begin(location, mark, block_tag)

            # Process the fields
            for field in block.field_list:
                self._insert_field(location, field, level, element)

            self.mDocument.insert_block_end(location, block_tag)


    #================================================================================
    # Change Listener

    def _handle_attribute_change(self, change):

        element = change.get_object()
        attr_name = change.get_attr_name()
        print "  - Changes: Attribute", attr_name, "Changed on", element
                
        # Find the element, if not present we ignore till it is inserted
        try:
            location_list = self.element_index[element]
            for block, level, mark in location_list:
                self.mark_changes.relocate_to_mark(mark)
                    
                # Check if this attribute is used by this block
                print "    Looking at block", block, level, mark
                for field in block.field_list:
                    print "      checking", field.name
                    field_tag = self._generate_field_tag(field, level)
                    if field.type == "Attribute":
                        if attr_name == field.attr_name:
                            content = field.get_content(element)
                            if self.mark_changes.at_tag_start(field_tag):
                                # Replace the contents of the existing field
                                print "      --> replacing field text"
                                self.mDocument.replace_field_text(self.mark_changes,
                                                                  field_tag, content)
                            else:
                                # Insert the field
                                print "      --> inserting field text"
                                self._insert_field(self.mark_changes, field, level, element)
                            break
                    if self.mark_changes.at_tag_start(field_tag):
                        self.mark_changes.relocate_to_tag_end(field_tag)
        except KeyError, n:
            print "Element lookup failed", n
            pass # ignore


    def _handle_role_change(self, change):

        element = change.get_object()
        role_name = change.get_role_name()
        print "  - Changes: Role", role_name, "Changed on", element
        print "    IGNORED***"

    def _handle_role_added(self, change):

        element = change.get_object()
        role_name = change.get_role_name()
        new_value = change.get_value()
        print "Changes: Role", role_name, "Added on", element

        # Look up the location of the element modified
        try:
            location_list = self.element_index[element]
            for block, level, mark in location_list:
                self.mark_changes.relocate_to_mark(mark)

                # Check if this attribute is used by this block
                print "    Looking at block", block, level, mark
                for field in block.field_list:
                    print "      checking", field.name
                    field_tag = self._generate_field_tag(field, level)
                    if field.type == "Role":
                        if role_name == field.role_name:
                            if self.mark_changes.at_tag_start(field_tag):
                                # update existing field contents
                                # I assume that this field has not content, only children
                                print "      --> adding new block to existing field"
                                assert field.has_children(), "Can only deal with limited roles"

                                # Only insert it if it is not already present
                                if not self.element_index.has_key(new_value):
                                    self.mark_changes.relocate_to_tag_end(field_tag)
                                    self._recursive_build(level + 1, self.mark_changes, new_value)
                            else:
                                # field not present, insert it now
                                print "      --> adding new field and block"
                                self._insert_field(self.mark_changes, field, level, element)
                            break
                    # Skip this field, if present
                    if self.mark_changes.at_tag_start(field_tag):
                        self.mark_changes.relocate_to_tag_end(field_tag)
        except KeyError, n:
            print "Element lookup failed", n
            pass # Ignore


    def _handle_role_removed(self, change):
        element = change.get_object()
        role_name = change.get_role_name()
        value = change.get_value()
        print "  - Changes: Role", role_name, "Remove on", element

        # Look up the location of the element modified
        try:
            location_list = self.element_index[value]
            for block, level, mark in location_list:
                print "    --> Delete block"
                block_tag = self._generate_field_tag(block, level)
                self.mDocument.delete_block(mark, block_tag)
            del self.element_index[value]
        except KeyError, n:
            print "Element lookup failed", n
            pass # Ignore


    def changes_occurred(self, change_list):
        print "Document ChangeListener: ", len(change_list), "changes"
        """Called for change in model with list of changes of the form
           get_type() returns: 'AttributeChanged'
                               'RoleChanged'
                               'RoleAdded'
                               'RoleRemoved'
           get_object() returns the class changed
           get_attr_name() gets the name of the attribute changed
           get_role_name() gets the name of the role changed/added/removed
           get_value() or get_new_value() gets the value changed/added/removed
           get_old_value() gets the old value except for role added
        """
        if not self.mContext:
            return

        for change in change_list:

            # Ignore all those strange things used to implement the diagram
            element = change.get_object()
            if element.is_a_type_of("ModelElement"):
                # Looks like it may be useful

                if change.get_type() == "AttributeChanged":
                    self._handle_attribute_change(change)
                    continue
                
                if change.get_type() == "RoleChanged":
                    self._handle_role_change(change)
                    continue

                if change.get_type() == "RoleAdded":
                    self._handle_role_added(change)
                    continue
            
                if change.get_type() == "RoleRemoved":
                    self._handle_role_removed(change)
                    continue
        print "Document Changes Complete"    

    #================================================================================
    # Document Maintainer
    def changed(self, node, mark_block, tag_block,
                mark_field_begin, mark_field_end, tag_field):
        print "DocumentMaintainer.changed"

    def apply_change(self, node, mark_block, tag_block,
                     mark_field_begin, mark_field_end, tag_field):
        element = Reflective.Class(node)
        print "DocumentMaintainer.apply_change"
        print "TAGS", tag_block.get_name(), tag_field.get_name()
        # Locate the block entry
        tag_block_name, tag_block_level = tag_block.get_name().split('-', 2)
        tag_field_name, tag_field_level = tag_field.get_name().split('-', 2)
        for block in formatting_table:
            print "  Checking ", block.name,
            if block.name == tag_block_name:
                print "  - matched"
                for field in block.field_list:
                    print "    Checking ", field.name
                    if field.name == tag_field_name:
                        print "    - matched"

                        if field.has_content():
                            print "Save content changes"
                            content = self.mDocument.get_field_text(mark_field_begin, tag_field)
                            field.set_content(element, content)

                            
    #================================================================================
    # Mouse Listener
    #
    
    #================================================================================

    def set_context(self, context):
        print "Document Set Context", context
        global global_context
        global_context = context
        if self.mContext != context:
            if self.mContext:
                print "Clear the old document"
                self.mDocument.clear()
                del self.element_index[:]
                
            self.mContext = context
            if self.mContext:
                print "Build a new document"
                root = self.mContext.get_root();

                # insert the contents of the model into the document
                mark = self.mDocument.get_insert_mark()
                self._recursive_build(0, mark, root)


# --------------------------------------------------------------------------------

def create(doc):
    #//doc.create_tags(tag_table)
    #//doc.clear()
    return Translate(doc)

