// Dachshund
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "uuid.h"

#include "foundation/macros.h"

#include <cstring>
#include <ctime>
#include <cstdlib>
#include <cerrno>
#include <cstdio>
#include <iostream>
using namespace std;

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>

#include <numeric>
using namespace std;

namespace Common {

    bool UUID::mInitialised = false;
    bool UUID::mHaveRandom;
    unsigned char UUID::mNodeID[6];

    namespace {

	// The stuff in this namespace was originally written by
	// Theodore Ts'o. Here is the header from that file:

	/*
	* gen_uuid.c --- generate a DCE-compatible uuid
	*
	* Copyright (C) 1996, 1997 Theodore Ts'o.
	*
	* %Begin-Header%
	* This file may be redistributed under the terms of the GNU Public
	* License.
	* %End-Header%
	*/

	void
	get_random_bytes(void * buf, int nbytes) {
	    static int fd = -2;
	    char * cp = static_cast<char *>(buf);

	    if(fd == -2) {
		fd = open("/dev/urandom", O_RDONLY);
		srand((getpid() << 16) ^ getuid() ^ time(0));
	    }

	    if(fd >= 0) {
		while (nbytes > 0) {
		    int i = read(fd, cp, nbytes);

		    if(i < 0) {
			if((errno == EINTR) || (errno == EAGAIN))
			    continue;
			break;
		    }

		    nbytes -= i;
		    cp += i;
		}
	    }

	    if(nbytes == 0) {
		return;
	    }

	    // put something better here if no /dev/random!
	    for(int i = 0; i < nbytes; i++) {
		*cp++ = rand() & 0xFF;
	    }

	    return;
	}

	int
	get_node_id(unsigned char *node_id) {
#ifdef HAVE_NET_IF_H
	    int sd;
	    struct ifreq ifr, * ifrp;
	    struct ifconf ifc;
	    char buf[1024];
	    int n, i;
	    unsigned char * a;

	    /*
	    * BSD 4.4 defines the size of an ifreq to be
	    * max(sizeof(ifreq), sizeof(ifreq.ifr_name)+ifreq.ifr_addr.sa_len
	    * However, under earlier systems, sa_len isn't present, so the size is 
	    * just sizeof(struct ifreq)
	     */
#ifdef HAVE_SA_LEN
#define ifreq_size(i) max(sizeof(struct ifreq),\
			  sizeof((i).ifr_name)+(i).ifr_addr.sa_len)
#else
#define ifreq_size(i) sizeof(struct ifreq)
#endif /* HAVE_SA_LEN*/

	    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);

	    if(sd < 0) {
		return 0;
	    }

	    memset(buf, 0, sizeof(buf));
	    ifc.ifc_len = sizeof(buf);
	    ifc.ifc_buf = buf;
	    if(ioctl(sd, SIOCGIFCONF, static_cast<char *>(&ifc)) < 0) {
		close(sd);
		return 0;
	    }
	    n = ifc.ifc_len;
	    for(i = 0; i < n; i+= ifreq_size(*ifr) ) {
		ifrp = (struct ifreq *)(static_cast<char *>(ifc.ifc_buf + i));
		strncpy(ifr.ifr_name, ifrp->ifr_name, IFNAMSIZ);
#ifdef SIOCGIFHWADDR
		if(ioctl(sd, SIOCGIFHWADDR, &ifr) < 0)
		    continue;
		a = static_cast<unsigned char *>(&ifr.ifr_hwaddr.sa_data);
#else
#ifdef SIOCGENADDR
		if(ioctl(sd, SIOCGENADDR, &ifr) < 0)
		    continue;
		a = static_cast<unsigned char *>(ifr.ifr_enaddr);
#else
		/*
		* XXX we don't have a way of getting the hardware
		* address
		 */
		close(sd);
		return 1;
#endif /* SIOCGENADDR */
#endif /* SIOCGIFHWADDR */
		if(!a[0] && !a[1] && !a[2] && !a[3] && !a[4] && !a[5])
		    continue;
		if(node_id) {
		    memcpy(node_id, a, 6);
		    close(sd);
		    return 0;
		}
	    }

	    close(sd);
#endif
	    return 1;
	}

	/* Assume that the gettimeofday() has microsecond granularity */
#define MAX_ADJUSTMENT 10

	int get_clock(unsigned int *clock_high,
		      unsigned int *clock_low,
		      unsigned short int *ret_clock_seq) {
	    static int adjustment = 0;
	    static struct timeval last = {0, 0};
	    static unsigned short int clock_seq;
	    struct timeval tv;
	    unsigned long long clock_reg;

try_again:
	    gettimeofday(&tv, 0);
	    if((last.tv_sec == 0) && (last.tv_usec == 0)) {
		get_random_bytes(&clock_seq, sizeof(clock_seq));
		clock_seq &= 0x1FFF;
		last = tv;
		last.tv_sec--;
	    }
	    if((tv.tv_sec < last.tv_sec) ||
	       ((tv.tv_sec == last.tv_sec) && (tv.tv_usec < last.tv_usec))) {
		clock_seq = (clock_seq + 1) & 0x1FFF;
		adjustment = 0;
	    } else if((tv.tv_sec == last.tv_sec) && (tv.tv_usec == last.tv_usec)) {
		if(adjustment >= MAX_ADJUSTMENT)
		    goto try_again;
		adjustment++;
	    } else {
		adjustment = 0;
	    }

	    clock_reg = tv.tv_usec*10 + adjustment;
	    clock_reg += ((unsigned long long) tv.tv_sec)*10000000;
	    clock_reg += (((unsigned long long) 0x01B21DD2) << 32) + 0x13814000;

	    *clock_high = clock_reg >> 32;
	    *clock_low = clock_reg;
	    *ret_clock_seq = clock_seq;

	    return 0;
	}

    } // namespace {anonymous}

    UUID::UUID() {
	clear();
    }

    UUID::~UUID() {
    }

    UUID
    UUID::create() {
	UUID uuid;
	uuid.generate();
	return uuid;
    }

    void
    UUID::generate() {
	if(!mInitialised) {
	    if(!get_node_id(mNodeID)) {
		get_random_bytes(mNodeID, 6);

		// set multicast bit, to prevent conflicts with
		// IEEE 802 addresses obtained from network cards

		mNodeID[0] |= 0x80;
	    }

	    if(access("/dev/urandom", R_OK) == 0) {
		mHaveRandom = true;
	    }
	    else {
		mHaveRandom = false;
	    }

	    mInitialised = true;
	}

	if(mHaveRandom) {
	    generate_random();
	}
	else {
	    generate_time();
	}
    }

    void
    UUID::generate_random() {
	get_random_bytes(mRaw, 16);

	mClockSeq = (mClockSeq & 0x3fff) | 0x8000;
	mTimeHiAndVersion = (mTimeHiAndVersion & 0x0fff) | 0x4000;
    }

    void
    UUID::generate_time() {
	unsigned int clock_mid;

	get_clock(&clock_mid, &mTimeLow, &mClockSeq);

	mClockSeq |= 0x8000;
	mTimeMid = clock_mid;
	mTimeHiAndVersion = (clock_mid >> 16) | 0x1000;
	memcpy(mNode, mNodeID, 6);
    }

    void
    UUID::clear() {
	memset(mRaw, 0, 16);
    }

    UUID::operator bool () const {
	for(int i = 0; i < 16; ++i) {
	    if(mRaw[i]) {
		return true;
	    }
	}

	return false;
    }

    bool
    UUID::operator ! () const {
	return ! operator bool();
    }

    string
    UUID::get_string() const {
	string s;

	s.resize(36);

	snprintf(&s[0], s.size() + 1, "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
		 mTimeLow,
		 mTimeMid,
		 mTimeHiAndVersion,
		 mClockSeq >> 8,
		 mClockSeq & 0xff,
		 mNode[0],
		 mNode[1],
		 mNode[2],
		 mNode[3],
		 mNode[4],
		 mNode[5]);

	return s;
    }

    void
    UUID::set_string(const string & s) {
	RETURN_IF_FAIL(s.size() == 36);

	const char * in = s.c_str();
	const char * cp;
	char buf[3];
	int i;

	// do a run through the string to check the format

	for(i = 0, cp = in; i < 36; i++, cp++) {
	    if((i == 8) || (i == 13) || (i == 18) || (i == 23)) {
		RETURN_IF_FAIL(*cp == '-');
	    }
	    else {
		RETURN_IF_FAIL(isxdigit(*cp));
	    }
	}

	// now modify this object

	mTimeLow = strtoul(in, 0, 16);
	mTimeMid = strtoul(in + 9, 0, 16);
	mTimeHiAndVersion = strtoul(in + 14, 0, 16);
	mClockSeq = strtoul(in + 19, 0, 16);
	cp = in + 24;
	buf[2] = '\0';

	for(i = 0; i < 6; i++) {
	    buf[0] = *cp++;
	    buf[1] = *cp++;
	    mNode[i] = strtoul(buf, 0, 16);
	}
    }

    bool
    operator == (const UUID & uuid1, const UUID & uuid2) {
	return uuid1.get_string() == uuid2.get_string();
    }

    bool
    operator < (const UUID & uuid1, const UUID & uuid2) {
	return uuid1.get_string() < uuid2.get_string();
    }

    ostream &
    operator << (ostream & ost, const UUID & uuid) {
	ost << uuid.get_string();

	return ost;
    }

    istream &
    operator >> (istream & ist, UUID & uuid) {
	string s;
	s.resize(36);

	for(int i = 0; i < 36; ++i) {
	    char c;
	    ist.get(c);

	    if(ist.good()) {
		s[i] = c;
	    }
	    else {
		ist.setstate(ios::failbit);
		return ist;
	    }
	}

	uuid.set_string(s);

	return ist;
    }

} // namespace Common
