#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_edge_text.h"

#include "fig_edge.h"

#include "foundation/macros.h"
#include "drawing/segment.h"
#include "drawing/line.h"

namespace Diagram {

    FigEdgeText::FigEdgeText(FigEdge * edge) : FigText(),
	mEdge(edge),
	mFraction(0.5),
	mDistance(0.0) {
	ASSERT(edge);
    }

    FigEdgeText::~FigEdgeText() {
    }

    void
    FigEdgeText::compute_position() {
	ASSERT(mEdge);
	ASSERT(mEdge->get_points().size() >= 2);

	double total_length1 = 0.0;

	for(unsigned int i = 0; i < mEdge->get_points().size() - 1; ++i) {
	    const Drawing::Point & p1 = mEdge->get_points()[i];
	    const Drawing::Point & p2 = mEdge->get_points()[i + 1];

	    Drawing::Segment segment(p1, p2);

	    double segment_length = segment.get_length();

	    total_length1 += segment_length;
	}

	ASSERT(mFraction >= 0.0);
	ASSERT(mFraction <= 1.0);

	double length = mFraction * total_length1;

	double total_length2 = 0.0;

	Drawing::Point location;

	for(unsigned int i = 0; i < mEdge->get_points().size() - 1; ++i) {
	    const Drawing::Point & p1 = mEdge->get_points()[i];
	    const Drawing::Point & p2 = mEdge->get_points()[i + 1];

	    Drawing::Segment segment(p1, p2);

	    double segment_length = segment.get_length();

	    total_length2 += segment_length;

	    if(total_length2 >= length) {
		double fraction =
		    (mFraction - (total_length2 - segment_length) / total_length1) *
		    (total_length1 / segment_length);

		Drawing::Point point = segment.get_point1() + fraction * (segment.get_point2() -
								segment.get_point1());

		Drawing::Vector normal = (segment.get_point1() - segment.get_point2()).get_normal();

		double normal_length = normal.get_length();

		Drawing::Vector offset = (mDistance / normal_length) * normal;

		location = point + offset;

		break;
	    }
	}

	Drawing::Rectangle new_bounds(location - Drawing::Vector(15, 6), Drawing::Vector(30, 12));

	if(mBounds != new_bounds) {
	    damage();
	    mBounds = new_bounds;
	    damage();
	}
    }

    void
    FigEdgeText::compute_relationship() {
	ASSERT(mEdge && mEdge->get_points().size() >= 2);

	Drawing::Point center;

	get_center(center);

	int closest_segment = -1;

	double total_length1 = 0.0;

	for(unsigned int i = 0; i < mEdge->get_points().size() - 1; ++i) {
	    const Drawing::Point & p1 = mEdge->get_points()[i];
	    const Drawing::Point & p2 = mEdge->get_points()[i + 1];

	    Drawing::Segment segment(p1, p2);

	    double segment_length = segment.get_length();

	    total_length1 += segment_length;

	    double distance_to_segment = segment.get_distance_to(center);

	    if(closest_segment < 0 || distance_to_segment < mDistance) {
		closest_segment = i;
		mDistance = distance_to_segment;
	    }
	}

	ASSERT(closest_segment >= 0);

	Drawing::Point p1(mEdge->get_points()[closest_segment]);
	Drawing::Point p2(mEdge->get_points()[closest_segment + 1]);

	Drawing::Line line1(p1, p2);
	Drawing::Line line2(center, (p2 - p1).get_normal());

	Drawing::Point location;

	ASSERT(line1.get_intercept(line2, location));

	Drawing::Vector v1(center - location);
	Drawing::Vector v2(line1.get_vector().get_normal());

	if(v1 * v2 > 0) {
	    mDistance = -mDistance;
	}

	double total_length2 = 0.0;

	for(unsigned int i = 0; i < mEdge->get_points().size() - 1; ++i) {
	    const Drawing::Point & p1 = mEdge->get_points()[i];
	    const Drawing::Point & p2 = mEdge->get_points()[i + 1];

	    Drawing::Segment segment(p1, p2);

	    double segment_length = segment.get_length();

	    total_length2 += segment_length;

	    if(int(i) == closest_segment) {
		mFraction = (total_length2 - segment_length +
			     (location - p1).get_length()) / total_length1;

		if(mFraction < 0.0) {
		    WARNING("Adjusting fraction: " << mFraction);
		    mFraction = 0.0;
		}
		else if(mFraction > 1.0) {
		    WARNING("Adjusting fraction: " << mFraction);
		    mFraction = 1.0;
		}

		break;
	    }
	}
    }

    void
    FigEdgeText::set_all(double fraction, double distance) {
	mFraction = fraction;
	mDistance = distance;

	if(mEdge) {
	    compute_position();
	}
    }

} // namespace Diagram
