// This code is a modified port of GEF (Graphical Editing Framework) // http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "fig_node.h"

#include "fig_edge.h"
#include "selection_resize.h"

#include "foundation/macros.h"

#include "drawing/drawable.h"
#include "drawing/rectangle.h"

#include <functional>
#include <algorithm>
#include <iterator>
using namespace std;

namespace Diagram {

    FigNode::FigNode(Object * owner) : FigGroup(owner),
	mHighlight(false) {
    }

    FigNode::~FigNode() {
	for_each(mFigEdges.begin(), mFigEdges.end(), bind2nd(mem_fun(&FigEdge::remove_node), this));
    }

    void
    FigNode::relocate(const Drawing::Point & location) {
	FigGroup::relocate(location);
	update_edges();
    }

    void
    FigNode::resize(const Drawing::Vector & size) {
	FigGroup::resize(size);
	update_edges();
    }

    void
    FigNode::add_edge(FigEdge * fig_edge) {
	mFigEdges.push_front(fig_edge);
    }

    void
    FigNode::remove_edge(FigEdge * fig_edge) {
	// we can't use mFigEdges.remove(fig_edge); because it will remove all instances
	// of the edge. This causes a problem if two ends of the same edge
	// are connected to a node.

	list<FigEdge *>::iterator iter = find(mFigEdges.begin(), mFigEdges.end(), fig_edge);

	ASSERT(iter != mFigEdges.end());

	mFigEdges.erase(iter);
    }

    void
    FigNode::paint(Drawing::Drawable * drawable) const {
	FigGroup::paint(drawable);

	if(mHighlight) {
	    const Drawing::Rectangle & bounds = get_bounds();

	    int x = bounds.get_x();
	    int y = bounds.get_y();
	    int w = bounds.get_w();
	    int h = bounds.get_h();

	    drawable->set_color("#00dd33");

	    drawable->draw_rectangle(Drawing::Rectangle(x, y, w, 2), true);		// top edge
	    drawable->draw_rectangle(Drawing::Rectangle(x, y, 2, h), true);		// left edge
	    drawable->draw_rectangle(Drawing::Rectangle(x + w - 2, y, 2, h), true);// right edge
	    drawable->draw_rectangle(Drawing::Rectangle(x, y + h - 2, w, 2), true);// right edge
	}
    }

    void
    FigNode::set_highlight(bool highlight) {
	if(mHighlight != highlight) {
	    mHighlight = highlight;
	    damage();
	}
    }

    void
    FigNode::update_edges() {
	for_each(mFigEdges.begin(), mFigEdges.end(), mem_fun(&FigEdge::update_edge_to_nodes));
    }

    Selection *
    FigNode::create_selection() {
	return new SelectionResize(this);
    }

    bool
    FigNode::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
#if 0
	set_highlight(true);
#endif

	return true;
    }

    bool
    FigNode::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
#if 0
	set_highlight(false);
#endif

	return true;
    }

    bool
    FigNode::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigNode::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigNode::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    FigNode::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    void
    FigNode::cleanup() {
	FigGroup::cleanup();
	for_each(mFigEdges.begin(), mFigEdges.end(), mem_fun(&Fig::cleanup));
    }

} // namespace Diagram
