// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "layer_diagram.h"

#include "fig.h"
#include "selection_manager.h"

#include "drawing/drawable.h"
#include "drawing/rectangle.h"

#include "foundation/macros.h"
#include "foundation/functional.h"
#include "foundation/algorithm.h"

#include <algorithm>
#include <functional>
using namespace std;

namespace Diagram {

    LayerDiagram::LayerDiagram() : Layer() {
    }

    LayerDiagram::~LayerDiagram() {
	clear();
    }

    Fig *
    LayerDiagram::get_presentation_for(Object * object) {
	ASSERT(object);

	list<Fig *>::const_iterator iter = find_if(mFigs.begin(), mFigs.end(),
						   bind2nd(mem_fun(&Fig::is_owner), object));

	return iter == mFigs.end() ? 0 : *iter;
    }

    Fig *
    LayerDiagram::hit(const Drawing::Rectangle & rect) {
	list<Fig *>::iterator iter = find_if(mFigs.begin(), mFigs.end(),
					     bind2nd_vr(mem_fun(&Fig::hit), rect));

	return iter == mFigs.end() ? 0 : *iter;
    }

    Fig *
    LayerDiagram::hit_not(const Drawing::Rectangle & rect, Fig * disregard) {
	list<Fig *>::iterator iter = mFigs.begin();

TRYAGAIN:
	iter = find_if(iter, mFigs.end(), bind2nd_vr(mem_fun(&Fig::hit), rect));

	if(*iter == disregard) {
	    ++iter;
	    goto TRYAGAIN;
	}

	return iter == mFigs.end() ? 0 : *iter;
    }

    void
    LayerDiagram::add(Fig * fig) {
	MESSAGE("add fig: " << fig);

	RETURN_IF_FAIL(fig);
	RETURN_IF_FAIL(find(mFigs.begin(), mFigs.end(), fig) == mFigs.end());

	Layer::add(fig);

	fig->ref();
	mFigs.push_front(fig);
	fig->set_layer(this);
    }

    void
    LayerDiagram::remove(Fig * fig) {
	MESSAGE("remove fig: " << fig);

	RETURN_IF_FAIL(fig);

	list<Fig *>::iterator iter = find(mFigs.begin(), mFigs.end(), fig);

	if(iter == mFigs.end()) {
	    ERROR("tried to remove a fig from a layer but the fig couldn't be found");
	}
	else {
	    ASSERT(*iter == fig);

	    SelectionManager * selection_manager = SelectionManager::instance();

	    if(selection_manager->contains(fig)) {
		selection_manager->damage();
		selection_manager->remove(fig);
	    }

	    fig->set_layer(0);
	    mFigs.erase(iter);
	    fig->unref();
	}

	Layer::remove(fig);
    }

    void
    LayerDiagram::clear() {
	for_each(mFigs.begin(), mFigs.end(),
		 bind2nd(mem_fun(&Fig::set_layer), static_cast<Layer *>(0)));
	for_each(mFigs.begin(), mFigs.end(), mem_fun(&Object::unref));

	mFigs.clear();
    }

    void
    LayerDiagram::paint(Drawing::Drawable * drawable) const {
	ASSERT(drawable);

	const Drawing::Rectangle & clip = drawable->get_clip();

	if(clip) {
	    for_each_if(mFigs.rbegin(), mFigs.rend(),
			bind2nd_vr(mem_fun(&Fig::intersects), clip),
			bind2nd(mem_fun(&Fig::paint), drawable));
	}
	else {
	    for_each(mFigs.rbegin(), mFigs.rend(), bind2nd(mem_fun(&Fig::paint), drawable));
	}

#if 0
	// debug the bounds

	drawable->set_color("#0088ff");

	for(list<Fig *>::iterator iter = mFigs.begin(); iter != mFigs.end(); ++iter) {
	    Fig * fig = *iter;
	    drawable->draw_rectangle(fig->get_bounds(), false);
	}
#endif
    }

    void
    LayerDiagram::send_to_back(Fig * fig) {
	ASSERT(fig);

	list<Fig *>::iterator iter = find(mFigs.begin(), mFigs.end(), fig);

	if(iter == mFigs.end()) {
	    ERROR("tried to send a fig to back in layer, but the fig couldn't be found");
	}
	else {
	    Fig * f = *iter;
	    mFigs.erase(iter);
	    mFigs.push_back(f);
	}

	fig->damage();
    }

    void
    LayerDiagram::send_backward(Fig * fig) {
	ASSERT(fig);

	list<Fig *>::iterator iter = find(mFigs.begin(), mFigs.end(), fig);

	if(iter == mFigs.end()) {
	    ERROR("tried to bring a fig forward, but the fig couldn't be found");
	}
	else {
	    Fig * f = *iter;
	    list<Fig *>::iterator i = mFigs.erase(iter);

	    if(i != mFigs.end()) {
		++i;
	    }

	    mFigs.insert(i, f);
	}

	fig->damage();
    }

    void
    LayerDiagram::send_behind(Fig * f1, Fig * f2) {
	ASSERT(f1);
	ASSERT(f2);

	if(f1 != f2) {
	    list<Fig *>::iterator iter1 = find(mFigs.begin(), mFigs.end(), f1);

	    if(iter1 != mFigs.end()) {
		list<Fig *>::iterator iter2 = find(mFigs.begin(), mFigs.end(), f2);

		if(iter2 != mFigs.end()) {
		    Fig * f = *iter1;
		    mFigs.erase(iter1);
		    ++iter2;
		    mFigs.insert(iter2, f);
		}
		else {
		    ERROR("couldn't find f2 in layer");
		}
	    }
	    else {
		ERROR("couldn't find f1 in layer");
	    }
	}
	else {
	    ERROR("they are the same fig");
	}

	f1->damage();
    }

    void
    LayerDiagram::bring_to_front(Fig * fig) {
	ASSERT(fig);

	list<Fig *>::iterator iter = find(mFigs.begin(), mFigs.end(), fig);

	if(iter == mFigs.end()) {
	    ERROR("tried to bring a fig to front in layer, but the fig couldn't be found");
	}
	else {
	    Fig * f = *iter;
	    mFigs.erase(iter);
	    mFigs.push_front(f);
	}

	fig->damage();
    }

    void
    LayerDiagram::bring_in_front_of(Fig * f1, Fig * f2) {
	ASSERT(f1);
	ASSERT(f2);

	if(f1 != f2) {
	    list<Fig *>::iterator iter1 = find(mFigs.begin(), mFigs.end(), f1);

	    if(iter1 != mFigs.end()) {
		list<Fig *>::iterator iter2 = find(mFigs.begin(), mFigs.end(), f2);

		if(iter2 != mFigs.end()) {
		    Fig * f = *iter1;
		    mFigs.erase(iter1);
		    mFigs.insert(iter2, f);
		}
		else {
		    ERROR("couldn't find f2 in layer");
		}
	    }
	    else {
		ERROR("couldn't find f1 in layer");
	    }
	}
	else {
	    ERROR("they are the same fig");
	}

	f1->damage();
    }

    void
    LayerDiagram::bring_forward(Fig * fig) {
	ASSERT(fig);

	list<Fig *>::iterator iter = find(mFigs.begin(), mFigs.end(), fig);

	if(iter == mFigs.end()) {
	    ERROR("tried to bring a fig forward, but the fig couldn't be found");
	}
	else {
	    Fig * f = *iter;

	    list<Fig *>::iterator i = mFigs.erase(iter);

	    if(i != mFigs.begin()) {
		--i;
	    }

	    mFigs.insert(i, f);
	}

	fig->damage();
    }

    const list<Fig *> &
    LayerDiagram::get_figs() {
	return mFigs;
    }

} // namespace Diagram
