// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "mode_create_fig_poly.h"

#include "foundation/macros.h"
#include "foundation/utility.h"
#include "toolkit/mouse_event.h"

#include "fig_poly_line.h"
#include "editor.h"
#include "handle.h"
#include "selection_manager.h"

namespace Diagram {

    ModeCreateFigPoly::ModeCreateFigPoly() : ModeCreate() {
	mCursor = Toolkit::C_DOT;
    }

    ModeCreateFigPoly::~ModeCreateFigPoly() {
    }

    bool
    ModeCreateFigPoly::can_start(const Toolkit::MouseEvent * mouse_event) {
	return true;
    }

    bool
    ModeCreateFigPoly::can_finish(const Toolkit::MouseEvent * mouse_event) {
	const Drawing::Point & point = mouse_event->get_point();

	return near_last(point);
    }

    bool
    ModeCreateFigPoly::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	if(!mNewItem) {
	    if(can_start(mouse_event)) {
		ModeCreate::mouse_pressed(mouse_event);
	    }
	    else {
		return true;
	    }
	}

	const Drawing::Point & point = mouse_event->get_point();

	if(!near_last(point)) {
	    FigPoly * fig_poly = dynamic_cast<FigPoly *>(mNewItem);

	    Editor * editor = Editor::current();

	    ASSERT(editor);

	    Drawing::Point snapped_point = point;

	    damage();

	    editor->snap(snapped_point);

	    fig_poly->add_point(snapped_point);

	    mBounds = fig_poly->get_bounds();

	    damage();
	}

	return true;
    }

    bool
    ModeCreateFigPoly::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	const Drawing::Point & point = mouse_event->get_point();

	if(mNewItem) {
	    FigPoly * fig_poly = known_cast<FigPoly *>(mNewItem);

	    if(fig_poly->get_points().size() == 2) {
		mouse_pressed(mouse_event);
	    }

	    if(can_finish(mouse_event)) {
		Editor * editor = Editor::current();

		ASSERT(editor);

		damage();

		mHandle.set_index(fig_poly->get_points().size() - 1);

		fig_poly->remove_point(mHandle.get_index());

		editor->add(fig_poly);

		fig_poly->damage();

		SelectionManager * selection_manager = SelectionManager::instance();

		if(selection_manager) {
		    selection_manager->damage();
		    selection_manager->set(fig_poly);
		    selection_manager->damage();
		}

		finish(mouse_event);

		mNewItem = 0;
	    }
	}

	mLast = point;

	return true;
    }

    bool
    ModeCreateFigPoly::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	const Drawing::Point & point = mouse_event->get_point();

	if(mNewItem) {
	    FigPoly * fig_poly = known_cast<FigPoly *>(mNewItem);

	    ASSERT(fig_poly);

	    Editor * editor = Editor::current();

	    ASSERT(editor);

	    Drawing::Point snapped_point = point;

	    damage();

	    editor->snap(snapped_point);

	    mHandle.set_index(fig_poly->get_points().size() - 1);

	    fig_poly->set_point(mHandle, snapped_point);
	    mBounds = fig_poly->get_bounds();

	    damage();
	}

	return true;
    }

    bool
    ModeCreateFigPoly::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	return mouse_dragged(mouse_event);
    }

    string
    ModeCreateFigPoly::get_instructions() const {
	return "Click to add points and double-click to finish";
    }

    bool
    ModeCreateFigPoly::near_last(const Drawing::Point & point) const {
	const int GRIP_SIZE = 8;		// FIXME where does it belong

	return (point - mLast).get_length() < GRIP_SIZE;
    }

} // namespace Diagram
