// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "selection_manager.h"

#include "foundation/functional.h"
#include "foundation/algorithm.h"
#include "foundation/macros.h"
#include "drawing/rectangle.h"
#include "drawing/point.h"
#include "drawing/vector.h"
#include "toolkit/key_event.h"		// needed ???

#include "fig.h"
#include "fig_node.h"
#include "fig_edge.h"
#include "selection.h"
#include "handle.h"
#include "layer.h"

#include <numeric>
#include <algorithm>
#include <functional>
#include <set>
using namespace std;

namespace Diagram {

    SelectionManager * SelectionManager::mInstance = 0;

    SelectionManager *
    SelectionManager::instance() {
	if(!mInstance) {
	    mInstance = new SelectionManager();
	}

	return mInstance;
    }

    void
    SelectionManager::shutdown() {
	ASSERT(mInstance);

	delete mInstance;

	mInstance = 0;
    }

    SelectionManager::SelectionManager() : Object(),
					   Toolkit::MouseListener(), Toolkit::KeyListener(),
	mCacheValid(false),
	mInInteraction(false) {
    }

    SelectionManager::~SelectionManager() {
	for_each(mSelections.begin(), mSelections.end(), mem_fun(&Object::unref));
    }

    void
    SelectionManager::get_bounds(Drawing::Rectangle & rect) const {
	rect = accumulate(mSelections.begin(), mSelections.end(), rect,
			  unite_area(mem_fun(&Selection::get_bounds)));
    }

    void
    SelectionManager::get_contents_bounds(Drawing::Rectangle & rect) const {
	rect = accumulate(mSelections.begin(), mSelections.end(), rect,
			  unite_area(mem_fun(&Selection::get_contents_bounds)));
    }

    void
    SelectionManager::translate(const Drawing::Vector & offset) {
	// FIXME I think this is broken

	if(!mCacheValid) {
	    mAffectedEdges.clear();

	    std::set<FigNode *> fig_nodes;
	    std::set<FigEdge *> fig_edges;

	    for(list<Selection *>::iterator iter = mSelections.begin();
		iter != mSelections.end();
		++iter) {
		Selection * selection = *iter;
		Fig * fig = selection->get_content();
		FigNode * fig_node = dynamic_cast<FigNode *>(fig);

		if(fig_node) {
		    fig_nodes.insert(fig_node);

		    for(list<FigEdge *>::const_iterator iter = fig_node->get_edges().begin();
			iter != fig_node->get_edges().end();
			++iter) {
			FigEdge * fig_edge = *iter;
			fig_edges.insert(fig_edge);
		    }
		}
	    }

	    for(std::set<FigEdge *>::iterator iter = fig_edges.begin();
		iter != fig_edges.end();
		++iter) {
		FigEdge * fig_edge = *iter;

		if(fig_nodes.find(fig_edge->get_source_fig()) != fig_nodes.end() &&
		fig_nodes.find(fig_edge->get_destination_fig()) != fig_nodes.end()) {
		    mAffectedEdges.push_back(fig_edge);
		}
	    }

	    mCacheValid = true;
	}

#if 0
	for_each(mAffectedEdges.begin(), mAffectedEdges.end(),
                 bind2nd_vr(mem_fun(&Fig::translate), offset));
#endif

	// FIXME
#if 0
	for_each_if(mSelections.begin(),
		    mSelections.end(),
		    mem_fun(&Selection::is_movable),
		    bind2nd_vr(mem_fun(&Selection::translate), offset));
#else
	for_each(mSelections.begin(),
		 mSelections.end(),
		 bind2nd_vr(mem_fun(&Selection::translate), offset));
#endif
    }

    void
    SelectionManager::drag_handle(const Drawing::Point & m, const Drawing::Point & an,
				  Handle & handle) {
	if(mSelections.size() == 1) {
	    Selection * selection = mSelections.front();
	    selection->drag_handle(m, an, handle);
	}
    }

    void
    SelectionManager::add(Fig * fig) {
	Selection * selection = make_selection_for(fig);

	if(selection) {
	    // we have to clear the selections if this fig
	    // is not on the same layer

	    if(!mSelections.empty()) {
		Fig * content = mSelections.front()->get_content();
		if(content->get_layer() != fig->get_layer()) {
		    clear();
		}
	    }

	    mSelections.push_front(selection);

	    mCacheValid = false;
	}
    }

    void
    SelectionManager::remove(Fig * fig) {
	list<Selection *>::iterator iter = find_if(mSelections.begin(), mSelections.end(),
						   bind2nd(mem_fun(&Selection::is_content), fig));

	if(iter == mSelections.end()) {
	    ERROR("Fig not found");
	}
	else {
	    Selection * selection = *iter;
	    mSelections.erase(iter);
	    selection->unref();
	}

	mCacheValid = false;
    }

    void
    SelectionManager::set(Fig * fig) {
	clear();
	add(fig);
    }

    void
    SelectionManager::clear() {
	mSelections.clear();

	mCacheValid = false;
    }

    void
    SelectionManager::set_selections(const list<Selection *> selections) {
	mSelections = selections;

	mCacheValid = false;
    }

    void
    SelectionManager::toggle(Fig * fig) {
	// we have to clear the selections if this fig
	// is not on the same layer

	if(!mSelections.empty()) {
	    Fig * content = mSelections.front()->get_content();
	    if(content->get_layer() != fig->get_layer()) {
		clear();
	    }
	}

	// do the toggling

	fig->damage();

	if(contains(fig)) {
	    remove(fig);
	}
	else {
	    add(fig);
	}

	fig->damage();

	mCacheValid = false;
    }

    bool
    SelectionManager::contains(Fig * fig) const {
	return find_if(mSelections.begin(), mSelections.end(),
		       bind2nd(mem_fun(&Selection::is_content), fig)) != mSelections.end();
    }

    Selection *
    SelectionManager::make_selection_for(Fig * fig) {
	// use the fig's factory method to create the selection
	Selection * selection = fig->create_selection();

	return selection;
    }

    void
    SelectionManager::damage() {
	if(!mSelections.empty()) {
	    Drawing::Rectangle rect;

	    get_bounds(rect);

	    Fig * content = mSelections.front()->get_content();
	    Layer * layer = content->get_layer();

	    ASSERT(layer);

	    layer->damage(rect);
	}
    }

    //

    bool
    SelectionManager::mouse_moved(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_moved), mouse_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::mouse_dragged(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_dragged), mouse_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::mouse_entered(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_entered), mouse_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::mouse_exited(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_exited), mouse_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::mouse_pressed(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_pressed), mouse_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::mouse_released(const Toolkit::MouseEvent * mouse_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::mouse_released), mouse_event));

	return iter != mSelections.end();
    }

    //

    bool
    SelectionManager::key_pressed(const Toolkit::KeyEvent * key_event) {
	TRACE(9, int(key_event->get_key()) << " " << char(key_event->get_key()));

	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::key_pressed), key_event));

	return iter != mSelections.end();
    }

    bool
    SelectionManager::key_released(const Toolkit::KeyEvent * key_event) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::key_released), key_event));

	return iter != mSelections.end();
    }

    //

    Selection *
    SelectionManager::find_selection_at(const Drawing::Point & point) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd_vr(mem_fun(&Selection::contains), point));

	return iter == mSelections.end() ? 0 : *iter;
    }

    Selection *
    SelectionManager::find_selection_for(Fig * fig) {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd(mem_fun(&Selection::is_content), fig));

	return iter == mSelections.end() ? 0 : *iter;
    }

    bool
    SelectionManager::hit(const Drawing::Rectangle & rect) const {
	list<Selection *>::const_iterator iter =
	    find_if(mSelections.begin(), mSelections.end(),
		    bind2nd_vr(mem_fun(&Selection::hit), rect));

	return iter != mSelections.end();
    }

    void
    SelectionManager::hit_handle(const Drawing::Rectangle & rect, Handle & handle) const {
	if(mSelections.size() == 1) {
	    Selection * selection = mSelections.front();
	    selection->hit_handle(rect, handle);
	}
	else {
	    handle.set_index(-1);
	}
    }

    void
    SelectionManager::paint(Drawing::Drawable * drawable) const {
	const Drawing::Rectangle & clip = drawable->get_clip();

	if(clip) {
	    for_each_if(mSelections.rbegin(), mSelections.rend(),
			bind2nd_vr(mem_fun(&Selection::intersects), clip),
			bind2nd(mem_fun(&Selection::paint), drawable));
	}
	else {
	    for_each(mSelections.rbegin(), mSelections.rend(),
		     bind2nd(mem_fun(&Selection::paint), drawable));
	}
    }

    void
    SelectionManager::cleanup() {
	for_each(mSelections.begin(), mSelections.end(), mem_fun(&Selection::cleanup));
    }

    void
    SelectionManager::begin_interaction() {
	ASSERT(!mInInteraction);

	mInInteraction = true;

	for_each(mSelections.begin(), mSelections.end(), mem_fun(&Selection::begin_interaction));
    }

    void
    SelectionManager::end_interaction() {
	ASSERT(mInInteraction);

	begin_transaction();

	for_each(mSelections.begin(), mSelections.end(), mem_fun(&Selection::end_interaction));

	mInInteraction = false;

	end_transaction();
    }

} // namespace Diagram
