// This code is a modified port of GEF (Graphical Editing Framework)
// http://gef.tigris.org/
// GEF is copyrighted by the Regents of the University of California

// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "selection_resize.h"

#include "foundation/macros.h"
#include "drawing/point.h"
#include "drawing/rectangle.h"
#include "drawing/vector.h"
#include "drawing/drawable.h"

#include "handle.h"
#include "fig.h"

#include <algorithm>
using namespace std;

//
// NOTE: 1, 3, 4, 6 are not implemented
//
//   0-------1-------2
//   |               |
//   3               4
//   |               |
//   5-------6-------7
//

namespace Diagram {

    SelectionResize::SelectionResize(Fig * fig) : Selection(fig) {
	update_bounds_to_content();
    }

    SelectionResize::~SelectionResize() {
    }

    void
    SelectionResize::hit_handle(const Drawing::Rectangle & rect, Handle & handle) const {
	const Drawing::Rectangle & bounds = mContent->get_bounds();

	bool left_edge = rect & Drawing::Rectangle(bounds.get_x() - HAND_SIZE / 2,
						   bounds.get_y() - HAND_SIZE / 2,
						   HAND_SIZE,
						   bounds.get_h() + HAND_SIZE);

	bool right_edge = rect & Drawing::Rectangle(bounds.get_x() + bounds.get_w() - HAND_SIZE / 2,
						    bounds.get_y() - HAND_SIZE / 2,
						    HAND_SIZE,
						    bounds.get_h() + HAND_SIZE);

	bool top_edge = rect & Drawing::Rectangle(bounds.get_x() - HAND_SIZE / 2,
						  bounds.get_y() - HAND_SIZE / 2,
						  bounds.get_w() + HAND_SIZE,
						  HAND_SIZE);

	bool bottom_edge = rect & Drawing::Rectangle(bounds.get_x() - HAND_SIZE / 2,
						     bounds.get_y() + bounds.get_h() -HAND_SIZE / 2,
						     bounds.get_w() + HAND_SIZE,
						     HAND_SIZE);

	if(left_edge && top_edge) {
	    handle.set_index(0);
	    handle.set_instructions("Resize top left");
	}
	else if(left_edge && bottom_edge) {
	    handle.set_index(5);
	    handle.set_instructions("Resize bottom left");
	}
	else if(right_edge && top_edge) {
	    handle.set_index(2);
	    handle.set_instructions("Resize top right");
	}
	else if(right_edge && bottom_edge) {
	    handle.set_index(7);
	    handle.set_instructions("Resize bottom right");
	}
	else {
	    handle.set_index(-1);
	    handle.set_instructions("Move object(s)");
	}
    }

    void
    SelectionResize::drag_handle(const Drawing::Point & m, const Drawing::Point & an,
				 Handle & handle) {
	const Drawing::Vector & min_size = mContent->get_minimum_size();
	const Drawing::Rectangle & old_bounds = mContent->get_bounds();
	Drawing::Rectangle new_bounds = old_bounds;

	ASSERT(handle.get_index() != -1);

	if(handle.get_index() == 0) {
	    new_bounds.set_w(max(old_bounds.get_xw() - m.get_x() + 1, min_size.get_x()));
	    new_bounds.set_h(max(new_bounds.get_yh() - m.get_y() + 1, min_size.get_y()));
	    new_bounds.set_x(old_bounds.get_x() + old_bounds.get_w() - new_bounds.get_w());
	    new_bounds.set_y(old_bounds.get_y() + old_bounds.get_h() - new_bounds.get_h());
	}
	else if(handle.get_index() == 2) {
	    new_bounds.set_w(max(m.get_x() - old_bounds.get_x() + 1, min_size.get_x()));
	    new_bounds.set_h(max(new_bounds.get_yh() - m.get_y() + 1, min_size.get_y()));
	    new_bounds.set_y(old_bounds.get_y() + old_bounds.get_h() - new_bounds.get_h());
	}
	else if(handle.get_index() == 5) {
	    new_bounds.set_w(max(old_bounds.get_xw() - m.get_x() + 1, min_size.get_x()));
	    new_bounds.set_h(max(m.get_y() - new_bounds.get_y() + 1, min_size.get_y()));
	    new_bounds.set_x(old_bounds.get_x() + old_bounds.get_w() - new_bounds.get_w());
	}
	else if(handle.get_index() == 7) {
	    new_bounds.set_w(max(m.get_x() - old_bounds.get_x() + 1, min_size.get_x()));
	    new_bounds.set_h(max(m.get_y() - new_bounds.get_y() + 1, min_size.get_y()));
	}
	else {
	    ERROR("Unknown handle index: " << handle.get_index());
	}

	mContent->rebound(new_bounds);

	update_bounds_to_content();
    }

    void
    SelectionResize::paint(Drawing::Drawable * drawable) const {
	const Drawing::Rectangle & bounds = mContent->get_bounds();

	int x = bounds.get_x();
	int y = bounds.get_y();
	int w = bounds.get_w();
	int h = bounds.get_h();

	paint_rectangle(Drawing::Point(x, y), drawable);
	paint_rectangle(Drawing::Point(x + w - 1, y), drawable);
	paint_rectangle(Drawing::Point(x + w - 1, y + h - 1), drawable);
	paint_rectangle(Drawing::Point(x, y + h - 1), drawable);
    }

    void
    SelectionResize::update_bounds_to_content() {
	mBounds = mContent->get_bounds();
	mBounds.translate(Drawing::Vector(-HAND_SIZE / 2, -HAND_SIZE / 2));
	mBounds.grow(HAND_SIZE + 1, HAND_SIZE + 1);
    }

} // namespace Diagram
