// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "line.h"

#include "foundation/macros.h"

#include "point.h"
#include "vector.h"

#include <cmath>
#include <iostream>
using namespace std;

namespace Drawing {

    Line::Line(const Point & p1, const Point & p2) : mPoint(p1), mVector(p2 - p1) {
    }

    Line::Line(const Point & p, const Vector & v) : mPoint(p), mVector(v) {
	if(!mVector) {
            ERROR("Bad vector");
	    throw;	// FIXME
	}
    }

    bool
    Line::get_intercept(const Line & line, Point & point) const {
	const Vector & v1 = get_vector();
	const Vector & v2 = line.get_vector();
	const Point & p1 = get_point();
	const Point & p2 = line.get_point();

	int a = v2.get_x() * v1.get_y() - v1.get_x() * v2.get_y();

	if(a == 0) {
	    // the lines are parallel
	    return false;
	}
	else {
	    if(abs(v1.get_y()) > abs(v1.get_x())) {
		// line 1 is more vertical than horizontal
		// so we solve for y then substitute for x
		int b = p1.get_y() * v1.get_x() - p1.get_x() * v1.get_y();
		int c = p2.get_y() * v2.get_x() - p2.get_x() * v2.get_y();

		int j = v1.get_y() * c - v2.get_y() * b;
		point.set_y((int)rint(double(j) / double(a)));

		int k = v1.get_x() * (point.get_y() - p1.get_y());
		point.set_x(p1.get_x() + (int)rint(double(k) / double(v1.get_y())));
	    }
	    else {
		// line 1 is more horizontal than vertical
		// so we solve for x then substitute for y
		int b = (p2.get_y() - p1.get_y()) * v1.get_x() * v2.get_x();
		int c = v1.get_x() * v2.get_y() * p2.get_x();
		int d = v2.get_x() * v1.get_y() * p1.get_x();

		int j = b - c + d;
		point.set_x((int)rint(double(j) / double(a)));

		int k = v1.get_y() * (point.get_x() - p1.get_x());
		point.set_y(p1.get_y() + (int)rint(double(k) / double(v1.get_x())));
	    }

	    return true;
	}
    }

    double
    Line::get_distance_to(const Point & p0) const {
	const Point & p1 = mPoint;
	Point p2 = p1 + mVector;

	double numerator = abs((p2.get_y() - p1.get_y()) * (p0.get_x() - p1.get_x()) -
			    (p2.get_x() - p1.get_x()) * (p0.get_y() - p1.get_y()));
	double denominator = sqrt(double((p2.get_x() - p1.get_x()) * (p2.get_x() - p1.get_x()) +
					(p2.get_y() - p1.get_y()) * (p2.get_y() - p1.get_y())));

	return denominator < 1e-10 ? 1e10 : numerator / denominator;
    }

    ostream &
    operator << (ostream & ost, const Line & l) {
	const Point & p = l.get_point();
	const Vector & v = l.get_vector();

	ost << "(x, y, dx, dy): ("
	    << p.get_x() << ", " << p.get_y() << ", "
	    << v.get_x() << ", " << v.get_y() << ")";

	return ost;
    }

} // namespace Drawing
