// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "enumeration_type.h"

#include "mof/enumeration_type.h"

#include "foundation/macros.h"

#include "utility.h"

#include "common/cpp_stream.h"
#include "common/diff_stream.h"

#include <iostream>
#include <typeinfo>
using namespace std;

namespace Generator {

    EnumerationType::EnumerationType(MOF::EnumerationType * enumeration_type,
				     const string & model_name) : Object(),
        mEnumerationType(enumeration_type),
	mModelName(model_name),
	mDetailsLoaded(false) {
	ASSERT(enumeration_type);
    }

    EnumerationType::~EnumerationType() {
    }

    void
    EnumerationType::generate(const string & output_dir, const string & include_dir) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}

	string full_path = output_dir + "/" + mEnumerationTypeDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    EnumerationType::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mEnumerationTypeDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	generate_header_includes(hpp);

	hpp << "\nnamespace " << mModelName << " {\n";

	hpp << "\nclass " << mEnumerationTypeDetails.name << " : public REF::Attribute {";

	hpp << "\npublic:";

	generate_header_public_members(hpp);

	hpp << "\nprivate:";

	generate_header_private_members(hpp);

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    EnumerationType::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	generate_source_includes(cpp);

	cpp << "\nnamespace " << mModelName << " {\n";

	generate_source_statics(cpp);
	generate_source_body(cpp);

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    EnumerationType::generate_header_includes(Common::CppStream & hpp) {
	hpp << "\n#include \"reflective/attribute.h\"\n";

	if(mEnumerationTypeDetails.name == "Name") {
	    hpp << "\n#include <string>"
		<< "\nusing namespace std;\n";
	}
    }

    void
    EnumerationType::generate_header_public_members(Common::CppStream & hpp) {
	hpp << "\n" << mEnumerationTypeDetails.name << "();"
	    << "\n~" << mEnumerationTypeDetails.name << "();"
	    << "\n"
	    << "\nvoid ref_set_value(const string & value);"
	    << "\nvoid ref_get_value(string & value) const;"
	    << "\n"
	    << "\nconst string & ref_get_type() const;\n";

	hpp << "\nconst string & get_value() const;"
	    << "\nvoid set_value(const string & value) { ref_set_value(value); }\n";
    }

    void
    EnumerationType::generate_header_private_members(Common::CppStream & hpp) {
	hpp << "\nstatic const string TYPE;\n";

	hpp << "\nint mIndex;"
	    << "\nstatic const string TABLE[];"
	    << "\nstatic const int TABLE_SIZE;";
    }

    //

    void
    EnumerationType::generate_source_includes(Common::CppStream & cpp) {
	cpp << "\n#include \"" << mEnumerationTypeDetails.role << ".h\"\n"
	    << "\n#include \"reflective/exceptions.h\""
	    << "\n"
	    << "\n#include <sstream>"
	    << "\nusing namespace std;\n";
    }

    void
    EnumerationType::generate_source_statics(Common::CppStream & cpp) {
	cpp << "\nconst string " << mEnumerationTypeDetails.name << "::TYPE = \""
	    << mEnumerationTypeDetails.name << "\";\n";

	cpp << "\nconst string " << mEnumerationTypeDetails.name << "::TABLE[] = {";

	for(list<string>::iterator iter = mEnumerationTypeDetails.labels.begin();
	    iter != mEnumerationTypeDetails.labels.end();
	    ++iter) {
	    cpp << "\n\"" << *iter << "\",";
	}

	cpp << "\n};\n";

	cpp << "\nconst int " << mEnumerationTypeDetails.name << "::TABLE_SIZE = "
	    << "sizeof(TABLE) / sizeof(TABLE[0]);\n";
    }

    void
    EnumerationType::generate_source_body(Common::CppStream & cpp) {
	cpp << "\n" << mEnumerationTypeDetails.name << "::"
	    << mEnumerationTypeDetails.name << "() : "
	    << "REF::Attribute(), mIndex(0) {"
	    << "\n}\n";

	cpp << "\n" << mEnumerationTypeDetails.name << "::~"
	    << mEnumerationTypeDetails.name << "() {"
	    << "\n}\n";

	cpp << "\nvoid"
	    << "\n" << mEnumerationTypeDetails.name << "::ref_set_value(const string & value) {"
	    << "\nistringstream ist(value);"
	    << "\n"
	    << "\nstring literal;"
	    << "\nist >> literal;"
	    << "\n"
	    << "\nfor(int i = 0; i < TABLE_SIZE; ++i) {"
	    << "\nif(TABLE[i] == literal) {"
	    << "\nmIndex = i;"
	    << "\nreturn;"
	    << "\n}"
	    << "\n}"
	    << "\n"
	    << "\nthrow REF::BadValue(\"" << mEnumerationTypeDetails.name << ": \" + value);"
	    << "\n}\n";

	cpp << "\nvoid"
	    << "\n" << mEnumerationTypeDetails.name << "::ref_get_value(string & value) const {"
	    << "\nvalue += TABLE[mIndex];"
	    << "\n}\n";

	cpp << "\nconst string &"
	    << "\n" << mEnumerationTypeDetails.name << "::ref_get_type() const {"
	    << "\n  return TYPE;"
	    << "\n}\n";

	cpp << "\nconst string &"
	    << "\n" << mEnumerationTypeDetails.name << "::get_value() const {"
	    << "\nreturn TABLE[mIndex];"
	    << "\n}\n";
    }

    void
    EnumerationType::load_details() {
	ASSERT(mEnumerationType->get_name());
	ASSERT(mEnumerationType->get_labels());

	//
	// load the details of this primitive type
	//

	mEnumerationTypeDetails.enumeration_type = mEnumerationType;
	mEnumerationTypeDetails.name = mEnumerationType->get_name()->get_value();
	mEnumerationTypeDetails.role = name_to_role(mEnumerationTypeDetails.name);
	mEnumerationTypeDetails.variable = role_to_variable(mEnumerationTypeDetails.role);
	mEnumerationTypeDetails.qualified_name = mEnumerationType->get_qualified_name();
	mEnumerationTypeDetails.qualified_path =
	    name_to_path(mEnumerationTypeDetails.qualified_name);

	istringstream ist(mEnumerationType->get_labels()->get_value() + "\n");

	for(;;) {
	    string label;
	    ist >> label;

	    if(ist.good()) {
		mEnumerationTypeDetails.labels.push_back(label);
	    }
	    else {
		break;
	    }
	}
    }

} // namespace Generator
