// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "io.h"

#include "mof/class.h"
#include "mof/package.h"
#include "mof/model.h"
#include "mof/reference.h"
#include "mof/association_end.h"
#include "mof/attribute.h"
#include "mof/data_type.h"
#include "mof/primitive_type.h"
#include "mof/structure_type.h"
#include "mof/collection_type.h"
#include "mof/alias_type.h"
#include "mof/enumeration_type.h"

#include "common/file_utilities.h"
#include "common/diff_stream.h"
#include "common/cpp_stream.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include <typeinfo>
#include <iostream>
using namespace std;

namespace Generator {

    IO::IO(MOF::Package * root_package, const string & model_name) : Object(),
	mPackage(root_package),
	mModelName(model_name),
	mDetailsLoaded(false) {
	ASSERT(mPackage);
	ASSERT(!mPackage->get_container());	// make sure it is the root
	mPackage->ref();
    }

    IO::~IO() {
	mPackage->unref();
    }

    void
    IO::generate(const string & output_dir, const string & include_dir) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}
    }

    void
    IO::load_details() {
	REF::Context * context = mPackage->ref_get_context();
	MOF::Model * model = known_cast<MOF::Model *>(context);

	const set<MOF::DataType *> & data_types = model->get_all_of_type_data_type();

	for(set<MOF::DataType *>::const_iterator iter = data_types.begin();
	    iter != data_types.end();
	    ++iter) {
	    MOF::DataType * data_type = *iter;

	    DataTypeDetails details;

	    details.type = data_type;
	    details.name = data_type->get_name()->get_value();
	    details.role = name_to_role(details.name);
	    details.variable = role_to_variable(details.role);
	    details.qualified_name = data_type->get_qualified_name();
	    details.qualified_path = name_to_path(details.qualified_name);

	    mDataTypeDetails.push_back(details);
	}

	//

	const set<MOF::Class *> & classes = model->get_all_of_type_class();

	for(set<MOF::Class *>::const_iterator iter = classes.begin();
	    iter != classes.end();
	    ++iter) {
	    MOF::Class * klass = *iter;

	    ClassDetails details;

	    details.klass = klass;
	    details.name = klass->get_name()->get_value();
	    details.role = name_to_role(details.name);
	    details.variable = role_to_variable(details.role);
	    details.qualified_name = klass->get_qualified_name();
	    details.qualified_path = name_to_path(details.qualified_name);

	    const list<MOF::GeneralizableElement *> & supertypes = klass->get_supertypes();

	    for(list<MOF::GeneralizableElement *>::const_iterator iter = supertypes.begin();
		iter != supertypes.end();
		++iter) {
		MOF::GeneralizableElement * ge = *iter;
		MOF::Class * supertype = known_cast<MOF::Class *>(ge);

		SuperTypeDetails details2;

		ASSERT(supertype->get_name());
		details2.name = (supertype->get_name()->get_value());
		details2.role = name_to_role(details2.name);
		details2.variable = role_to_variable(details2.role);

		details.supertypes.push_back(details2);
	    }

	    const list<MOF::ModelElement *> & contained_elements = klass->get_contained_elements();

	    for(list<MOF::ModelElement *>::const_iterator iter = contained_elements.begin();
		iter != contained_elements.end();
		++iter) {
		MOF::ModelElement * contained_element = *iter;

		if(typeid(*contained_element) == typeid(MOF::Reference)) {
		    MOF::Reference * r = known_cast<MOF::Reference *>(contained_element);

		    ReferenceDetails details2;

		    details2.other_end = r->get_referenced_end();
		    details2.other_end_navigable = details2.other_end->is_navigable()->get_value();
		    details2.other_end_name = details2.other_end->get_name()->get_value();
		    details2.other_end_role = name_to_role(details2.other_end_name);
		    details2.other_end_variable = role_to_variable(details2.other_end_role);
		    details2.other_type = known_cast<MOF::Class *>(details2.other_end->get_type());
		    details2.other_type_name = details2.other_type->get_name()->get_value();
		    details2.other_type_role = name_to_role(details2.other_type_name);

		    details2.this_end = details2.other_end->get_other_end();
		    details2.this_end_aggregate =
			details2.this_end->get_aggregation()->get_value() == "composite";
		    //

		    const MOF::MultiplicityType * other_mult =
			details2.other_end->get_multiplicity();

		    ASSERT(other_mult);

		    ASSERT(other_mult->is_unique()->get_value());		// all we can handle

		    if(other_mult->get_upper()->get_value() == 1) {
			details2.other_cardinality = UNARY;
		    }
		    else {
			details2.other_cardinality = MULTI;
		    }

		    if(other_mult->get_ordered()->get_value()) {
			details2.other_ordering = ORDERED;
		    }
		    else {
			//details2.other_ordering = UNORDERED;		FIXME
			details2.other_ordering = ORDERED;
		    }

		    details.references.push_back(details2);
		}
		else if(typeid(*contained_element) == typeid(MOF::Attribute)) {
		    MOF::Attribute * a = known_cast<MOF::Attribute *>(contained_element);

		    AttributeDetails details2;

		    details2.name = a->get_name()->get_value();
		    details2.role = name_to_role(details2.name);
		    details2.variable = role_to_variable(details2.role);
		    details2.member = name_to_member(details2.name);

		    details2.type = known_cast<MOF::DataType *>(a->get_type());
		    details2.type_name = details2.type->get_name()->get_value();

		    details.attributes.push_back(details2);
		}
	    }

	    ASSERT(klass->is_abstract());

	    if(klass->is_abstract()->get_value()) {
		mAbstractClasses.push_back(details);
	    }
	    else {
		mConcreteClasses.push_back(details);
	    }

	    mAllClasses.push_back(details);
	}
    }

} // namespace Generator
