// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "primitive_type.h"

#include "mof/primitive_type.h"

#include "foundation/macros.h"

#include "utility.h"

#include "common/cpp_stream.h"
#include "common/diff_stream.h"

#include <iostream>
#include <typeinfo>
using namespace std;

namespace Generator {

    PrimitiveType::PrimitiveType(MOF::PrimitiveType * alias_type,
				 const string & model_name) : Object(),
        mPrimitiveType(alias_type),
	mModelName(model_name),
	mDetailsLoaded(false) {
	ASSERT(alias_type);
    }

    PrimitiveType::~PrimitiveType() {
    }

    void
    PrimitiveType::generate(const string & output_dir, const string & include_dir) {
	if(!mDetailsLoaded) {
	    load_details();
	    mDetailsLoaded = true;
	}

	string full_path = output_dir + "/" + mPrimitiveTypeDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    PrimitiveType::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mPrimitiveTypeDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	generate_header_includes(hpp);

	hpp << "\nnamespace " << mModelName << " {\n";

	hpp << "\nclass " << mPrimitiveTypeDetails.name << " : public REF::Attribute {";

	hpp << "\npublic:";

	generate_header_public_members(hpp);

	hpp << "\nprivate:";

	generate_header_private_members(hpp);

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    PrimitiveType::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	generate_source_includes(cpp);

	cpp << "\nnamespace " << mModelName << " {\n";

	generate_source_statics(cpp);
	generate_source_body(cpp);

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    PrimitiveType::generate_header_includes(Common::CppStream & hpp) {
	hpp << "\n#include \"reflective/attribute.h\"\n";

	if(mPrimitiveTypeDetails.name == "Name") {
	    hpp << "\n#include <string>"
		<< "\nusing namespace std;\n";
	}
    }

    void
    PrimitiveType::generate_header_public_members(Common::CppStream & hpp) {
	hpp << "\n" << mPrimitiveTypeDetails.name << "();"
	    << "\n~" << mPrimitiveTypeDetails.name << "();"
	    << "\n"
	    << "\nvoid ref_set_value(const string & value);"
	    << "\nvoid ref_get_value(string & value) const;"
	    << "\n"
	    << "\nconst string & ref_get_type() const;\n";

	if(mPrimitiveTypeDetails.lightweight) {
	    hpp << "\n" << mPrimitiveTypeDetails.c_type << " get_value() const { return mValue; }"
		<< "\nvoid set_value(" << mPrimitiveTypeDetails.c_type
		<< " value) { mValue = value; }\n";
	}
	else {
	    hpp << "\nconst " << mPrimitiveTypeDetails.c_type
		<< " & get_value() const { return mValue; }"
		<< "\nvoid set_value(const " << mPrimitiveTypeDetails.c_type
		<< " & value) { mValue = value; }\n";
	}

	hpp << "\nconst string & get_type() const;\n";
    }

    void
    PrimitiveType::generate_header_private_members(Common::CppStream & hpp) {
	hpp << "\n" << mPrimitiveTypeDetails.c_type << " mValue;";

	hpp << "\nstatic const string TYPE;";
    }

    //

    void
    PrimitiveType::generate_source_includes(Common::CppStream & cpp) {
	cpp << "\n#include \"" << mPrimitiveTypeDetails.role << ".h\"\n"
	    << "\n#include \"reflective/exceptions.h\""
	    << "\n"
	    << "\n#include <sstream>"
	    << "\nusing namespace std;\n";
    }

    void
    PrimitiveType::generate_source_statics(Common::CppStream & cpp) {
	cpp << "\nconst string " << mPrimitiveTypeDetails.name << "::TYPE = \""
	    << mPrimitiveTypeDetails.name << "\";\n";
    }

    void
    PrimitiveType::generate_source_body(Common::CppStream & cpp) {
	cpp << "\n" << mPrimitiveTypeDetails.name << "::" << mPrimitiveTypeDetails.name << "() : "
	    << "REF::Attribute(), mValue(" << mPrimitiveTypeDetails.initial_value << ") {"
	    << "\n}\n";

	cpp << "\n" << mPrimitiveTypeDetails.name << "::~" << mPrimitiveTypeDetails.name << "() {"
	    << "\n}\n";

	if(mPrimitiveTypeDetails.name == "Name") {
	    cpp << "\nvoid"
		<< "\n" << mPrimitiveTypeDetails.name << "::ref_set_value(const string & value) {"
		<< "\nmValue = value;"
		<< "\n}\n";
	}
	else {
	    cpp << "\nvoid"
		<< "\n" << mPrimitiveTypeDetails.name << "::ref_set_value(const string & value) {"
		<< "\nistringstream ist(value + \"\\n\");"
		<< "\nist >> mValue;"
		<< "\nif(!ist.good()) {"
		<< "\nthrow REF::BadValue(\"" << mPrimitiveTypeDetails.name << ": \" + value);"
		<< "\n}"
		<< "\n}\n";
	}

	cpp << "\nvoid"
	    << "\n" << mPrimitiveTypeDetails.name << "::ref_get_value(string & value) const {"
	    << "\nostringstream ost;"
	    << "\nost << mValue;"
	    << "\nvalue += ost.str();"
	    << "\n}\n";

	cpp << "\nconst string &"
	    << "\n" << mPrimitiveTypeDetails.name << "::ref_get_type() const {"
	    << "\n  return TYPE;"
	    << "\n}\n";
    }

    void
    PrimitiveType::load_details() {
	ASSERT(mPrimitiveType->get_name());

	//
	// load the details of this primitive type
	//

	mPrimitiveTypeDetails.primitive_type = mPrimitiveType;
	mPrimitiveTypeDetails.name = mPrimitiveType->get_name()->get_value();
	mPrimitiveTypeDetails.role = name_to_role(mPrimitiveTypeDetails.name);
	mPrimitiveTypeDetails.variable = role_to_variable(mPrimitiveTypeDetails.role);
	mPrimitiveTypeDetails.qualified_name = mPrimitiveType->get_qualified_name();
	mPrimitiveTypeDetails.qualified_path = name_to_path(mPrimitiveTypeDetails.qualified_name);

	if(mPrimitiveTypeDetails.name == "Boolean") {
	    mPrimitiveTypeDetails.c_type = "bool";
	    mPrimitiveTypeDetails.lightweight = true;
	    mPrimitiveTypeDetails.initial_value = "false";
	}
	else if(mPrimitiveTypeDetails.name == "Integer") {
	    mPrimitiveTypeDetails.c_type = "int";
	    mPrimitiveTypeDetails.lightweight = true;
	    mPrimitiveTypeDetails.initial_value = "0";
	}
	else if(mPrimitiveTypeDetails.name == "Long") {
	    mPrimitiveTypeDetails.c_type = "long";
	    mPrimitiveTypeDetails.lightweight = true;
	    mPrimitiveTypeDetails.initial_value = "0";
	}
	else if(mPrimitiveTypeDetails.name == "Float") {
	    mPrimitiveTypeDetails.c_type = "float";
	    mPrimitiveTypeDetails.lightweight = true;
	    mPrimitiveTypeDetails.initial_value = "0.0f";
	}
	else if(mPrimitiveTypeDetails.name == "Double") {
	    mPrimitiveTypeDetails.c_type = "double";
	    mPrimitiveTypeDetails.lightweight = true;
	    mPrimitiveTypeDetails.initial_value = "0.0";
	}
	else if(mPrimitiveTypeDetails.name == "Name") {
	    mPrimitiveTypeDetails.c_type = "string";
	    mPrimitiveTypeDetails.lightweight = false;
	    mPrimitiveTypeDetails.initial_value = "\"\"";
	}
	else {
	    FATAL("Unknown primitive: " << mPrimitiveTypeDetails.name);
	}
    }

} // namespace Generator
