// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "root_package.h"

#include "handler.h"
#include "writer.h"

#include "mof/model.h"
#include "mof/package.h"
#include "mof/class.h"
#include "mof/data_type.h"

#include "common/diff_stream.h"
#include "common/cpp_stream.h"
#include "common/file_utilities.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include <iostream>
using namespace std;

namespace Generator {

    RootPackage::RootPackage(MOF::Package * package,
			     const string & model_name) : Package(package, model_name) {
    }

    RootPackage::~RootPackage() {
    }

    void
    RootPackage::generate(const string & output_dir, const string & include_dir, int depth) {
	Package::generate(output_dir, include_dir, depth);

	string full_path = output_dir + "/" + mPackageDetails.path;

	Common::DiffStream hpp_stream(full_path + "/" + mPackageDetails.path + ".h");
	ASSERT(hpp_stream);

	Common::DiffStream cpp_stream(full_path + "/" + mPackageDetails.path + ".cc");
	ASSERT(cpp_stream);

	//

	Common::CppStream cpp(cpp_stream);
	Common::CppStream hpp(hpp_stream);

	//

	cout << "    Generating model context..." << endl;

	generate_context(hpp, cpp);

	cout << "    Generating XMI writer..." << endl;

	Writer * writer = new Writer(mPackage, mModelName);
	writer->generate(output_dir, include_dir);
	writer->unref();

	cout << "    Generating SAX handler..." << endl;

	Handler * handler = new Handler(mPackage, mModelName);
	handler->generate(output_dir, include_dir);
	handler->unref();
    }

    void
    RootPackage::load_details() {
	Package::load_details();

	MOF::Model * model = known_cast<MOF::Model *>(mPackage->ref_get_context());

	for(set<MOF::Class *>::const_iterator iter = model->get_all_of_type_class().begin();
	    iter != model->get_all_of_type_class().end();
	    ++iter) {
	    MOF::Class * klass = *iter;

	    ASSERT(klass->get_name());
	    ASSERT(klass->is_abstract());

	    ClassDetails details;

	    details.klass = klass;
	    details.name = klass->get_name()->get_value();
	    details.path = name_to_path(details.name);
	    details.qualified_name = klass->get_qualified_name();
	    details.qualified_path = name_to_path(details.qualified_name);
	    details.is_abstract = klass->is_abstract()->get_value();

	    mAllClassDetails.push_back(details);
	}

	for(set<MOF::DataType *>::const_iterator iter = model->get_all_of_type_data_type().begin();
	    iter != model->get_all_of_type_data_type().end();
	    ++iter) {
	    MOF::DataType * data_type = *iter;

	    DataTypeDetails details;

	    details.data_type = data_type;
	    details.name = data_type->get_name()->get_value();
	    details.path = name_to_path(details.name);
	    details.qualified_name = data_type->get_qualified_name();
	    details.qualified_path = name_to_path(details.qualified_name);

	    mAllDataTypeDetails.push_back(details);
	}

	//
	//
	//

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    string path = name_to_path(trim_path(iter->qualified_name));
	    mAllSources.insert(path + ".cc");
	    mAllHeaders.insert(path + ".h");
	}

	for(vector<DataTypeDetails>::iterator iter = mAllDataTypeDetails.begin();
	    iter != mAllDataTypeDetails.end();
	    ++iter) {
	    string path = name_to_path(trim_path(iter->qualified_name));
	    mAllSources.insert(path + ".cc");
	    mAllHeaders.insert(path + ".h");
	}

	mAllSources.insert(mPackageDetails.path + ".cc");
	mAllHeaders.insert(mPackageDetails.path + ".h");

	mAllSources.insert("writer.cc");
	mAllHeaders.insert("writer.h");

	mAllSources.insert("handler.cc");
	mAllHeaders.insert("handler.h");

	ASSERT(!mAllSources.empty());
	ASSERT(!mAllHeaders.empty());

	mBuiltSources.insert(mPackageDetails.path + ".cc");
	mBuiltHeaders.insert(mPackageDetails.path + ".h");

	mBuiltSources.insert("writer.cc");
	mBuiltHeaders.insert("writer.h");

	mBuiltSources.insert("handler.cc");
	mBuiltHeaders.insert("handler.h");
    }

    void
    RootPackage::generate_makefile(Common::DiffStream & mam) {
	Package::generate_makefile(mam);

	mam << "\nAUTOMAKE_OPTIONS = subdir-objects\n";

	mam << "\nnoinst_LTLIBRARIES = libdachshund-" << mPackageDetails.path << ".la\n";

	mam << "\nlibdachshund_" << mPackageDetails.path << "_la_SOURCES =";

	for(set<string>::iterator iter = mAllSources.begin();
	    iter != mAllSources.end();
	    ++iter) {
	    mam << " \\\n\t" << *iter;
	}

	mam << "\n";
    }

    void
    RootPackage::generate_cvsignore(Common::DiffStream & ign) {
	Package::generate_cvsignore(ign);

	ign << mPackageDetails.path << ".h\n"
	    << mPackageDetails.path << ".cc\n";
    }

    void
    RootPackage::generate_context(Common::CppStream & hpp, Common::CppStream & cpp) {
	string guard = path_to_guard(mPackageDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard
	    << "\n"
	    << "\n#include \"reflective/context.h\""
	    << "\n"
	    << "\n#include <set>"
	    << "\nusing namespace std;"
	    << "\n"
	    << "\nnamespace " << mPackageDetails.name << " {\n"
	    << "\nusing namespace REF;\n";

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    hpp << "\nclass " << iter->name << ";";
	}

	hpp << "\n\nclass " << mPackageDetails.name << " : public Context {"
	    << "\npublic:"
	    << "\n" << mPackageDetails.name << "();\n"
	    << "\nREF::Class * ref_create_class(const string & class_name);\n"
	    << "\nREF::Attribute * ref_create_attribute(const string & attribute_name);\n";

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    string class_name = iter->name;
	    string class_role = name_to_role(class_name);

	    hpp << "\nconst set<" << class_name << " *> & get_all_of_type_" << class_role
		<< "() {"
		<< "\nreturn mAllOfType" << class_name << ";"
		<< "\n}\n";

	    hpp << "\nvoid internal_add_" << class_role << "(" << class_name << " * element);"
		<< "\nvoid internal_remove_" << class_role << "(" << class_name
		<< " * element);\n";
	}

	hpp << "\n"
	    << "\nprotected:"
	    << "\n~" << mPackageDetails.name << "();\n";

	hpp << "\nprivate:";

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    hpp << "\nset<" << iter->name << " *> mAllOfType" << iter->name << ";";
	}

	hpp << "\n};"
	    << "\n"
	    << "\n} // namespace " << mPackageDetails.name
	    << "\n"
	    << "\n#endif // " << guard << "\n";

	//
	//
	//

	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	cpp << "\n#include \"" << mPackageDetails.path << ".h\"\n";

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    if(!iter->is_abstract) {
		cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	    }
	}

	cpp << "\n";

	for(vector<DataTypeDetails>::iterator iter = mAllDataTypeDetails.begin();
	    iter != mAllDataTypeDetails.end();
	    ++iter) {
	    cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	}

	cpp << "\n"
	    << "\n#include \"reflective/exceptions.h\""
	    << "\n"
	    << "\n#include \"foundation/macros.h\""
	    << "\n";

	cpp << "\nnamespace " << mPackageDetails.name << " {";

	// constructor

	cpp << "\n"
	    << "\n" << mPackageDetails.name << "::" << mPackageDetails.name << "() : Context() {"
	    << "\n}\n";

	// destructor

	cpp << "\n"
	    << "\n" << mPackageDetails.name << "::~" << mPackageDetails.name << "() {"
	    << "\n}\n";

	// ref_create_class

	cpp << "\nREF::Class *"
	    << "\n" << mPackageDetails.name << "::ref_create_class(const string & class_name) {";

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {
	    if(!iter->is_abstract) {
		string name = iter->name;
		string role = name_to_role(name);

		cpp << "\nif(class_name == \"" << role << "\") {"
		    << "\nreturn new " << name << "();"
		    << "\n}\n";
	    }
	}

	cpp << "\nERROR(\"No such class: \" + class_name);"
	    << "\nthrow REF::BadType(class_name);"
	    << "\n}\n";

	// ref_create_attribute

	cpp << "\nREF::Attribute *"
	    << "\n" << mPackageDetails.name
	    << "::ref_create_attribute(const string & attribute_name) {";

	for(vector<DataTypeDetails>::iterator iter = mAllDataTypeDetails.begin();
	    iter != mAllDataTypeDetails.end();
	    ++iter) {
	    string name = iter->name;
	    string role = name_to_role(name);

	    cpp << "\nif(attribute_name == \"" << role << "\") {"
		<< "\nreturn new " << name << "();"
		<< "\n}\n";
	}

	cpp << "\nERROR(\"No such attribute: \" + attribute_name);"
	    << "\nthrow REF::BadType(attribute_name);"
	    << "\n}\n";

	//

	for(vector<ClassDetails>::iterator iter = mAllClassDetails.begin();
	    iter != mAllClassDetails.end();
	    ++iter) {

	    string name = iter->name;
	    string role = name_to_role(name);

	    cpp << "\nvoid"
		<< "\n" << mPackageDetails.name << "::internal_add_" << role
		<< "(" << name << " * element) {"
		<< "\nmAllOfType" << name << ".insert(element);"
		<< "\n}\n";

	    cpp << "\nvoid"
		<< "\n" << mPackageDetails.name << "::internal_remove_" << role
		<< "(" << name << " * element) {"
		<< "\nmAllOfType" << name << ".erase(element);"
		<< "\n}\n";
	}

	cpp << "\n} // namespace " << mPackageDetails.name << "\n";
    }

} // namespace Generator
