// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "writer.h"

#include "mof/class.h"
#include "mof/package.h"
#include "mof/model.h"
#include "mof/reference.h"
#include "mof/association_end.h"
#include "mof/attribute.h"
#include "mof/data_type.h"
#include "mof/primitive_type.h"
#include "mof/structure_type.h"
#include "mof/collection_type.h"
#include "mof/alias_type.h"
#include "mof/enumeration_type.h"
#include "mof/structure_field.h"

#include "common/file_utilities.h"
#include "common/diff_stream.h"
#include "common/cpp_stream.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "utility.h"

#include <typeinfo>
#include <iostream>
using namespace std;

namespace Generator {

    Writer::Writer(MOF::Package * root_package,
		   const string & model_name) : IO(root_package, model_name) {
    }

    Writer::~Writer() {
    }

    void
    Writer::generate(const string & output_dir, const string & include_dir) {
	IO::generate(output_dir, include_dir);

	string full_path = output_dir + "/" + mWriterDetails.qualified_path;

	Common::DiffStream hpp_stream(full_path + ".h");
	ASSERT(hpp_stream);
	Common::CppStream hpp(hpp_stream);

	Common::DiffStream cpp_stream(full_path + ".cc");
	ASSERT(cpp_stream);
	Common::CppStream cpp(cpp_stream);

	generate_header(hpp);
	generate_source(cpp);
    }

    void
    Writer::generate_header(Common::CppStream & hpp) {
	string guard = path_to_guard(mWriterDetails.qualified_path);

	hpp << get_prologue()
	    << "\n#ifndef " << guard
	    << "\n#define " << guard << "\n";

	hpp << "\n#include \"foundation/object.h\"\n";

	hpp << "\n#include \"reflective/class.h\"\n";

	hpp << "\n#include <string>"
	    << "\n#include <map>"
	    << "\n#include <set>"
	    << "\n#include <iosfwd>"
	    << "\n#include <typeinfo>"
	    << "\nusing namespace std;\n";

	hpp << "\nnamespace XML {"
	    << "\n"
	    << "\nclass Stream;"
	    << "\n"
	    << "\n} // namespace XML\n";

	hpp << "\nnamespace " << mModelName << " {\n";

	// class forward declarations

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    hpp << "\nclass " << iter->name << ";";
	}

	hpp << "\n";

	// data_type forward declarations

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    hpp << "\nclass " << iter->name << ";";
	}

	hpp << "\n";

	//

	hpp << "\nclass Writer : public Object {";

	hpp << "\npublic:";

	hpp << "\nWriter(REF::Class * root);"
	    << "\n~Writer();\n";

	//

	hpp << "\nvoid write(ostream & ost);\n";

	hpp << "\nprotected:";

	hpp << "\nvoid invoke_class_handler(REF::Class * object, bool reference);"
	    << "\n"
	    << "\nvoid invoke_data_type_handler(const REF::Attribute * object);\n";

	// uncasted concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    hpp << "\nvoid write_" << iter->role << "(REF::Class * object, bool reference);";
	}

	hpp << "\n"
	    << "\n//"
	    << "\n//"
	    << "\n//"
	    << "\n";

	// casted concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    hpp << "\nvoid write_" << iter->role << "(" << iter->name << " * " << iter->variable
		<< ", bool reference);";
	}

	hpp << "\n"
	    << "\n//"
	    << "\n//"
	    << "\n//"
	    << "\n";

	// class contents

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    hpp << "\nvoid write_" << iter->role << "(" << iter->name << " * " << iter->variable
		<< ", set<string> & visited);";
	}

	hpp << "\n"
	    << "\n//"
	    << "\n//"
	    << "\n//"
	    << "\n";

	// uncasted data types

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    hpp << "\nvoid write_" << iter->role << "(const REF::Attribute * "
		<< iter->variable << ");";
	}

	hpp << "\n";

	// casted data types

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    hpp << "\nvoid write_" << iter->role << "(const " << iter->name << " * "
		<< iter->variable << ");";
	}

	hpp << "\n";

	//

	hpp << "\nprivate:";

	hpp << "\nREF::Class * mRoot;"
	    << "\ntypedef const type_info * TypeInfo;"
	    << "\ntypedef void (Writer::*ClassHandler)(REF::Class *, bool);"
	    << "\ntypedef void (Writer::*DataTypeHandler)(const REF::Attribute *);"
	    << "\n"
	    << "\nstruct Compare { bool operator() (TypeInfo a, TypeInfo b) "
	    << "const { return a->before(*b); } };"
	    << "\n"
	    << "\nmap<TypeInfo, ClassHandler, Compare> mClassHandlerMap;"
	    << "\nmap<TypeInfo, DataTypeHandler, Compare> mDataTypeHandlerMap;"
	    << "\n"
	    << "\nXML::Stream * mStream;";

	//

	hpp << "\n};\n";

	hpp << "\n} // namespace " << mModelName << "\n";

	hpp << "\n#endif // " << guard << "\n";
    }

    void
    Writer::generate_source(Common::CppStream & cpp) {
	cpp << get_prologue()
	    << "\n#ifdef HAVE_CONFIG_H"
	    << "\n#include \"config.h\""
	    << "\n#endif // HAVE_CONFIG_H\n";

	cpp << "\n#include \"writer.h\"\n";

	cpp << "\n#include \"xml/stream.h\"\n";

	cpp << "\n#include \"foundation/macros.h\""
	    << "\n#include \"foundation/utility.h\"\n";

	// include all the headers of the classes in the model

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	}

	cpp << "\n";

	// include all the headers of the data_types in the model

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    cpp << "\n#include \"" << iter->qualified_path << ".h\"";
	}

	cpp << "\n";

	cpp << "\n#include <iostream>"
	    << "\nusing namespace std;\n";

	//

	cpp << "\nnamespace " << mModelName << " {\n";

	//

	cpp << "\nWriter::Writer(REF::Class * root) : Object(), mRoot(root), mStream(0) {"
	    << "\nASSERT(mRoot);"
	    << "\nmRoot->ref();";

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    cpp << "\nmClassHandlerMap[&typeid(" << iter->name << ")] = &Writer::write_"
		<< iter->role << ";";
	}

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    cpp << "\nmDataTypeHandlerMap[&typeid(" << iter->name << ")] = &Writer::write_"
		<< iter->role << ";";
	}

	cpp << "\n}\n";

	//

	cpp << "\nWriter::~Writer() {"
	    << "\nmRoot->unref();"
	    << "\n}\n";

	//

	cpp << "\nvoid"
	    << "\nWriter::write(ostream & ost) {"
	    << "\nmStream = new XML::Stream(ost, \"XMI\", XML::DTDK_SYSTEM, \"xmi.dtd\");"
	    /*
	    << "\n*mStream << XML::XmlBeginOpen(\"XMI\");"
	    << "\n*mStream << XML::XmlBeginClose();"
	    */
	    << "\ninvoke_class_handler(mRoot, false);"
	    /*
	    << "\n*mStream << XML::XmlEnd();"
	    */
	    << "\ndelete mStream;"
	    << "\n}\n";

	//

	cpp << "\nvoid"
	    << "\nWriter::invoke_class_handler(REF::Class * object, bool reference) {"
	    << "\nmap<TypeInfo, ClassHandler, Compare>::iterator iter = "
	    << "mClassHandlerMap.find(&typeid(*object));"
	    << "\n"
	    << "\nif(iter == mClassHandlerMap.end()) {"
	    << "\nERROR(\"No handler for class\");"
	    << "\n}"
	    << "\nelse {"
	    << "\nClassHandler handler = iter->second;"
	    << "\n(this->*handler)(object, reference);"
	    << "\n}"
	    << "\n}\n";

	cpp << "\nvoid"
	    << "\nWriter::invoke_data_type_handler(const REF::Attribute * object) {"
	    << "\nmap<TypeInfo, DataTypeHandler, Compare>::iterator iter = "
	    << "mDataTypeHandlerMap.find(&typeid(*object));"
	    << "\n"
	    << "\nif(iter == mDataTypeHandlerMap.end()) {"
	    << "\nERROR(\"No handler for attribute\");"
	    << "\n}"
	    << "\nelse {"
	    << "\nDataTypeHandler handler = iter->second;"
	    << "\n(this->*handler)(object);"
	    << "\n}"
	    << "\n}\n";

	// uncasted concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    cpp << "\nvoid"
		<< "\nWriter::write_" << iter->role << "(REF::Class * object, bool reference) {";

	    cpp << "\nwrite_" << iter->role << "(known_cast<" << iter->name
		<< " *>(object), reference);";

	    cpp << "\n}\n";
	}

	cpp << "\n";

	// casted concrete classes

	for(list<ClassDetails>::iterator iter = mConcreteClasses.begin();
	    iter != mConcreteClasses.end();
	    ++iter) {
	    cpp << "\nvoid"
		<< "\nWriter::write_" << iter->role << "(" << iter->name << " * " << iter->variable
		<< ", bool reference) {";

	    cpp << "\n*mStream << XML::XmlBeginOpen(\"" << mModelName << ":" << iter->name << "\");"
		<< "\n"
		<< "\nif(reference) {"
		<< "\n*mStream << XML::XmlAttribute(\"xmi.idref\", "
		<< iter->variable << "->ref_get_uuid().get_string());"
		<< "\n*mStream << XML::XmlBeginClose();"
		<< "\n}"
		<< "\nelse {"
		<< "\n*mStream << XML::XmlAttribute(\"xmi.id\", "
		<< iter->variable << "->ref_get_uuid().get_string());"
		<< "\n*mStream << XML::XmlBeginClose();"
		<< "\nset<string> visited;"
		<< "\nwrite_" << iter->role << "(" << iter->variable << ", visited);"
		<< "\n}"
		<< "\n"
		<< "\n*mStream << XML::XmlEnd();";

	    cpp << "\n}\n";
	}

	cpp << "\n";

	// class contents

	for(list<ClassDetails>::iterator iter = mAllClasses.begin();
	    iter != mAllClasses.end();
	    ++iter) {
	    cpp << "\nvoid"
		<< "\nWriter::write_" << iter->role << "(" << iter->name << " * " << iter->variable
		<< ", set<string> & visited) {";

	    cpp << "\nif(visited.insert(\"" << iter->name << "\").second) {";

	    for(list<AttributeDetails>::iterator iter2 = iter->attributes.begin();
		iter2 != iter->attributes.end();
		++iter2) {
		string tag_name = mModelName + ":" + iter->name + "." + iter2->name;

		cpp << "\nconst " << iter2->type_name << " * " << iter2->variable << " = "
		    << iter->variable << "->get_" << iter2->role << "();"
		    << "\n"
		    << "\nif(" << iter2->variable << ") {"
		    << "\n*mStream << XML::XmlBeginOpen(\"" << tag_name << "\");"
		    << "\n*mStream << XML::XmlBeginClose();"
		    << "\n"
		    << "\ninvoke_data_type_handler(" << iter2->variable << ");"
		    << "\n"
		    << "\n*mStream << XML::XmlEnd();\n"
		    << "\n}\n";
	    }

	    for(list<SuperTypeDetails>::iterator iter2 = iter->supertypes.begin();
		iter2 != iter->supertypes.end();
		++iter2) {
		cpp << "\nWriter::write_" << iter2->role << "(" << iter->variable << ", visited);";
	    }

	    for(list<ReferenceDetails>::iterator iter2 = iter->references.begin();
		iter2 != iter->references.end();
		++iter2) {
		if(!iter2->other_end_navigable) {
		    continue;
		}

		string tag_name = mModelName + ":" + iter->name + "." + iter2->other_end_name;

		if(iter2->other_cardinality == UNARY) { 
		    cpp << "\n" << iter2->other_type_name << " * "
			<< iter2->other_end_variable << " = "
			<< iter->variable << "->get_" << iter2->other_end_role << "();"
			<< "\n"
			<< "\nif(" << iter2->other_end_variable << ") {"
			<< "\n*mStream << XML::XmlBeginOpen(\"" << tag_name << "\");"
			<< "\n*mStream << XML::XmlBeginClose();"
			<< "\n"
			<< "\ninvoke_class_handler(" << iter2->other_end_variable
			<< ", " << (iter2->this_end_aggregate ? "false" : "true") << ");"
			<< "\n"
			<< "\n*mStream << XML::XmlEnd();\n"
			<< "\n}\n";
		}
		else if(iter2->other_cardinality == MULTI) {
		    if(iter2->other_ordering == ORDERED) {
			cpp << "\nconst list<" << iter2->other_type_name << " *> & "
			    << iter2->other_end_variable << " = "
			    << iter->variable << "->get_" << iter2->other_end_role << "();"
			    << "\n"
			    << "\nif(!" << iter2->other_end_variable << ".empty()) {"
			    << "\n*mStream << XML::XmlBeginOpen(\"" << tag_name << "\");"
			    << "\n*mStream << XML::XmlBeginClose();"
			    << "\n"
			    << "\nfor(list<" << iter2->other_type_name << " *>::const_iterator "
			    << "iter = " << iter2->other_end_variable << ".begin();"
			    << "\niter != " << iter2->other_end_variable << ".end();"
			    << "\n++iter) {"
			    << "\ninvoke_class_handler(*iter, "
			    << (iter2->this_end_aggregate ? "false" : "true") << ");"
			    << "\n}"
			    << "\n"
			    << "\n*mStream << XML::XmlEnd();\n"
			    << "\n}\n";
		    }
		    else if(iter2->other_ordering == UNORDERED) {
			cpp << "\nconst set<" << iter2->other_type_name << " *> & "
			    << iter2->other_end_variable << " = "
			    << iter->variable << "->get_" << iter2->other_end_role << "();"
			    << "\n"
			    << "\nif(!" << iter2->other_end_variable << ".empty()) {"
			    << "\n*mStream << XML::XmlBeginOpen(\"" << tag_name << "\");"
			    << "\n*mStream << XML::XmlBeginClose();"
			    << "\n"
			    << "\nfor(set<" << iter2->other_type_name << " *>::const_iterator "
			    << "iter = " << iter2->other_end_variable << ".begin();"
			    << "\niter != " << iter2->other_end_variable << ".end();"
			    << "\n++iter) {"
			    << "\ninvoke_class_handler(*iter, "
			    << (iter2->this_end_aggregate ? "false" : "true") << ");"
			    << "\n}"
			    << "\n"
			    << "\n*mStream << XML::XmlEnd();\n"
			    << "\n}\n";
		    }
		    else {
			ASSERT_NOT_REACHED();
		    }
		}
		else {
		    ASSERT_NOT_REACHED();
		}
	    }

	    cpp << "\n}";

	    cpp << "\n}\n";
	}

	// uncasted data types

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    cpp << "\nvoid"
		<< "\nWriter::write_" << iter->role << "(const REF::Attribute * "
		<< iter->variable << ") {"
		<< "\nwrite_" << iter->role << "(known_cast<const " << iter->name << " *>("
		<< iter->variable << "));"
		<< "\n}\n";
	}

	// casted data types

	for(list<DataTypeDetails>::iterator iter = mDataTypeDetails.begin();
	    iter != mDataTypeDetails.end();
	    ++iter) {
	    cpp << "\nvoid"
		<< "\nWriter::write_" << iter->role << "(const " << iter->name << " * "
		<< iter->variable << ") {";

	    if(typeid(*iter->type) == typeid(MOF::PrimitiveType)) {
		cpp << "\n*mStream << " << iter->variable << "->get_value();";
	    }
	    else if(typeid(*iter->type) == typeid(MOF::EnumerationType)) {
		cpp << "\n*mStream << " << iter->variable << "->get_value();";
	    }
	    else if(typeid(*iter->type) == typeid(MOF::StructureType)) {
		MOF::StructureType * structure_type = known_cast<MOF::StructureType *>(iter->type);

		const list<MOF::ModelElement *> & contained_elements =
		    structure_type->get_contained_elements();

#if 0
		for(list<MOF::ModelElement *>::const_iterator iter2 = contained_elements.begin();
		    iter2 != contained_elements.end();
		    ++iter2) {
		    MOF::ModelElement * contained_element = *iter2;
		    MOF::StructureField * structure_field =
			known_cast<MOF::StructureField *>(contained_element);

		    string tag_name = mModelName + ":" + iter->name + "." + structure_field->get_name()->get_value();
		    MOF::Classifier * type = structure_field->get_type();

		    cpp << "\n*mStream << XML::XmlBeginOpen(\"" << tag_name << "\");"
			<< "\n*mStream << XML::XmlBeginClose();"
			<< "\n";

		    cpp << "\nwrite_" << type->get_name()->get_value()
			<< "(" << iter->variable << "->get_" << structure_field->get_name()->get_value() << ");";

		    cpp << "\n"
			<< "\n*mStream << XML::XmlEnd();"
			<< "\n";
		}
#endif
	    }
	    else if(typeid(*iter->type) == typeid(MOF::CollectionType)) {
		NYI();
	    }
	    else if(typeid(*iter->type) == typeid(MOF::AliasType)) {
		NYI();
	    }
	    else {
		ASSERT_NOT_REACHED();
	    }

	    cpp << "\n}\n";
	}

	//

	cpp << "\n} // namespace " << mModelName << "\n";
    }

    void
    Writer::load_details() {
	IO::load_details();

	mWriterDetails.qualified_path = name_to_path(mPackage->get_qualified_name()) + "/writer";
    }

} // namespace Generator
