// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "canvas.h"

#include "gtk_canvas.h"     // A C specialisation of a GTK+ widget

#include "image.h"
#include "text.h"
#include "palette.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "toolkit/mouse_event.h"
#include "toolkit/canvas_event.h"
#include "toolkit/key_event.h"

#include <gtk/gtksignal.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkhbox.h>

#include <algorithm>
#include <functional>
#include <iostream>
#include <cstring>
using namespace std;

namespace GTK {

    const unsigned int Canvas::SCROLL_TIMEOUT = 50;
    const int Canvas::SCROLL_THRESHOLD = 10;

    Canvas::Canvas(int width, int height) : Toolkit::Canvas(), Widget(),
	mDownButton(0),
	mTimeoutId(0),
	mTest(0) {
        mBaseWidget = gtk_scrolled_window_new(0, 0);
        gtk_widget_set_size_request(mBaseWidget, 340, 400);

	mCanvas = gtk_canvas_new(width, height);

	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(mBaseWidget), mCanvas);

	//

        mHandlers[HID_EXPOSE] =
            g_signal_connect_after(G_OBJECT(mCanvas), "expose_event",
                                   G_CALLBACK(&Canvas::static_expose_event), this);

        mHandlers[HID_CONFIGURE] =
            g_signal_connect_after(G_OBJECT(mCanvas), "configure_event",
                                   G_CALLBACK(&Canvas::static_configure_event), this);

        mHandlers[HID_MOTION_NOTIFY] =
            g_signal_connect_after(G_OBJECT(mCanvas), "motion_notify_event",
                                   G_CALLBACK(&Canvas::static_motion_notify_event),
                                   this);

        mHandlers[HID_ENTER_NOTIFY] =
            g_signal_connect_after(G_OBJECT(mCanvas), "enter_notify_event",
                                   G_CALLBACK(&Canvas::static_enter_notify_event),
                                   this);

        mHandlers[HID_LEAVE_NOTIFY] =
            g_signal_connect_after(G_OBJECT(mCanvas), "leave_notify_event",
                                   G_CALLBACK(&Canvas::static_leave_notify_event),
                                   this);

        mHandlers[HID_BUTTON_PRESS] =
            g_signal_connect_after(G_OBJECT(mCanvas), "button_press_event",
                                   G_CALLBACK(&Canvas::static_button_press_event),
                                   this);

        mHandlers[HID_BUTTON_RELEASE] =
            g_signal_connect_after(G_OBJECT(mCanvas), "button_release_event",
                                   G_CALLBACK(&Canvas::static_button_release_event),
                                   this);

        mHandlers[HID_KEY_PRESS] =
            g_signal_connect_after(G_OBJECT(mCanvas), "key_press_event",
                                   G_CALLBACK(&Canvas::static_key_press_event),
                                   this);

        mHandlers[HID_KEY_RELEASE] =
            g_signal_connect_after(G_OBJECT(mCanvas), "key_release_event",
                                   G_CALLBACK(&Canvas::static_key_release_event),
                                   this);

	//

	mClipStack.push(Drawing::Rectangle(0, 0, 0, 0));

	//

	gtk_widget_show(mCanvas);

	post_init();
    }

    Canvas::~Canvas() {
	if(mBaseWidget) {
	    for(int i = 0; i < HID_LAST; ++i) {
		g_signal_handler_disconnect(G_OBJECT(mCanvas), mHandlers[i]);
	    }
	}
    }

    void
    Canvas::refresh() {
	GdkRectangle rect = {
	    0,
	    0,
	    get_size().get_x(),
	    get_size().get_y()
	};

	gtk_canvas_damage(GTK_CANVAS(mCanvas), &rect);
    }

    void
    Canvas::damage(const Drawing::Rectangle & rect1) {
	GdkRectangle rect2 = {
	    rect1.get_x() << mTest,
	    rect1.get_y() << mTest,
	    ((rect1.get_w() - 1) << mTest) + 1,
	    ((rect1.get_h() - 1) << mTest) + 1
	};

	gtk_canvas_damage(GTK_CANVAS(mCanvas), &rect2);
    }

    void
    Canvas::set_cursor(Toolkit::Cursor cursor) {
	if(GTK_WIDGET_REALIZED(mCanvas)) {
	    GdkCursorType cursor_type;

	    switch(cursor) {
	    case Toolkit::C_DEFAULT:
		cursor_type = GDK_ARROW;
		break;
	    case Toolkit::C_CROSSHAIR:
		cursor_type = GDK_CROSSHAIR;
		break;
	    case Toolkit::C_DOT:
		cursor_type = GDK_DOT;
		break;
	    case Toolkit::C_HAND:
		cursor_type = GDK_HAND1;
		break;
	    default:
		ASSERT_NOT_REACHED();
	    }

	    GdkCursor * gdk_cursor = gdk_cursor_new(cursor_type);

	    GdkWindow * window = mCanvas->window;

	    gdk_window_set_cursor(window, gdk_cursor);

	    gdk_cursor_unref(gdk_cursor);
	}
    }

    void
    Canvas::set_color(const string & color_string) {
	gtk_canvas_set_color(GTK_CANVAS(mCanvas), color_string.c_str());
    }

    void
    Canvas::set_line_width(int line_width) {
	gtk_canvas_set_line_width(GTK_CANVAS(mCanvas), line_width);
    }

    void
    Canvas::set_line_style(Drawing::LineStyle line_style) {
	GdkLineStyle gdk_line_style;

	switch(line_style) {
	case Drawing::LS_SOLID:
	    gdk_line_style = GDK_LINE_SOLID;
	    break;
	case Drawing::LS_DASHED:
	    gdk_line_style = GDK_LINE_ON_OFF_DASH;
	    break;
	case Drawing::LS_DOTTED:
	    gdk_line_style = GDK_LINE_DOUBLE_DASH;
	    break;
	default:
	    ASSERT_NOT_REACHED();
	}

	gtk_canvas_set_line_style(GTK_CANVAS(mCanvas), gdk_line_style);
    }

    //
    //
    //

    void
    Canvas::draw_point(const Drawing::Point & point) {
	gtk_canvas_draw_point(GTK_CANVAS(mCanvas), point.get_x() << mTest, point.get_y() << mTest);
    }

    void
    Canvas::draw_rectangle(const Drawing::Rectangle & rect1, bool filled) {
	GdkRectangle rect2 = {
	    rect1.get_x() << mTest,
	    rect1.get_y() << mTest,
	    ((rect1.get_w() - 1) << mTest) + 1,
	    ((rect1.get_h() - 1) << mTest) + 1
	};

	gtk_canvas_draw_rectangle(GTK_CANVAS(mCanvas), &rect2, filled ? TRUE : FALSE);
    }

    void
    Canvas::draw_line(const Drawing::Point & p1, const Drawing::Point & p2) {
	gtk_canvas_draw_line(GTK_CANVAS(mCanvas),
			    p1.get_x() << mTest,
			    p1.get_y() << mTest,
			    p2.get_x() << mTest,
			    p2.get_y() << mTest);
    }

    void
    Canvas::draw_oval(const Drawing::Rectangle & rect1, bool filled) {
	GdkRectangle rect2 = {
	    rect1.get_x() << mTest,
	    rect1.get_y() << mTest,
	    ((rect1.get_w() - 1) << mTest) + 1,
	    ((rect1.get_h() - 1) << mTest) + 1
	};

	gtk_canvas_draw_oval(GTK_CANVAS(mCanvas),
			    rect2.x, rect2.y, rect2.width, rect2.height,
			    filled ? TRUE : FALSE);
    }

    void
    Canvas::draw_polygon(const vector<Drawing::Point> & points, bool filled) {
	// only resize the buffer if it is smaller then required
	if(mPointBuffer.size() < points.size()) {
	    mPointBuffer.resize(points.size());
	}

	for(unsigned int i = 0; i < points.size(); ++i) {
	    mPointBuffer[i].x = points[i].get_x() << mTest;
	    mPointBuffer[i].y = points[i].get_y() << mTest;
	}

	gtk_canvas_draw_polygon(GTK_CANVAS(mCanvas),
				&mPointBuffer[0],
				points.size(),		// note that we use the size of the arg
				filled ? TRUE : FALSE);
    }

    void
    Canvas::draw_lines(const vector<Drawing::Point> & points) {
	// only resize the buffer if it is smaller then required
	if(mPointBuffer.size() < points.size()) {
	    mPointBuffer.resize(points.size());
	}

	for(unsigned int i = 0; i < points.size(); ++i) {
	    mPointBuffer[i].x = points[i].get_x() << mTest;
	    mPointBuffer[i].y = points[i].get_y() << mTest;
	}

	gtk_canvas_draw_lines(GTK_CANVAS(mCanvas),
			      &mPointBuffer[0],
			      points.size());		// note that we use the size of the arg
    }

    void
    Canvas::draw_text(Drawing::Text * _text, const Drawing::Point & point) {
	Text * text = known_cast<Text *>(_text);

	gtk_canvas_draw_layout(GTK_CANVAS(mCanvas),
			       text->get_layout(),
			       point.get_x(), point.get_y());
    }

    void
    Canvas::draw_image(Drawing::Image * _image, const Drawing::Point & point) {
	Image * image = known_cast<Image *>(_image);

	gtk_canvas_draw_pixbuf(GTK_CANVAS(mCanvas),
			       image->get_pixbuf(),
			       point.get_x(), point.get_y());
    }

    //
    //
    //

    void
    Canvas::push_clip(const Drawing::Rectangle & rect) {
	if(mClipStack.size() == 1) {
	    mClipStack.push(rect);
	}
	else {
	    mClipStack.push(rect & mClipStack.top());
	}

	const Drawing::Rectangle & rect1 = mClipStack.top();

	GdkRectangle rect2 = {
	    rect1.get_x() << mTest,
	    rect1.get_y() << mTest,
	    ((rect1.get_w() - 1) << mTest) + 1,
	    ((rect1.get_h() - 1) << mTest) + 1
	};

	gtk_canvas_set_clip(GTK_CANVAS(mCanvas), &rect2);
    }

    void
    Canvas::pop_clip() {
	mClipStack.pop();

        ASSERT(!mClipStack.empty());

	if(mClipStack.size() == 1) {
	    gtk_canvas_unset_clip(GTK_CANVAS(mCanvas));
	}
	else {
	    const Drawing::Rectangle & rect1 = mClipStack.top();

	    GdkRectangle rect2 = {
		rect1.get_x() << mTest,
		rect1.get_y() << mTest,
		((rect1.get_w() - 1) << mTest) + 1,
		((rect1.get_h() - 1) << mTest) + 1
	    };

	    gtk_canvas_set_clip(GTK_CANVAS(mCanvas), &rect2);
	}
    }

    const Drawing::Rectangle &
    Canvas::get_clip() const {
	return mClipStack.top();
    }

    const Drawing::Vector &
    Canvas::get_size() const {
	return mSize;
    }

    bool
    Canvas::expose_event(const GdkEventExpose * event) {
	TRACE(10, "");

	Drawing::Rectangle rect(event->area.x >> mTest,
				event->area.y >> mTest,
				((event->area.width - 1) >> mTest) + 1,
				((event->area.height - 1) >> mTest) + 1);

	Toolkit::CanvasEvent canvas_event(this, rect);

	return exposed(&canvas_event);
    }

    bool
    Canvas::configure_event(const GdkEventConfigure * event) {
	TRACE(10, "");

	mSize.set(event->width >> mTest, event->height >> mTest);

	return true;
    }

    bool
    Canvas::motion_notify_event(const GdkEventMotion * event) {
	TRACE(10, "");

	bool dragged = mDownButton != 0;
	bool handled;

	if(dragged) {
	    Toolkit::MouseEvent mouse_event(Toolkit::MET_DRAGGED,
					    Drawing::Point(int(event->x + 0.5) >> mTest,
							   int(event->y + 0.5) >> mTest),
					    this,
					    translate_modifiers(event->state),
					    event->time);

	    handled = mouse_dragged(&mouse_event);
	}
	else {
	    Toolkit::MouseEvent mouse_event(Toolkit::MET_MOVED,
					    Drawing::Point(int(event->x + 0.5) >> mTest,
							   int(event->y + 0.5) >> mTest),
					    this,
					    translate_modifiers(event->state),
					    event->time);

	    handled = mouse_moved(&mouse_event);
	}

	//return handled;
        return true;
    }

    bool
    Canvas::enter_notify_event(const GdkEventCrossing * event) {
	TRACE(10, "");

	gtk_widget_grab_focus(mCanvas);		// FIXME TEMPORARY HACK

	bool dragged = mDownButton != 0;
	bool handled = false;

	if(mTimeoutId != 0) {
	    gtk_timeout_remove(mTimeoutId);
	    mTimeoutId = 0;
	}

	if(event->mode == GDK_CROSSING_NORMAL && !dragged) {
	    Toolkit::MouseEvent mouse_event(Toolkit::MET_ENTERED,
					    Drawing::Point(int(event->x + 0.5) >> mTest,
							   int(event->y + 0.5) >> mTest),
					    this,
					    translate_modifiers(event->state),
					    event->time);

	    handled = mouse_entered(&mouse_event);
	}

	//return handled;
        return true;
    }

    bool
    Canvas::leave_notify_event(const GdkEventCrossing * event) {
	TRACE(10, "");

	bool dragged = mDownButton != 0;
	bool handled = false;

	if(dragged && mTimeoutId == 0) {
#if 0
	    // this is making things hard to debug for now (FIXME put it back in)
	    mTimeoutId = gtk_timeout_add(SCROLL_TIMEOUT, &Canvas::static_timeout, this);
#endif
	}

	switch(event->mode) {
	case GDK_CROSSING_NORMAL:
	    if(!dragged) {
		Toolkit::MouseEvent mouse_event(Toolkit::MET_EXITED,
						Drawing::Point(int(event->x + 0.5) >> mTest,
							       int(event->y + 0.5) >> mTest),
						this,
						translate_modifiers(event->state),
						event->time);

		handled = mouse_exited(&mouse_event);
	    }
	    break;
	case GDK_CROSSING_GRAB:
	    // We might get here if a popup menu grabbed the cursor from us
	    // We have to reset the mouse button
	    mDownButton = 0;
	    break;
	case GDK_CROSSING_UNGRAB:
	    // we might get here if a popup menu returned the grab to us
	    break;
	}

	//return handled;
        return true;
    }

    bool
    Canvas::button_press_event(const GdkEventButton * event) {
	TRACE(10, "");

	if(mDownButton == 0) {
	    // FIXME I don't think grabs are necessary
	    // because the cursor is grabbed anyway when the button 
	    // is pressed
	    //gtk_grab_add(mCanvas);

	    mDownButton = event->button;

	    Toolkit::MouseEvent mouse_event(Toolkit::MET_PRESSED,
					    Drawing::Point(int(event->x + 0.5) >> mTest,
							   int(event->y + 0.5) >> mTest),
					    this,
					    translate_modifiers(event->state),
					    event->time,
					    event->button,
					    translate_presses(event->type));

	    /*return*/ mouse_pressed(&mouse_event); return true;
	}
	else {
	    //return false;
            return true;
	}
    }

    bool
    Canvas::button_release_event(const GdkEventButton * event) {
	TRACE(10, "");

	if(mDownButton == event->button) {
	    //gtk_grab_remove(mCanvas);

	    if(mTimeoutId != 0) {
		gtk_timeout_remove(mTimeoutId);
		mTimeoutId = 0;
	    }

	    mDownButton = 0;

	    Toolkit::MouseEvent mouse_event(Toolkit::MET_RELEASED,
					    Drawing::Point(int(event->x + 0.5) >> mTest,
							   int(event->y + 0.5) >> mTest),
					    this,
					    translate_modifiers(event->state),
					    event->time,
					    event->button,
					    translate_presses(event->type));

	    /*return*/ mouse_released(&mouse_event); return true;
	}
	else {
	    //return false;
            return true;
	}
    }

    bool
    Canvas::key_press_event(const GdkEventKey * event) {
	TRACE(10, "");

	Toolkit::KeyEvent key_event(Toolkit::KET_PRESSED,
				    event->keyval,
				    this,
				    translate_modifiers(event->state),
				    event->time);

	return key_pressed(&key_event);
    }

    bool
    Canvas::key_release_event(const GdkEventKey * event) {
	TRACE(10, "");

	Toolkit::KeyEvent key_event(Toolkit::KET_RELEASED,
				    event->keyval,
				    this,
				    translate_modifiers(event->state),
				    event->time);

	return key_released(&key_event);
    }

    bool
    Canvas::timeout() {
	GtkScrolledWindow * scrolled_window = GTK_SCROLLED_WINDOW(mBaseWidget);

	GtkAdjustment * hadjustment = gtk_scrolled_window_get_hadjustment(scrolled_window);
	GtkAdjustment * vadjustment = gtk_scrolled_window_get_vadjustment(scrolled_window);

	ASSERT(hadjustment);
	ASSERT(vadjustment);

	gdouble old_hvalue = gtk_adjustment_get_value(hadjustment);
	gdouble old_vvalue = gtk_adjustment_get_value(vadjustment);

	GtkWidget * viewport = GTK_BIN(mBaseWidget)->child;
	int w = viewport->allocation.width;
	int h = viewport->allocation.height;

	int x, y;
	GdkModifierType mask;

	gdk_window_get_pointer(viewport->window, &x, &y, &mask);

	int hoffset = x < 0 ? x / 3 : x > w ? (x - w) / 3 : 0;
	int voffset = y < 0 ? y / 3 : y > h ? (y - h) / 3 : 0;

	gdouble new_hvalue = CLAMP(old_hvalue + hoffset,
				hadjustment->lower,
				hadjustment->upper - hadjustment->page_size);

	gdouble new_vvalue = CLAMP(old_vvalue + voffset,
				vadjustment->lower,
				vadjustment->upper - vadjustment->page_size);

	bool need_event = false;

	if(old_hvalue != new_hvalue) {
	    gtk_adjustment_set_value(hadjustment, new_hvalue);
	    need_event = true;
	}

	if(old_vvalue != new_vvalue) {
	    gtk_adjustment_set_value(vadjustment, new_vvalue);
	    need_event = true;
	}

	if(need_event) {
	    // create a dummy event for the apparent motion caused
	    // by the viewport changing

	    GdkEventMotion event;

	    memset(&event, 0, sizeof(event));

	    gdk_window_get_pointer(mCanvas->window, &x, &y, &mask);

	    event.type = GDK_MOTION_NOTIFY;
	    event.send_event = TRUE;
	    event.window = mCanvas->window;
	    event.x = x;
	    event.y = y;
	    event.time = 0;
	    event.state = mask;
	    event.is_hint = 0;

	    motion_notify_event(&event);
	}

	return true;
    }

    gint
    Canvas::static_expose_event(GtkWidget * w, GdkEventExpose * event, Canvas * canvas) {
	return canvas->expose_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_configure_event(GtkWidget * w, GdkEventConfigure * event, Canvas * canvas) {
	return canvas->configure_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_motion_notify_event(GtkWidget * w, GdkEventMotion * event, Canvas * canvas) {
	return canvas->motion_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_enter_notify_event(GtkWidget * w, GdkEventCrossing * event, Canvas * canvas) {
	return canvas->enter_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_leave_notify_event(GtkWidget * w, GdkEventCrossing * event, Canvas * canvas) {
	return canvas->leave_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_button_press_event(GtkWidget * w, GdkEventButton * event, Canvas * canvas) {
	return canvas->button_press_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_button_release_event(GtkWidget * w, GdkEventButton * event, Canvas * canvas) {
	return canvas->button_release_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_key_press_event(GtkWidget * w, GdkEventKey * event, Canvas * canvas) {
	return canvas->key_press_event(event) ? TRUE : FALSE;
    }

    gint
    Canvas::static_key_release_event(GtkWidget * w, GdkEventKey * event, Canvas * canvas) {
	return canvas->key_release_event(event) ? TRUE : FALSE;
    }

    gboolean
    Canvas::static_timeout(gpointer data) {
	Canvas * canvas = static_cast<Canvas *>(data);
	return canvas->timeout();
    }

} // namespace GTK
