// Dachshund,			      -*- mode: C++; c-basic-offset: 4; -*-
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef GTK_DOCUMENT_H
#define GTK_DOCUMENT_H

#include "toolkit/document.h"
#include "widget.h"

//#include <stack>
//#include <vector>
using namespace std;

#include <gtk/gtktextview.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextiter.h>
#include <gtk/gtktextmark.h>


namespace GTK {
  
    class Tag;
    class Mark;

    class Document : public Toolkit::Document, public Widget {
        class LocalChange;
	friend class Document::LocalChange;

    public:
	Document();
	~Document();

	GtkTextBuffer * get_text_buffer() { return mTextBuffer; }
    protected:
	class MarkTag;

    private:

	GtkTextBuffer *	mTextBuffer;

	// Working marks used to replace text
	GtkTextMark *	mMarkRG;		// right hand gravity
	GtkTextMark *	mMarkLG;		// left hand gravity

	// Our copy of standard marks
	GtkTextMark *	mGtkTextMarkInsert;
	GtkTextMark *	mGtkTextMarkChangeBegin;
	GtkTextMark *	mGtkTextMarkChangeEnd;
	Toolkit::Mark *	mMarkInsert;
	Toolkit::Mark *	mMarkChangeBegin;
	Toolkit::Mark *	mMarkChangeEnd;
	Toolkit::Mark *	mMarkSelection;

	// Change related variables, maintained by changed()
	bool		mHaveChange;
	Toolkit::Tag *	mChangeBlock;
	Toolkit::Tag *	mChangeField;
	Toolkit::Mark *	mMarkBlock;

	// Insert stack used by insert_* methods
	list<MarkTag *>	mInsertStack;

	// Temporary marks for inserting fields and blocks
	list<GtkTextMark *> mTemporaryMarks;


	// Flag to mask local changes from change signals
	bool	mLocalChange;

  public:

	//------------------------------------------------------------
	// Functions dealing with the special marks
	//
	Toolkit::Mark * get_insert_mark();
	Toolkit::Mark * get_change_mark();
	Toolkit::Mark * get_selection_mark();


	//------------------------------------------------------------
	// Methods that implement an insertion type model for the document.
	
	void insert_block_begin(Toolkit::Mark * location, Toolkit::Mark * block_mark, 
				Toolkit::Tag * block);
	void insert_block_end(Toolkit::Mark * location, Toolkit::Tag * block);

	void insert_field_begin(Toolkit::Mark * location, Toolkit::Tag * field);
	void insert_field_end(Toolkit::Mark * location, Toolkit::Tag * block);

	void insert_field_text(Toolkit::Mark * location, Toolkit::Tag * field, 
			       const char * text);
	void insert_text(Toolkit::Mark * location, const char * text);
	void insert_image(Toolkit::Mark * location, Drawing::Image * image);
	void insert_widget(Toolkit::Mark * location, Toolkit::Widget * widget);


	//------------------------------------------------------------
	// Collection of methods used to maintain the document.

	void get_field_text(Toolkit::Mark * location, Toolkit::Tag * field, 
			    string & result);
    
	void replace_field_text(Toolkit::Mark * location, Toolkit::Tag * field, 
				const char * new_text);
	void replace_field_image(Toolkit::Mark * location, Toolkit::Tag * field, 
				 Drawing::Image * image);
	void replace_field_widget(Toolkit::Mark * location, Toolkit::Tag * field, 
				  Toolkit::Widget * widget);
    
	void delete_field(Toolkit::Mark * location, Toolkit::Tag * field);
	void delete_block(Toolkit::Mark * block_mark, Toolkit::Tag * block);

	void clear();

	//------------------------------------------------------------
	// Debug methods for sanity and consistancy checking

	//void verify_layout(Toolkit::Mark * location, const list<Toolkit::Tag *> & layout);


    protected: // Internal functions
	GtkTextMark *	allocate_temp_mark(GtkTextIter * iter);
	void		release_temp_mark(GtkTextMark * mark);


	// Methods to mask change events when *we* are changing the text buffer
	class LocalChange {
	    Document * mDoc;
	public:
	    LocalChange(Document * doc) : mDoc(doc) {
		mDoc->begin_local_change();
	    }
	    ~LocalChange() {
		mDoc->end_local_change();
	    }
	};

	void	begin_local_change() { mLocalChange = true; }
	void	end_local_change() { mLocalChange = false; }


	// Convenience and sanity checking methods
	
    public:
	// States driven by various signals
	void	changed_check();
	void	changes_sync();

	/*
	bool configure_event(const GdkEventConfigure * event);
	bool motion_notify_event(const GdkEventMotion * event);
	bool enter_notify_event(const GdkEventCrossing * event);
	bool leave_notify_event(const GdkEventCrossing * event);
	bool button_press_event(const GdkEventButton * event);
	bool button_release_event(const GdkEventButton * event);
	bool key_press_event(const GdkEventKey * event);
	bool key_release_event(const GdkEventKey * event);
	bool timeout();

	static gint static_configure_event(GtkWidget * widget, GdkEventConfigure * event,
					   Textview * textview);
	static gint static_motion_notify_event(GtkWidget * widget, GdkEventMotion * event,
					       Textview * textview);
	static gint static_enter_notify_event(GtkWidget * widget, GdkEventCrossing * event,
					      Textview * textview);
	static gint static_leave_notify_event(GtkWidget * widget, GdkEventCrossing * event,
					      Textview * textview);
	static gint static_button_press_event(GtkWidget * widget, GdkEventButton * event,
					      Textview * textview);
	static gint static_button_release_event(GtkWidget * widget, GdkEventButton * event,
						Textview * textview);
	static gint static_key_press_event(GtkWidget * widget, GdkEventKey * event,
					   Textview * textview);
	static gint static_key_release_event(GtkWidget * widget, GdkEventKey * event,
					     Textview * textview);

	static gboolean static_timeout(gpointer data);

	enum {
	    HID_CONFIGURE,
	    HID_MOTION_NOTIFY,
	    HID_ENTER_NOTIFY,
	    HID_LEAVE_NOTIFY,
	    HID_BUTTON_PRESS,
	    HID_BUTTON_RELEASE,
	    HID_KEY_PRESS,
	    HID_KEY_RELEASE,
	    HID_LAST
	};

	guint mHandlers[HID_LAST];

	unsigned int mDownButton;
	unsigned int mTimeoutId;

	int mTest;

	static const unsigned int SCROLL_TIMEOUT;
	static const int SCROLL_THRESHOLD;
	*/	
    };

} // namespace GTK

#endif // GTK_DOCUMENT_H
