// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef CONFIG_H
#include "config.h"
#endif // CONFIG_H

#include "gtk_canvas.h"

#include <gdk/gdk.h>

#include <gtk/gtkmain.h>

#include <string.h>
#include <stdlib.h>

#define BUFFER_SIZE 256

enum {
    PROP_0,
    PROP_WIDTH,
    PROP_HEIGHT
};

static void gtk_canvas_class_init(GtkCanvasClass * klass);
static void gtk_canvas_init(GtkCanvas * canvas);
static void gtk_canvas_realize(GtkWidget * widget);
static void gtk_canvas_destroy(GtkObject * object);
static gboolean gtk_canvas_idle_handler(gpointer data);
static guint gtk_canvas_string_hash(gconstpointer key);
static gboolean gtk_canvas_string_cmp(gconstpointer a, gconstpointer b);
static void gtk_canvas_color_hash_free(gpointer key, gpointer value, gpointer user_data);
static void gtk_canvas_send_configure(GtkCanvas * canvas);
static void gtk_canvas_size_request(GtkWidget * widget, GtkRequisition * requisition);
static void gtk_canvas_size_allocate(GtkWidget * widget, GtkAllocation * allocation);
static gboolean gtk_canvas_focus_in_event(GtkWidget * widget, GdkEventFocus * event);
static gboolean gtk_canvas_focus_out_event(GtkWidget * widget, GdkEventFocus * event);
static void gtk_canvas_grab_focus(GtkWidget * widget);
static gboolean gtk_canvas_focus(GtkWidget * widget, GtkDirectionType direction);


static GObject * gtk_canvas_constructor(GType type,
					guint n_construct_properties,
					GObjectConstructParam * construct_params);
static void gtk_canvas_set_property(GObject * object,
				    guint prop_id,
				    const GValue * value,
				    GParamSpec * pspec);
static void gtk_canvas_get_property(GObject * object,
				    guint prop_id,
				    GValue * value,
				    GParamSpec * pspec);

static GObjectClass * parent_class = NULL;

GtkType
gtk_canvas_get_type(void) {
    static GtkType canvas_type = 0;

    if (!canvas_type)
    {
	static const GtkTypeInfo canvas_info =
	{
	    "GtkCanvas",
	    sizeof (GtkCanvas),
	    sizeof (GtkCanvasClass),
	    (GtkClassInitFunc) gtk_canvas_class_init,
	    (GtkObjectInitFunc) gtk_canvas_init,
	    /* reserved_1 */ NULL,
	    /* reserved_2 */ NULL,
	    (GtkClassInitFunc)NULL,
	};

	canvas_type = gtk_type_unique (GTK_TYPE_DRAWING_AREA, &canvas_info);
    }

    return canvas_type;
}

static void
gtk_canvas_class_init(GtkCanvasClass * klass) {
    GtkDrawingAreaClass * drawing_area_class = GTK_DRAWING_AREA_CLASS(klass);
    GtkWidgetClass * widget_class = GTK_WIDGET_CLASS(drawing_area_class);
    GtkObjectClass * object_class = GTK_OBJECT_CLASS(widget_class);
    GObjectClass * g_object_class = G_OBJECT_CLASS(object_class);

    widget_class->realize = &gtk_canvas_realize;
    widget_class->size_request = &gtk_canvas_size_request;
    widget_class->size_allocate = &gtk_canvas_size_allocate;
    widget_class->focus_in_event = &gtk_canvas_focus_in_event;
    widget_class->focus_out_event = &gtk_canvas_focus_out_event;
    widget_class->grab_focus = &gtk_canvas_grab_focus;
    widget_class->focus = &gtk_canvas_focus;

    object_class->destroy = &gtk_canvas_destroy;

    g_object_class->constructor = &gtk_canvas_constructor;
    g_object_class->set_property = &gtk_canvas_set_property;
    g_object_class->get_property = &gtk_canvas_get_property;

    parent_class = g_type_class_peek_parent(klass);

    g_object_class_install_property(g_object_class,
				    PROP_WIDTH,
				    g_param_spec_int("width",
						     "Width",
						     "Width of the canvas.",
						     0,
						     G_MAXINT,
						     1,
						     G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
    g_object_class_install_property(g_object_class,
				    PROP_HEIGHT,
				    g_param_spec_int("height",
						     "Height",
						     "Height of the canvas.",
						     0,
						     G_MAXINT,
						     1,
						     G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
}

static void
gtk_canvas_init(GtkCanvas * canvas) {
    GtkWidget * widget;

    widget = GTK_WIDGET(canvas);

    GTK_WIDGET_SET_FLAGS(widget, GTK_CAN_FOCUS);

    canvas->gc = NULL;
    canvas->layout = NULL;
    canvas->idle_id = 0;
    canvas->grabbed = FALSE;

    canvas->width = 1;
    canvas->height = 1;

    canvas->damage.x = 0;
    canvas->damage.y = 0;
    canvas->damage.width = 0;
    canvas->damage.height = 0;

    canvas->color_table = g_hash_table_new(&gtk_canvas_string_hash, &gtk_canvas_string_cmp);

    canvas->current_color_string = g_new(gchar, BUFFER_SIZE);
    g_assert(canvas->current_color_string);
    canvas->current_color_string[0] = '\0';
    canvas->current_color_string[BUFFER_SIZE - 1] = '\0';

    canvas->current_line_style = GDK_SOLID;
}

GtkWidget *
gtk_canvas_new(int width, int height) {
    return g_object_new(GTK_TYPE_CANVAS, "width", width, "height", height, NULL);
}

GObject *
gtk_canvas_constructor(GType type,
		       guint n_construct_properties,
		       GObjectConstructParam * construct_params) {
    GObject * object = (*G_OBJECT_CLASS(parent_class)->constructor)(type,
								    n_construct_properties,
								    construct_params);

    return object;
}

void
gtk_canvas_set_property(GObject * object,
			guint prop_id,
			const GValue * value,
			GParamSpec * pspec) {
    GtkCanvas * canvas = GTK_CANVAS(object);

    switch(prop_id) {
    case PROP_WIDTH:
	gtk_canvas_set_width(canvas, g_value_get_int(value));
	break;
    case PROP_HEIGHT:
	gtk_canvas_set_height(canvas, g_value_get_int(value));
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    }
}

void
gtk_canvas_get_property(GObject * object,
			guint prop_id,
			GValue * value,
			GParamSpec * pspec) {
    GtkCanvas * canvas = GTK_CANVAS(object);

    switch(prop_id) {
    case PROP_WIDTH:
	g_value_set_int(value, canvas->width);
	break;
    case PROP_HEIGHT:
	g_value_set_int(value, canvas->height);
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
    }
}

void
gtk_canvas_destroy(GtkObject * object) {

    GtkCanvas * canvas = GTK_CANVAS(object);
    g_assert(object);

    g_message("gtk_canvas_destroy");

    return;			/* FIXME hack to get around multiple destroys */

    if(canvas->idle_id) {
	gtk_idle_remove(canvas->idle_id);
    }

    g_free(canvas->current_color_string);

    g_object_unref(G_OBJECT(canvas->gc));
    g_object_unref(G_OBJECT(canvas->layout));

    g_hash_table_foreach(canvas->color_table, &gtk_canvas_color_hash_free, (void *)canvas);
    g_hash_table_destroy(canvas->color_table);

    g_message("(destroy parent)");
    GTK_OBJECT_CLASS(parent_class)->destroy(object);
    g_message("end canvas destroy");
}

void
gtk_canvas_realize(GtkWidget * widget) {

    GtkCanvas * canvas = GTK_CANVAS(widget);

    GdkWindowAttr attributes;
    gint attributes_mask;

    g_assert(widget);

    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);
    attributes.event_mask = (gtk_widget_get_events(widget) |
			     GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK |
			     GDK_BUTTON_MOTION_MASK |
			     GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			     GDK_KEY_PRESS_MASK |  GDK_KEY_RELEASE_MASK |
			     GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
			     GDK_FOCUS_CHANGE_MASK);

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

    widget->window = gdk_window_new(gtk_widget_get_parent_window(widget), &attributes,
				    attributes_mask);

    GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

    gdk_window_set_user_data(widget->window, canvas);

    widget->style = gtk_style_attach(widget->style, widget->window);
    gdk_window_set_background(widget->window, &widget->style->white);

    gtk_canvas_send_configure(canvas);

    /* */

    canvas->gc = gdk_gc_new(widget->window);

    canvas->layout = gtk_widget_create_pango_layout(widget, NULL);
}

void
gtk_canvas_damage(GtkCanvas * canvas, const GdkRectangle * area) {
    g_return_if_fail(canvas);
    g_return_if_fail(area);

    /*
    g_message("gtk_canvas_damage");
    */

    if(!canvas->idle_id) {
	canvas->idle_id = gtk_idle_add(&gtk_canvas_idle_handler, canvas);
    }

    if(area->width == 0 && area->height == 0) {
	/* do nothing */
    }
    else if(canvas->damage.width == 0 && canvas->damage.height == 0) {
	canvas->damage.x = area->x;
	canvas->damage.y = area->y;
	canvas->damage.width = area->width;
	canvas->damage.height = area->height;
    }
    else {
	/* for some reason the input args to gdk_rectangle_union are not const,
	 * even though it never modifies the inputs. So I have to cast the
	 * constness away... */
	gdk_rectangle_union(&canvas->damage, (GdkRectangle *)area, &canvas->damage);
    }
}

void
gtk_canvas_set_color(GtkCanvas * canvas, const gchar * color_string) {
    gpointer data;
    GtkWidget * widget;

    g_return_if_fail(canvas);
    g_return_if_fail(color_string);

    if(strncmp(color_string, canvas->current_color_string, BUFFER_SIZE - 1)) {
	widget = GTK_WIDGET(canvas);


	data = g_hash_table_lookup(canvas->color_table, color_string);

	if(!data) {
	    GdkColor color;

	    if(gdk_color_parse(color_string, &color)) {
		GdkColormap * colormap = gtk_widget_get_colormap(widget);

		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_hash_table_insert(canvas->color_table,
				    g_strdup(color_string),
				    gdk_color_copy(&color));
		gdk_gc_set_foreground(canvas->gc, &color);
	    }
	    else {
		g_error("couldn't parse colour: %s", color_string);
	    }
	}
	else {
	    gdk_gc_set_foreground(canvas->gc, (GdkColor *)data);
	}

	strncpy(canvas->current_color_string, color_string, BUFFER_SIZE - 1);
    }
}


#if 1
void
gtk_canvas_set_line_width(GtkCanvas * canvas, int line_width) {

    GdkGCValues values;
    gint values_mask = GDK_GC_LINE_WIDTH;

    g_return_if_fail(canvas);
    g_return_if_fail(line_width > 0);

    values.line_width = line_width;

    gdk_gc_set_values(canvas->gc, &values, values_mask);
}
#endif

void
gtk_canvas_set_line_style(GtkCanvas * canvas, GdkLineStyle line_style) {
    g_return_if_fail(canvas);

    if(canvas->current_line_style != line_style) {
	GdkGCValues values;
	gint values_mask = GDK_GC_LINE_STYLE;

	values.line_style = line_style;

	gdk_gc_set_values(canvas->gc, &values, values_mask);

	canvas->current_line_style = line_style;
    }
}

void
gtk_canvas_draw_polygon(GtkCanvas * canvas, const GdkPoint * points, int num, int filled) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(points);
    g_return_if_fail(num > 0);

    gdk_draw_polygon(widget->window,
		     canvas->gc,
		     filled,
		     (GdkPoint *)points,
		     num);

    if(filled) {
	// make the filled version consistent with the outline version

	gdk_draw_polygon(widget->window,
			 canvas->gc,
			 FALSE,
			 (GdkPoint *)points,
			 num);
    }
}

void
gtk_canvas_draw_lines(GtkCanvas * canvas, const GdkPoint * points, int num) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(points);
    g_return_if_fail(num > 0);
    gdk_draw_lines(widget->window,
		   canvas->gc,
		   (GdkPoint *)points,
		   num);
}

void
gtk_canvas_draw_pixbuf(GtkCanvas * canvas, GdkPixbuf * pixbuf, int x, int y) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(pixbuf);
    // FIXME do w and h need to be adjusted to be consistent with the
    // rectangle size concept

    // FIXME by using (-1,-1) as the width/height, will it use the natural size ??

    gdk_pixbuf_render_to_drawable(pixbuf,
				  widget->window,
				  canvas->gc,
				  0,		// src_x
				  0,		// src_y
				  x,		// dest_x
				  y,		// dest_y
				  -1,//w,		// width
				  -1,//h,		// height
				  GDK_RGB_DITHER_NORMAL,
				  x,		// x_dither
				  y);		// y_dither
}

void
gtk_canvas_draw_layout(GtkCanvas * canvas, PangoLayout * layout, int x, int y) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(layout);
    gdk_draw_layout(widget->window,
		    canvas->gc,
		    x, y,
		    layout);
}

void
gtk_canvas_draw_point(GtkCanvas * canvas, int x, int y) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    gdk_draw_point(widget->window,
		   canvas->gc,
		   x, y);
}

void
gtk_canvas_draw_rectangle(GtkCanvas * canvas, const GdkRectangle * rect, int filled) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(rect);
    if(rect->width > 0 && rect->height > 0) {
	// make the filled version consistent with the outline version

	if(filled) {
	    gdk_draw_rectangle(widget->window,
			       canvas->gc,
			       TRUE,
			       rect->x, rect->y, rect->width, rect->height);
	}
	else {
	    gdk_draw_rectangle(widget->window,
			       canvas->gc,
			       FALSE,
			       rect->x, rect->y, rect->width - 1, rect->height - 1);
	}
    }
}

void
gtk_canvas_draw_string(GtkCanvas * canvas, const gchar * str, int length, int x, int y) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    g_return_if_fail(str);
    g_return_if_fail(length > 0);

    pango_layout_set_text(canvas->layout, str, length);

    gdk_draw_layout(widget->window,
		    canvas->gc,
		    x, y,
		    canvas->layout);
}

void
gtk_canvas_draw_line(GtkCanvas * canvas, int x0, int y0, int x1, int y1) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    gdk_draw_line(widget->window,
		  canvas->gc,
		  x0, y0, x1, y1);
}

void
gtk_canvas_draw_oval(GtkCanvas * canvas, int x, int y, int w, int h, int filled) {

    GtkWidget * widget = GTK_WIDGET(canvas);

    g_return_if_fail(canvas);
    if(w > 0 && h > 0) {
	// make the filled version consistent with the outline version

	if(filled) {
	    gdk_draw_arc(widget->window,
			 canvas->gc,
			 TRUE,
			 x, y,
			 w, h,
			 0, 360 << 6);
	}
	else {
	    gdk_draw_arc(widget->window,
			 canvas->gc,
			 FALSE,
			 x, y,
			 w - 1, h - 1,
			 0, 360 << 6);
	}
    }
}

void
gtk_canvas_set_clip(GtkCanvas * canvas, const GdkRectangle * rect) {
    g_return_if_fail(canvas);

    gdk_gc_set_clip_rectangle(canvas->gc, (GdkRectangle *)rect);
}

void
gtk_canvas_unset_clip(GtkCanvas * canvas) {
    g_return_if_fail(canvas);

    gdk_gc_set_clip_rectangle(canvas->gc, NULL);
}

void
gtk_canvas_set_width(GtkCanvas * canvas, int width) {
    g_return_if_fail(canvas);

    gtk_canvas_set_dimensions(canvas, width, canvas->height);
}

void
gtk_canvas_set_height(GtkCanvas * canvas, int height) {
    g_return_if_fail(canvas);

    gtk_canvas_set_dimensions(canvas, canvas->width, height);
}

void
gtk_canvas_set_dimensions(GtkCanvas * canvas, int width, int height) {
    g_return_if_fail(canvas);
    g_return_if_fail(width > 0);
    g_return_if_fail(height > 0);

    if(canvas->width != width || canvas->height != height) {
	canvas->width = width;
	canvas->height = height;

	gtk_widget_queue_resize(GTK_WIDGET(canvas));
    }
}

//
//
//

static gboolean
gtk_canvas_idle_handler(gpointer data) {

    GtkWidget * widget = GTK_WIDGET(data);
    GtkCanvas * canvas = GTK_CANVAS(widget);

    g_return_val_if_fail(data, FALSE);
    canvas->idle_id = 0;

    if(canvas->damage.width > 0 && canvas->damage.height > 0) {
        g_assert(widget->window);
        g_assert(GTK_WIDGET_REALIZED(widget));
	gdk_window_invalidate_rect(widget->window, &canvas->damage, FALSE);
    }

    // reset the damage rectangle

    canvas->damage.width = 0;
    canvas->damage.height = 0;

    return FALSE;	// prevent the idle handle from being called again
}

static guint
gtk_canvas_string_hash(gconstpointer key) {

    const GString gstr = {
	(gchar *)key,
	strlen((const gchar *)key)
    };

    return g_string_hash(&gstr);
}

static gboolean
gtk_canvas_string_cmp(gconstpointer a, gconstpointer b) {
    g_assert(a);
    g_assert(b);

    return !strcmp((const char *)a, (const char *)b);
}

static void
gtk_canvas_color_hash_free(gpointer key, gpointer value, gpointer user_data) {
    g_assert(key);
    g_assert(value);

    g_free(key);
    gdk_color_free((GdkColor *)value);
}

static void
gtk_canvas_send_configure(GtkCanvas * canvas) {

    GtkWidget * widget;
    GdkEventConfigure event;

    g_return_if_fail(canvas);
    widget = GTK_WIDGET(canvas);

    event.type = GDK_CONFIGURE;
    event.window = widget->window;
    event.send_event = TRUE;
    event.x = widget->allocation.x;
    event.y = widget->allocation.y;
    event.width = widget->allocation.width;
    event.height = widget->allocation.height;

    gtk_widget_event(widget, (GdkEvent *) &event);
}

static void
gtk_canvas_size_request(GtkWidget * widget, GtkRequisition * requisition) {

    GtkCanvas * canvas = GTK_CANVAS(widget);

    g_return_if_fail(widget);
    g_return_if_fail(requisition);
    requisition->width = canvas->width;
    requisition->height = canvas->height;
}

static void
gtk_canvas_size_allocate(GtkWidget * widget, GtkAllocation * allocation) {

    GtkCanvas * canvas = GTK_CANVAS(widget);

    g_return_if_fail(widget);
    g_return_if_fail(allocation);
    allocation->width = canvas->width;
    allocation->height = canvas->height;

    widget->allocation = *allocation;

    if(GTK_WIDGET_REALIZED(widget)) {
	gdk_window_move_resize(widget->window,
			       allocation->x, allocation->y,
			       allocation->width, allocation->height);

	gtk_canvas_send_configure(canvas);
    }
}

static gboolean
gtk_canvas_focus_in_event(GtkWidget * widget, GdkEventFocus * event) {
    g_return_val_if_fail(widget, FALSE);
    g_return_val_if_fail(event, FALSE);

    gtk_widget_queue_draw (widget);

    //g_message("Getting focus");

    return FALSE;
}

static gboolean
gtk_canvas_focus_out_event(GtkWidget * widget, GdkEventFocus * event) {
    g_return_val_if_fail(widget, FALSE);
    g_return_val_if_fail(event, FALSE);

    gtk_widget_queue_draw (widget);

    //g_message("Losing focus");

    return FALSE;
}

//
//
//

static void gtk_canvas_grab_focus(GtkWidget * widget);
static gboolean gtk_canvas_focus(GtkWidget * widget, GtkDirectionType direction);

static void
gtk_canvas_grab_focus(GtkWidget * widget) {
    g_return_if_fail(widget);

    //g_message("grab focus");

    GTK_WIDGET_CLASS(parent_class)->grab_focus(widget);
}

static gboolean
gtk_canvas_focus(GtkWidget * widget, GtkDirectionType direction) {
    g_return_val_if_fail(widget, FALSE);

    //g_message("focus");

    return GTK_WIDGET_CLASS(parent_class)->focus(widget, direction);
}
