// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "menu.h"

#include "foundation/macros.h"
#include "foundation/command.h"
#include "foundation/utility.h"

#include "toolkit/mouse_event.h"

#include <gtk/gtkmenubar.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkradiomenuitem.h>
#include <gtk/gtkseparatormenuitem.h>
#include <gtk/gtkmain.h>

namespace GTK {

    Menu::Menu(bool bar) : Toolkit::Menu(), Widget() {
	if(bar) {
	    mBaseWidget = gtk_menu_bar_new();
	}
	else {
	    mBaseWidget = gtk_menu_new();
	}

	post_init();
    }

    Menu::~Menu() {
    }

    void
    Menu::popup(const Toolkit::MouseEvent * mouse_event) {
	//Widget * _widget = known_cast<Widget *>(mouse_event->get_widget());

	g_signal_connect(G_OBJECT(mBaseWidget), "destroy",
			 G_CALLBACK(Menu::static_destroy_callback), this);

	gtk_menu_popup(GTK_MENU(mBaseWidget),
		       0,
		       0,
		       0,
		       0,
		       mouse_event->get_button(),
		       mouse_event->get_timestamp());

	//gtk_main();
    }

    void
    Menu::add_push(Command * command, const string & label, bool sensitive) {
	GtkWidget * item = gtk_menu_item_new_with_label(label.c_str());

	gtk_menu_shell_append(GTK_MENU_SHELL(mBaseWidget), item);

	if(command) {
	    command->ref();
	    g_object_set_data_full(G_OBJECT(item), "command", command, &Menu::static_unref_command);
	    g_signal_connect(G_OBJECT(item), "activate",
			     G_CALLBACK(&Menu::static_menu_callback),
			     this);
	}
	else {
	    WARNING("Push button has no commned");
	}

	if(!sensitive) {
	    gtk_widget_set_sensitive(item, false);
	}

	gtk_widget_show(item);
    }

    void
    Menu::add_check(Command * command, const string & label, bool sensitive) {
	GtkWidget * item = gtk_check_menu_item_new_with_label(label.c_str());

	gtk_menu_shell_append(GTK_MENU_SHELL(mBaseWidget), item);

	if(command) {
	    command->ref();
	    g_object_set_data_full(G_OBJECT(item), "command", command, &Menu::static_unref_command);
	    g_signal_connect(G_OBJECT(item), "activate",
			     G_CALLBACK(&Menu::static_menu_callback),
			     this);
	}
	else {
	    WARNING("Push button has no commned");
	}

	if(!sensitive) {
	    gtk_widget_set_sensitive(item, false);
	}

	gtk_widget_show(item);
    }

    void
    Menu::add_radio(Command * command, const string & label, Command * group, bool sensitive) {
	GSList * list;

	if(group) {
	    GtkWidget * other = find(group);
	    ASSERT(other);

	    list = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(other));
	}
	else {
	    list = 0;
	}

	GtkWidget * item = gtk_radio_menu_item_new_with_label(list, label.c_str());

	if(command) {
	    command->ref();
	    g_object_set_data_full(G_OBJECT(item), "command", command, &Menu::static_unref_command);
	    g_signal_connect(G_OBJECT(item), "activate",
			     G_CALLBACK(&Menu::static_menu_callback),
			     this);
	}
	else {
	    WARNING("Push button has no commned");
	}

	gtk_menu_shell_append(GTK_MENU_SHELL(mBaseWidget), item);

	if(!sensitive) {
	    gtk_widget_set_sensitive(item, false);
	}

	gtk_widget_show(item);
    }

    void
    Menu::add_submenu(Toolkit::Menu * submenu, const string & label) {
	GtkWidget * item = gtk_menu_item_new_with_label(label.c_str());

	gtk_menu_shell_append(GTK_MENU_SHELL(mBaseWidget), item);

	Menu * gtk_submenu = known_cast<Menu *>(submenu);

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), gtk_submenu->get_base_widget());

	gtk_widget_show(item);
    }

    void
    Menu::add_separator() {
	GtkWidget * item = gtk_separator_menu_item_new();

	gtk_menu_shell_append(GTK_MENU_SHELL(mBaseWidget), item);

	gtk_widget_show(item);
    }

    void
    Menu::set_sensitive(Command * command, bool sensitive) {
	GtkWidget * item = find(command);

	ASSERT(item);

	gtk_widget_set_sensitive(item, sensitive ? 1 : 0);
    }

    void
    Menu::set_state(Command * command, bool state) {
	GtkWidget * item = find(command);

	ASSERT(item);

	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), state ? 1 : 0);
    }

    GtkWidget *
    Menu::find(const Command * command) {
	ASSERT(command);

	GList * children = GTK_MENU_SHELL(mBaseWidget)->children;

	while(children) {
	    GtkMenuItem * item = static_cast<GtkMenuItem *>(children->data);

	    gpointer data = g_object_get_data(G_OBJECT(item), "command");

	    if(command == data) {
		return GTK_WIDGET(item);
	    }

	    children = children->next;
	}

	WARNING("Couldn't find item by command: " << command->get_instructions());

	return 0;
    }

    void
    Menu::static_destroy_callback(GtkWidget * widget, Menu * menu) {
	cerr << "Menu::static_destroy_callback" << endl;
	menu->destroy_callback(widget);
    }

    void
    Menu::destroy_callback(GtkWidget * widget) {
	cerr << "Menu::destroy_callback" << endl;
	gtk_main_quit();
    }

    void
    Menu::static_menu_callback(GtkWidget * widget, Menu * menu) {
	menu->menu_callback(widget);
    }

    void
    Menu::menu_callback(GtkWidget * widget) {
	GtkMenuItem * item = GTK_MENU_ITEM(widget);

	gpointer data = g_object_get_data(G_OBJECT(item), "command");

	ASSERT(data);

	Command * command = static_cast<Command *>(data);

	command->execute();
    }

    void
    Menu::static_unref_command(gpointer data) {
	Command * command = static_cast<Command *>(data);
	ASSERT(command);
	command->unref();
    }

} // namespace GTK
