// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h" 

#endif // HAVE_CONFIG_H

#include "tag.h"
#include "document.h"

#include "foundation/macros.h"

#include <gtk/gtktexttag.h>
#include <gtk/gtkenums.h>

namespace GTK {

  enum Attribute_Type {
    Attribute_Type_Int,
    Attribute_Type_Double,
    Attribute_Type_Bool,
    Attribute_Type_String,
    Attribute_Type_Token
  };

  struct Attribute_Table {
    const char *	name;
    Attribute_Type	type;
  };

  Attribute_Table attribute_table[] = {
    { "background",		Attribute_Type_String },
    { "foreground",		Attribute_Type_String },
    { "font",			Attribute_Type_String },
    { "weight",			Attribute_Type_Int },
    { "stretch",		Attribute_Type_Token },
    { "size",			Attribute_Type_Int },
    { "size-points",		Attribute_Type_Double },
    { "pixels-above-lines",	Attribute_Type_Int },
    { "pixels-below-lines",	Attribute_Type_Int },
    { "pixels-inside-wrap",	Attribute_Type_Int },
    { "editable",		Attribute_Type_Bool },
    { "wrap-mode",		Attribute_Type_Token },
    { "justification",		Attribute_Type_Token },
    { "left-margin",		Attribute_Type_Int },
    { "indent",			Attribute_Type_Int },
    { "strikethrough",		Attribute_Type_Bool },
    { "right-margin",		Attribute_Type_Int },
    { "underline",		Attribute_Type_Token },
    { "rise",			Attribute_Type_Int },
    { "background-full-height", Attribute_Type_Bool },
    { "invisible",		Attribute_Type_Bool },

    //"background-stipple"   GdkPixmap : Read / Write
    //"foreground-stipple"   GdkPixmap : Read / Write
    //"font-desc"            PangoFontDescription : Read / Write
    //"family"               gchararray : Read / Write
    //"style"                PangoStyle : Read / Write
    //"variant"              PangoVariant : Read / Write
    //"scale"                gdouble : Read / Write
    //"direction"            GtkTextDirection : Read / Write
    //"language"             gchararray : Read / Write
    //"tabs"                 PangoTabArray : Read / Write

    { 0, Attribute_Type_Int }
  };

  struct Attribute_Tokens {
    const char *	name;
    int			value;
  };

  Attribute_Tokens boolean_tokens[] = {
    { "0", 0 }, { "false", 0 }, { "no", 0 },
    { "1", 1 }, { "true", 1 }, { "yes", 1 },
    { 0, 0}
  };


  Attribute_Tokens attribute_tokens[] = {

    // enum PangoStretch
    { "ultra_condensed",	PANGO_STRETCH_ULTRA_CONDENSED },
    { "extra_condensed",	PANGO_STRETCH_EXTRA_CONDENSED },
    { "condensed",		PANGO_STRETCH_CONDENSED },
    { "semi_condensed",		PANGO_STRETCH_SEMI_CONDENSED },
    { "normal",			PANGO_STRETCH_NORMAL },
    { "semi_expanded",		PANGO_STRETCH_SEMI_EXPANDED },
    { "expanded",		PANGO_STRETCH_EXPANDED },
    { "extra_expanded",		PANGO_STRETCH_EXTRA_EXPANDED },
    { "ultra_expanded",		PANGO_STRETCH_ULTRA_EXPANDED },

    // enum GtkWrapMode
    { "wrap_none",		GTK_WRAP_NONE },
    { "wrap_char",		GTK_WRAP_CHAR },
    { "wrap_word",		GTK_WRAP_WORD },

    // enum GtkJustification
    { "left",			GTK_JUSTIFY_LEFT },
    { "right",			GTK_JUSTIFY_RIGHT },
    { "centre",			GTK_JUSTIFY_CENTER },
    { "center",			GTK_JUSTIFY_CENTER },
    { "fill",			GTK_JUSTIFY_FILL },

    // enum PangoUnderline
    { "underline_none",		PANGO_UNDERLINE_NONE },
    { "underline_single",	PANGO_UNDERLINE_SINGLE },
    { "underline_double",	PANGO_UNDERLINE_DOUBLE },
    { "underline_low",		PANGO_UNDERLINE_LOW },

    { 0, 0 }
  };


  static bool
  lookup_token(Attribute_Tokens * tp, const char * token, int & value) {
    while (tp->name) {
      //MESSAGE("Checking \"" << tp->name << "\" against \"" << token << "\"");
      if (strcasecmp(token, tp->name) == 0)
	break;
      tp++;
    }
    value = tp->value;
    if (tp->name)
      return true;
    else {
      WARNING("Unrecognised token \"" << token << "\" ignored");
      return false;
    }
  }


  static void destroy_notify(void * data) {
    // Unref the data when the owning object is destroyed
    Object * obj = (Object *) data;
    obj->unref();
  }


  Tag::Tag(Toolkit::Document * document, const char * name) : Toolkit::Tag(name) {
    //TRACE(2, "Create tag " << name);

    // Get the text buffer from the GTK::Document
    GTK::Document * gtkdoc = dynamic_cast<GTK::Document *>(document);
    GtkTextBuffer * textbuff = gtkdoc->get_text_buffer();

    // Create the tag, and add to the document tagtable
    mGtkTag = gtk_text_buffer_create_tag(textbuff, name, 0);
    assert(mGtkTag); // FIXUP

    // Set data in the Tag to allow relating back to this Tag
    this->ref();
    g_object_set_data_full(G_OBJECT(mGtkTag), "this", dynamic_cast<Toolkit::Tag *>(this), 
			   & destroy_notify);
  }



  Tag::~Tag() {
    if(mGtkTag) {
      g_object_unref(G_OBJECT(mGtkTag));
    }
  }


  // Toolkit::Tag interface methods
  bool
  Tag::set_attribute(const char * attr, const char * value) {
    //TRACE(2, "\t" << mName << " attribute " << attr << " = " << value);

    // Lookup the attr in the table
    Attribute_Table * atp = attribute_table;
    while (atp->name) {
      if (strcasecmp(attr, atp->name) == 0)
	break;
      atp++;
    }

    // Deal with the bad apples
    if (atp == 0) {
      WARNING("Bad attribute \"" << attr << "\" in Tag " << mName << " ignored");
      return false;
    }

    // Convert the value, and assign it to the tag
    int int_value = 0;
    bool conversion_result = true;
    switch (atp->type) {
    case Attribute_Type_Int:
      // Set the attribute as an iteger argument
      g_object_set(G_OBJECT(mGtkTag), attr, atoi(value), NULL);
      break;
    case Attribute_Type_Double:
      // Set the attribute as an iteger argument
      g_object_set(G_OBJECT(mGtkTag), attr, atof(value), NULL);
      break;
    case Attribute_Type_String:
      // Set the attribute as an string argument
      g_object_set(G_OBJECT(mGtkTag), attr, value, NULL);
      break;
    case Attribute_Type_Bool:
      // Set the attribute as an string argument
      conversion_result = lookup_token(boolean_tokens, value, int_value);
      g_object_set(G_OBJECT(mGtkTag), attr, int_value, NULL);
      break;	
    case Attribute_Type_Token:
      // Set the attribute as an string argument
      conversion_result = lookup_token(attribute_tokens, value, int_value);
      g_object_set(G_OBJECT(mGtkTag), attr, int_value, NULL);
      break;
    }
    return conversion_result;
  }


} // namespace GTK
