// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "toolbar.h"

#include "foundation/macros.h"
#include "foundation/utility.h"
#include "foundation/command.h"
#include "foundation/algorithm.h"

#include "image.h"

#include <gtk/gtktoolbar.h>
#include <gtk/gtkimage.h>
#include <gtk/gtktogglebutton.h>

namespace GTK {

    Toolbar::Toolbar() : Toolkit::Toolbar(), Widget() {
	mBaseWidget = gtk_toolbar_new();


	/*
	GTK_TOOLBAR_ICONS,
	GTK_TOOLBAR_TEXT,
	GTK_TOOLBAR_BOTH,
	GTK_TOOLBAR_BOTH_HORIZ
	*/

	gtk_toolbar_set_style(GTK_TOOLBAR(mBaseWidget), GTK_TOOLBAR_BOTH);
	gtk_toolbar_set_orientation(GTK_TOOLBAR(mBaseWidget), GTK_ORIENTATION_VERTICAL);

	post_init();
    }

    Toolbar::~Toolbar() {
    }

    void
    Toolbar::add_element(Command * command, Drawing::Image * image, bool sensitive) {
	RETURN_IF_FAIL(command);
	RETURN_IF_FAIL(image);

	GtkWidget * element;

	const string & instructions = command->get_instructions();
	Image * gtk_image = known_cast<GTK::Image *>(image);

	GdkPixbuf * gdk_pixbuf = gtk_image->get_pixbuf();
	ASSERT(gdk_pixbuf);

	GtkWidget * icon = gtk_image_new_from_pixbuf(gdk_pixbuf);
	ASSERT(icon);

	command->ref();
	image->ref();

	element = gtk_toolbar_append_element(GTK_TOOLBAR(mBaseWidget),
					     GTK_TOOLBAR_CHILD_BUTTON,
					     0,
					     0,
					     instructions.c_str(),
					     0,
					     icon,
					     GTK_SIGNAL_FUNC(&Toolbar::static_toolbar_callback),
					     this);

	ASSERT(element);

	g_object_set_data_full(G_OBJECT(element), "command", static_cast<Object *>(command),
			       &Toolbar::static_unref);
	g_object_set_data_full(G_OBJECT(element), "image", static_cast<Object *>(image),
			       &Toolbar::static_unref);

	if(!sensitive) {
	    gtk_widget_set_sensitive(element, FALSE);
	}

	gtk_widget_show(icon);
    }

    void
    Toolbar::add_separator() {
	gtk_toolbar_append_space(GTK_TOOLBAR(mBaseWidget));
    }

    void
    Toolbar::set_sensitive(Command * command, bool sensitive) {
	GtkWidget * element = find(command);

	RETURN_IF_FAIL(element);

	gtk_widget_set_sensitive(element, sensitive ? TRUE : FALSE);
    }

    GtkWidget *
    Toolbar::find(const Command * command) {
	GList * children = GTK_TOOLBAR(mBaseWidget)->children;

	while(children) {
	    GtkToolbarChild * child = static_cast<GtkToolbarChild *>(children->data);

	    if(child->type == GTK_TOOLBAR_CHILD_BUTTON) {
		GtkButton * button = GTK_BUTTON(child->widget);

		gpointer data = g_object_get_data(G_OBJECT(button), "command");

		ASSERT(data);

		if(command == data) {
		    return GTK_WIDGET(button);
		}
	    }

	    children = children->next;
	}

	ERROR("Couldn't find item by command: " << command->get_instructions());

	return 0;
    }

    void
    Toolbar::static_toolbar_callback(GtkWidget * widget, Toolbar * toolbar) {
	toolbar->toolbar_callback(widget);
    }

    void
    Toolbar::toolbar_callback(GtkWidget * widget) {
	GtkButton * button = GTK_BUTTON(widget);

	if(button->in_button) {
	    gpointer data = g_object_get_data(G_OBJECT(button), "command");

	    ASSERT(data);

	    Command * command = static_cast<Command *>(data);

	    command->execute();
	}
    }

    void
    Toolbar::static_unref(gpointer data) {
	ASSERT(data);
	Object * object = static_cast<Object *>(data);
	object->unref();
    }

} // namespace GTK
