// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H

#include "tree.h"

#include "image.h"

#include "foundation/macros.h"
#include "foundation/utility.h"

#include "drawing/point.h"

#include "toolkit/tree_maintainer.h"
#include "toolkit/mouse_event.h"

#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkcellrendererpixbuf.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtkenums.h>
#include <gtk/gtktreednd.h>

namespace {

    enum {
	TARGET_GTK_TREE_MODEL_ROW
    };

    GtkTargetEntry row_targets[] = {
	{ "GTK_TREE_MODEL_ROW", GTK_TARGET_SAME_APP, TARGET_GTK_TREE_MODEL_ROW }
    };

} // namespace {anonymous}


namespace GTK {

    Tree::Tree() : Toolkit::Tree(), Widget() {
	// the tree has 4 data elements,
	//   the pixbuf,
	//   the label,
	//   the object that the node represents
	//   is the text editable

	mTreeStore = gtk_tree_store_new(NUM_COLUMNS,
					GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_POINTER,
					G_TYPE_BOOLEAN);

	// create the base widget (the scrolled window)

	mBaseWidget = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(mBaseWidget),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	// create the tree view

	mTreeView = gtk_tree_view_new_with_model(GTK_TREE_MODEL(mTreeStore));

	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(mTreeView), TRUE);
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(mTreeView), FALSE);

	// add the tree to the scrolled window

	gtk_container_add(GTK_CONTAINER(mBaseWidget), mTreeView);

	// there is one column in the tree

	GtkTreeViewColumn * column = gtk_tree_view_column_new();

	gtk_tree_view_column_set_title(column, "Node");

	// create the pixbuf renderer

	GtkCellRenderer * pixbuf_renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(column, pixbuf_renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, pixbuf_renderer,
					    "pixbuf", PIXBUF_COLUMN, 0);

	// create the text renderer

	GtkCellRenderer * text_renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, text_renderer, TRUE);
	gtk_tree_view_column_set_attributes(column, text_renderer,
					    "text", TEXT_COLUMN,
					    "editable", EDITABLE_COLUMN, 0);

	// we want to get notifed when the node gets edited

	g_signal_connect(G_OBJECT(text_renderer), "edited",
			 G_CALLBACK(&Tree::static_edited), this);

	// add the column to the tree

	gtk_tree_view_append_column(GTK_TREE_VIEW(mTreeView), column);

	// sort the column by text

	gtk_tree_view_column_set_sort_column_id(column, TEXT_COLUMN);

	// setup the selection stuff

	GtkTreeSelection * selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(mTreeView));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

	g_signal_connect(G_OBJECT(selection), "changed",
			 G_CALLBACK(&Tree::static_selection_changed), this);

	// a hash table that maps from objects to iterators

	mHashTable = g_hash_table_new_full(g_direct_hash, g_direct_equal,
					   &Tree::hash_key_destroy,
					   &Tree::hash_value_destroy);

	//

	g_signal_connect(G_OBJECT(mTreeView), "button_press_event",
			 G_CALLBACK(&Tree::static_button_press_event),
			 this);

	//

#if 0
	gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(mTreeView),
					       GDK_BUTTON1_MASK,
					       row_targets,
					       G_N_ELEMENTS(row_targets),
					       GDK_ACTION_MOVE);

	gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(mTreeView),
					     row_targets,
					     G_N_ELEMENTS(row_targets),
					     GDK_ACTION_MOVE);
#endif


#if 0
	GtkTreeDragSourceIface * source_interface = GTK_TREE_DRAG_SOURCE_GET_IFACE(mTreeStore);

	g_signal_connect(G_OBJECT(mTreeStore), "row_draggable",
			 G_CALLBACK(&Tree::static_is_draggable), this);

	g_signal_connect(G_OBJECT(mTreeView), "row_draggable",
			 G_CALLBACK(&Tree::static_is_draggable), this);

	//


	cerr << source_interface << endl;
	cerr << source_interface->row_draggable << endl;
	cerr << source_interface->drag_data_get << endl;
	cerr << source_interface->drag_data_delete << endl;


	g_signal_connect(G_OBJECT(mTreeView), "drag_leave",
			 G_CALLBACK(&Tree::static_drag_leave), this);

	g_signal_connect(G_OBJECT(mTreeView), "drag_motion",
			 G_CALLBACK(&Tree::static_drag_motion), this);

	g_signal_connect(G_OBJECT(mTreeView), "drag_drop",
			 G_CALLBACK(&Tree::static_drag_leave), this);

	g_signal_connect(G_OBJECT(mTreeView), "drag_data_received",
			 G_CALLBACK(&Tree::static_drag_leave), this);
#endif

	gtk_widget_show(mTreeView);

	post_init();
    }

    Tree::~Tree() {
	g_hash_table_destroy(mHashTable);

	g_object_unref(G_OBJECT(mTreeStore));
    }

    //

    void
    Tree::set_image(Object * node, Drawing::Image * image) {
	RETURN_IF_FAIL(node);
	RETURN_IF_FAIL(image);

	Image * image2 = known_cast<GTK::Image *>(image);
	image2->scale(GTK_ICON_SIZE_SMALL_TOOLBAR);
	GdkPixbuf * pixbuf = image2->get_pixbuf();

	GtkTreeIter * iter = object_to_iter(node);
	RETURN_IF_FAIL(iter);

	gtk_tree_store_set(mTreeStore, iter, PIXBUF_COLUMN, pixbuf, -1);
    }

    void
    Tree::set_label(Object * node, const string & label) {
	RETURN_IF_FAIL(node);

	const gchar * text = label.c_str();

	GtkTreeIter * iter = object_to_iter(node);
	RETURN_IF_FAIL(iter);

	gtk_tree_store_set(mTreeStore, iter, TEXT_COLUMN, text, -1);
    }

    void
    Tree::add(Object * node, Drawing::Image * image, const string & label, bool expanded,
	      Object * parent, Object * position) {
	RETURN_IF_FAIL(node);
	RETURN_IF_FAIL(image);

	const gchar * text = label.c_str();
	Image * image2 = known_cast<GTK::Image *>(image);
	image2->scale(GTK_ICON_SIZE_SMALL_TOOLBAR);
	GdkPixbuf * pixbuf = image2->get_pixbuf();

	GtkTreeIter * parent_iter = parent ? object_to_iter(parent) : 0;

	GtkTreeIter iter;

	gtk_tree_store_append(mTreeStore, &iter, parent_iter);

	gtk_tree_store_set(mTreeStore, &iter,
			   PIXBUF_COLUMN, pixbuf,
			   TEXT_COLUMN, text,
			   NODE_COLUMN, node,
			   EDITABLE_COLUMN, TRUE, -1);

	hash_insert(node, gtk_tree_iter_copy(&iter));

	if(expanded && parent_iter) {
	    GtkTreePath * path = gtk_tree_model_get_path(GTK_TREE_MODEL(mTreeStore), parent_iter);
	    gtk_tree_view_expand_row(GTK_TREE_VIEW(mTreeView), path, FALSE);
	    gtk_tree_path_free(path);
	}

	image->unref();
    }

    void
    Tree::remove(Object * node) {
	RETURN_IF_FAIL(node);

	GtkTreeIter * iter = object_to_iter(node);
	RETURN_IF_FAIL(iter);

	// shouldn't delete branches, only leaves:
	ASSERT(gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mTreeStore), iter) == 0);

	gtk_tree_store_remove(mTreeStore, iter);

	hash_remove(node);
    }

    void
    Tree::clear() {
	// clear the tree

	gtk_tree_store_clear(mTreeStore);

	// clear the hash table (by destroying it and creating a new one)

	g_hash_table_destroy(mHashTable);
	mHashTable = g_hash_table_new_full(g_direct_hash, g_direct_equal,
					   &Tree::hash_key_destroy,
					   &Tree::hash_value_destroy);
    }

    Object *
    Tree::get_root() {
	GtkTreeIter iter;

	if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(mTreeStore), &iter)) {
	    return iter_to_object(&iter);
	}
	else {
	    return 0;		// the tree must be empty
	}
    }

    Object *
    Tree::get_parent(Object * node) {
	RETURN_VAL_IF_FAIL(node, 0);

	GtkTreeIter * iter = object_to_iter(node);
	GtkTreeIter parent;

	if(gtk_tree_model_iter_parent(GTK_TREE_MODEL(mTreeStore), iter, &parent)) {
	    return iter_to_object(&parent);
	}
	else {
	    return 0;		// the object must be the root... (it has no parent)
	}
    }

    bool
    Tree::contains(Object * node) const {
	return bool(g_hash_table_lookup(mHashTable, static_cast<gconstpointer>(node)));
    }

    Object *
    Tree::point_to_node(const Drawing::Point & point) {
	GtkTreePath * path = 0;

	if(gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(mTreeView),
					 point.get_x(), point.get_y(),
					 &path,
					 0,
					 0, 0)) {

	    ASSERT(path);

	    GtkTreeIter iter;

	    ASSERT(gtk_tree_model_get_iter(GTK_TREE_MODEL(mTreeStore),
					   &iter, path));

	    gtk_tree_path_free(path);

	    return iter_to_object(&iter);
	}
	else {
	    ASSERT(!path);

	    return 0;
	}
    }

    //

    void
    Tree::selection_changed(GtkTreeSelection * selection) {
	//cerr << "Selection changed" << endl;

#if 0
	mSelection.clear();

	GtkTreeIter iter;

	if(gtk_tree_selection_get_selected(selection, 0, &iter)) {
	    Object * object = iter_to_object(&iter);
	    mSelection.insert(object);
	    //cerr << object << endl;
	}

	TreeViewEvent tree_view_event(TVET_SELECTION_CHANGED, this, mSelection);
	TreeView::selection_changed(&tree_view_event);
#endif
    }

    //

    bool
    Tree::motion_notify_event(const GdkEventMotion * event) {
	return false;
    }

    bool
    Tree::enter_notify_event(const GdkEventCrossing * event) {
	return false;
    }

    bool
    Tree::leave_notify_event(const GdkEventCrossing * event) {
	return false;
    }

    bool
    Tree::button_press_event(const GdkEventButton * event) {
	//MESSAGE("Button pressed in tree");

	TRACE(10, "");

	Toolkit::MouseEvent mouse_event(Toolkit::MET_PRESSED,
					Drawing::Point(int(event->x + 0.5), int(event->y + 0.5)),
					this,
					translate_modifiers(event->state),
					event->time,
					event->button,
					translate_presses(event->type));

	//MESSAGE("Hit: " << mouse_event.get_point() << ", "
	//<< point_to_node(mouse_event.get_point()));

	//g_signal_stop_emission_by_name(G_OBJECT(mTreeView), "button_press_event");

	return mouse_pressed(&mouse_event);
    }

    bool
    Tree::button_release_event(const GdkEventButton * event) {
	return false;
    }

    //

    void
    Tree::hash_key_destroy(gpointer key) {
	g_assert(key);
	Object * object = static_cast<Object *>(key);
	object->unref();
    }

    void
    Tree::hash_value_destroy(gpointer value) {
	g_assert(value);
	GtkTreeIter * iter = static_cast<GtkTreeIter *>(value);
	gtk_tree_iter_free(iter);
    }

    gboolean
    Tree::is_draggable(GdkDragContext * drag_context, GtkTreePath * path) {
	MESSAGE("Is draggable");

	return TRUE;
    }

    gboolean
    Tree::is_droppable(GdkDragContext * drag_context, GtkTreePath * path,
		       GtkTreeViewDropPosition * position) {
	MESSAGE("Is droppable");

	return TRUE;
    }

    void
    Tree::edited(GtkCellRendererText * text_renderer, const char * path_string, const char * text) {
	Toolkit::TreeMaintainer * tree_maintainer = get_maintainer();

	if(tree_maintainer) {
	    GtkTreeIter iter;

	    ASSERT(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(mTreeStore),
						       &iter, path_string));

	    Object * node = iter_to_object(&iter);

	    ASSERT(node);

	    tree_maintainer->renamed(this, node, text);
	}
    }

    //

    void
    Tree::static_selection_changed(GtkTreeSelection * selection, Tree * tree) {
	tree->selection_changed(selection);
    }

    gboolean
    Tree::static_is_draggable(GtkTreeView * tree_view,
			      GdkDragContext * drag_context,
			      GtkTreePath * path,
			      gpointer user_data) {
	Tree * tree = static_cast<Tree *>(user_data);
	return tree->is_draggable(drag_context, path);
    }

    gboolean
    Tree::static_is_droppable(GtkTreeView * tree_view,
			      GdkDragContext * drag_context,
			      GtkTreePath * path,
			      GtkTreeViewDropPosition * position,
			      gpointer user_data) {
	Tree * tree = static_cast<Tree *>(user_data);
	return tree->is_droppable(drag_context, path, position);
    }

    void
    Tree::static_edited(GtkCellRendererText * text_renderer,
			const char * path, const char * text, Tree * tree) {
	tree->edited(text_renderer, path, text);
    }

    gint
    Tree::static_motion_notify_event(GtkWidget * widget, GdkEventMotion * event, Tree * window) {
	return window->motion_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Tree::static_enter_notify_event(GtkWidget * widget, GdkEventCrossing * event, Tree * window) {
	return window->enter_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Tree::static_leave_notify_event(GtkWidget * widget, GdkEventCrossing * event, Tree * window) {
	return window->leave_notify_event(event) ? TRUE : FALSE;
    }

    gint
    Tree::static_button_press_event(GtkWidget * widget, GdkEventButton * event, Tree * window) {
	return window->button_press_event(event) ? TRUE : FALSE;
    }

    gint
    Tree::static_button_release_event(GtkWidget * widget, GdkEventButton * event, Tree * window) {
	return window->button_release_event(event) ? TRUE : FALSE;
    }

    //

    GtkTreeIter *
    Tree::object_to_iter(Object * object) {
	gpointer value = g_hash_table_lookup(mHashTable, static_cast<gconstpointer>(object));

	GtkTreeIter * iter = static_cast<GtkTreeIter *>(value);

	ASSERT(iter);

	return iter;
    }

    Object *
    Tree::iter_to_object(GtkTreeIter * iter) {
	Object * object = 0;

	gtk_tree_model_get(GTK_TREE_MODEL(mTreeStore), iter, 2, &object, -1);

	return object;
    }

    void
    Tree::hash_insert(Object * object, GtkTreeIter * iter) {
	g_assert(object);
	g_assert(iter);

	object->ref();
	g_hash_table_insert(mHashTable, static_cast<gpointer>(object), static_cast<gpointer>(iter));
    }

    void
    Tree::hash_remove(Object * object) {
	g_assert(object);

	g_hash_table_remove(mHashTable, static_cast<gconstpointer>(object));
    }

} // namespace GTK
