// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "generalizable_element.h"

#include "model.h"

#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <algorithm>
#include <vector>
#include <iterator>
using namespace std;

namespace MOF {

    const string GeneralizableElement::TYPE = "GeneralizableElement";

    //

    GeneralizableElement::GeneralizableElement() : Namespace(),
        mVisibility(0),
	mAbstract(0),
	mRoot(0),
	mLeaf(0) {
    }

    GeneralizableElement::~GeneralizableElement() {
    }

    void
    GeneralizableElement::add_supertype(GeneralizableElement * supertype, int index) {
	RETURN_IF_FAIL(index <= static_cast<int>(mSupertypes.size()));

	if(index < 0) {
	    index = mSupertypes.size();
	}

	list<GeneralizableElement *>::iterator iter = mSupertypes.begin();

	for(int i = 0; i < static_cast<int>(mSupertypes.size()); ++i) {
	    RETURN_IF_FAIL(*iter != supertype);

	    if(i < index) {
		++iter;
	    }
	}

	begin_transaction();

	Command * command = command_flip_method(this,
						&GeneralizableElement::add_supertype,
						&GeneralizableElement::remove_supertype,
						supertype, index);
	log_command(command);
	command->unref();

	// make the changes

	supertype->ref();

	mSupertypes.insert(iter, supertype);
	supertype->internal_add_subtype(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "supertype", supertype, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    GeneralizableElement::remove_supertype(GeneralizableElement * supertype) {
	int index = 0;

	list<GeneralizableElement *>::iterator iter = mSupertypes.begin();

	for(; iter != mSupertypes.end(); ++iter, ++index) {
	    if(*iter == supertype) {
		break;
	    }
	}

	RETURN_IF_FAIL(iter != mSupertypes.end());

	begin_transaction();

	Command * command = command_flip_method(this,
						&GeneralizableElement::remove_supertype,
						&GeneralizableElement::add_supertype,
						supertype, index);
	log_command(command);
	command->unref();

	// make the changes

	supertype->internal_remove_subtype(this);
	mSupertypes.erase(iter);

	supertype->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "supertype", supertype, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void modify_supertype(GeneralizableElement * old_supertype,
			  GeneralizableElement * new_supertype) {
	NYI();
    }

    void
    GeneralizableElement::set_supertypes(const list<GeneralizableElement *> & supertypes) {
	set<GeneralizableElement *> before;
	set<GeneralizableElement *> after;

	before.insert(mSupertypes.begin(), mSupertypes.end());
	after.insert(supertypes.begin(), supertypes.end());

	// check for duplicates
	RETURN_IF_FAIL(before.size() == mSupertypes.size());
	RETURN_IF_FAIL(after.size() == supertypes.size());

	vector<GeneralizableElement *> added;
	vector<GeneralizableElement *> removed;

	set_difference(after.begin(), after.end(),
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	// transaction time

	begin_transaction();

	Command * command = command_flip_argument(this,
						  &GeneralizableElement::set_supertypes,
						  supertypes, mSupertypes);
	log_command(command);
	command->unref();

	// make the changes

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&GeneralizableElement::internal_add_subtype), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&GeneralizableElement::internal_remove_subtype), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mSupertypes = supertypes;

	// changes made

	for(vector<GeneralizableElement *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    GeneralizableElement * supertype = *iter;
	    ChangeEvent * change = change_role_added(this, "supertype", supertype, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<GeneralizableElement *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    GeneralizableElement * supertype = *iter;
	    ChangeEvent * change = change_role_removed(this, "supertype", supertype, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    GeneralizableElement::set_visibility(const VisibilityKind * visibility) {
	begin_transaction();

	//

	VisibilityKind * old_value = mVisibility ? new VisibilityKind(*mVisibility) : 0;
	VisibilityKind * new_value = visibility ? new VisibilityKind(*visibility) : 0;

	//

	Command * command = command_flip_argument(this, &GeneralizableElement::set_visibility,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mVisibility) {
	    mVisibility->unref();
	}

	mVisibility = new_value ? new VisibilityKind(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "visible", new_value, old_value);
	log_command(command);
	log_change(change);

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    GeneralizableElement::set_abstract(const Boolean * abstract) {
	begin_transaction();

	//

	Boolean * old_value = mAbstract ? new Boolean(*mAbstract) : 0;
	Boolean * new_value = abstract ? new Boolean(*abstract) : 0;

	//

	Command * command = command_flip_argument(this, &GeneralizableElement::set_abstract,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mAbstract) {
	    mAbstract->unref();
	}

	mAbstract = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "abstract", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    GeneralizableElement::set_root(const Boolean * root) {
	begin_transaction();

	//

	Boolean * old_value = mRoot ? new Boolean(*mRoot) : 0;
	Boolean * new_value = root ? new Boolean(*root) : 0;

	//

	Command * command = command_flip_argument(this, &GeneralizableElement::set_root,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mRoot) {
	    mRoot->unref();
	}

	mRoot = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "root", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    void
    GeneralizableElement::set_leaf(const Boolean * leaf) {
	begin_transaction();

	//

	Boolean * old_value = mLeaf ? new Boolean(*mLeaf) : 0;
	Boolean * new_value = leaf ? new Boolean(*leaf) : 0;

	//

	Command * command = command_flip_argument(this, &GeneralizableElement::set_leaf,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mLeaf) {
	    mLeaf->unref();
	}

	mLeaf = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "leaf", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //
    //
    //

    void
    GeneralizableElement::internal_add_subtype(GeneralizableElement * subtype) throw() {
	subtype->ref();
	mSubtypes.insert(subtype);
    }

    void
    GeneralizableElement::internal_remove_subtype(GeneralizableElement * subtype) throw() {
	mSubtypes.erase(subtype);
	subtype->unref();
    }

#if 0
    list<GeneralizableElement *>
    GeneralizableElement::get_all_supertypes() {
	list<GeneralizableElement *> super_types;

	// TODO

	return super_types;
    }
#endif

    void
    GeneralizableElement::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    list<GeneralizableElement *> empty_generalizable_element_list;
	    set<GeneralizableElement *> subtypes_copy = mSubtypes;

	    set_supertypes(empty_generalizable_element_list);
	    for_each(subtypes_copy.begin(), subtypes_copy.end(),
		     bind2nd(mem_fun(&GeneralizableElement::remove_supertype), this));

	    Namespace::ref_delete_impl(visited);
	}
    }

    bool
    GeneralizableElement::ref_is_a_type_of_impl(set<string> & visited,
						    const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || Namespace::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    GeneralizableElement::ref_get_type() const {
	return TYPE;
    }

    void
    GeneralizableElement::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_generalizable_element(this);

	    Namespace::ref_implant_impl(visited);
	}
    }

    void
    GeneralizableElement::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_generalizable_element(this);

	    Namespace::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
