// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "model_handler.h"

#include "foundation/iterator.h"
#include "foundation/utility.h"
#include "foundation/macros.h"

#include "xml/sax_parser.h"

#include "common/dumb_handler.h"
#include "common/type_code_handler.h"

#include "model.h"

#include "association_end.h"
#include "association.h"
#include "attribute.h"
#include "behavioral_feature.h"
#include "class.h"
#include "classifier.h"
#include "constant.h"
#include "constraint.h"
#include "data_type.h"
#include "exception.h"
#include "feature.h"
#include "generalizable_element.h"
#include "import.h"
#include "model_element.h"
#include "namespace.h"
#include "operation.h"
#include "package.h"
#include "parameter.h"
#include "reference.h"
#include "structural_feature.h"
#include "tag.h"
#include "typed_element.h"

#include "alias_type.h"
#include "primitive_type.h"
#include "collection_type.h"
#include "structure_type.h"
#include "enumeration_type.h"
#include "structure_field.h"

#include <iterator>
#include <functional>
#include <algorithm>
#include <sstream>
using namespace std;

namespace MOF {

    ModelHandler::ModelHandler() : XML::SAXHandler() {
	mHandlerMap["MOF:Namespace.contents"] =
	    make_pair(&ModelHandler::mof__namespace__contents__start,
		      &ModelHandler::mof__namespace__contents__end);

	mHandlerMap["MOF:Feature.visibility"] =
	    make_pair(&ModelHandler::mof__feature__visibility__start,
		      &ModelHandler::mof__feature__visibility__end);

	mHandlerMap["MOF:Feature.scope"] =
	    make_pair(&ModelHandler::mof__feature__scope__start,
		      &ModelHandler::mof__feature__scope__end);

	mHandlerMap["MOF:Feature"] =
	    make_pair(&ModelHandler::mof__feature__start,
		      &ModelHandler::mof__feature__end);

	mHandlerMap["MOF:Classifier"] =
	    make_pair(&ModelHandler::mof__classifier__start,
		      &ModelHandler::mof__classifier__end);

	mHandlerMap["MOF:Namespace"] =
	    make_pair(&ModelHandler::mof__namespace__start,
		      &ModelHandler::mof__namespace__end);

	mHandlerMap["MOF:StructuralFeature.multiplicity"] =
	    make_pair(&ModelHandler::mof__structural_feature__multiplicity__start,
		      &ModelHandler::mof__structural_feature__multiplicity__end);

	mHandlerMap["MOF:StructuralFeature.isChangeable"] =
	    make_pair(&ModelHandler::mof__structural_feature__is_changeable__start,
		      &ModelHandler::mof__structural_feature__is_changeable__end);

	mHandlerMap["MOF:StructuralFeature"] =
	    make_pair(&ModelHandler::mof__structural_feature__start,
		      &ModelHandler::mof__structural_feature__end);

	mHandlerMap["MOF:Attribute.isDerived"] =
	    make_pair(&ModelHandler::mof__attribute__is_derived__start,
		      &ModelHandler::mof__attribute__is_derived__end);

	mHandlerMap["MOF:Attribute"] =
	    make_pair(&ModelHandler::mof__attribute__start,
		      &ModelHandler::mof__attribute__end);

	mHandlerMap["MOF:Operation.isQuery"] =
	    make_pair(&ModelHandler::mof__operation__is_query__start,
		      &ModelHandler::mof__operation__is_query__end);

	mHandlerMap["MOF:Operation.exceptions"] =
	    make_pair(&ModelHandler::mof__operation__exceptions__start,
		      &ModelHandler::mof__operation__exceptions__end);

	mHandlerMap["MOF:Operation"] =
	    make_pair(&ModelHandler::mof__operation__start,
		      &ModelHandler::mof__operation__end);

	mHandlerMap["MOF:Class.isSingleton"] =
	    make_pair(&ModelHandler::mof__class__is_singleton__start,
		      &ModelHandler::mof__class__is_singleton__end);

	mHandlerMap["MOF:Class"] =
	    make_pair(&ModelHandler::mof__class__start,
		      &ModelHandler::mof__class__end);

	mHandlerMap["MOF:ModelElement.name"] =
	    make_pair(&ModelHandler::mof__model_element__name__start,
		      &ModelHandler::mof__model_element__name__end);

	mHandlerMap["MOF:ModelElement.annotation"] =
	    make_pair(&ModelHandler::mof__model_element__annotation__start,
		      &ModelHandler::mof__model_element__annotation__end);

	mHandlerMap["MOF:ModelElement.container"] =
	    make_pair(&ModelHandler::mof__model_element__container__start,
		      &ModelHandler::mof__model_element__container__end);

	mHandlerMap["MOF:ModelElement.constraints"] =
	    make_pair(&ModelHandler::mof__model_element__constraints__start,
		      &ModelHandler::mof__model_element__constraints__end);

	mHandlerMap["MOF:ModelElement"] =
	    make_pair(&ModelHandler::mof__model_element__start,
		      &ModelHandler::mof__model_element__end);

	mHandlerMap["MOF:Package"] =
	    make_pair(&ModelHandler::mof__package__start,
		      &ModelHandler::mof__package__end);

	mHandlerMap["MOF:Constraint.expression"] =
	    make_pair(&ModelHandler::mof__constraint__expression__start,
		      &ModelHandler::mof__constraint__expression__end);

	mHandlerMap["MOF:Constraint.language"] =
	    make_pair(&ModelHandler::mof__constraint__language__start,
		      &ModelHandler::mof__constraint__language__end);

	mHandlerMap["MOF:Constraint.evaluationPolicy"] =
	    make_pair(&ModelHandler::mof__constraint__evaluation_policy__start,
		      &ModelHandler::mof__constraint__evaluation_policy__end);

	mHandlerMap["MOF:Constraint.constrainedElements"] =
	    make_pair(&ModelHandler::mof__constraint__constrained_elements__start,
		      &ModelHandler::mof__constraint__constrained_elements__end);

	mHandlerMap["MOF:Constraint"] =
	    make_pair(&ModelHandler::mof__constraint__start,
		      &ModelHandler::mof__constraint__end);

	mHandlerMap["MOF:Association.isDerived"] =
	    make_pair(&ModelHandler::mof__association__is_derived__start,
		      &ModelHandler::mof__association__is_derived__end);

	mHandlerMap["MOF:Association"] =
	    make_pair(&ModelHandler::mof__association__start,
		      &ModelHandler::mof__association__end);

	mHandlerMap["MOF:Parameter.direction"] =
	    make_pair(&ModelHandler::mof__parameter__direction__start,
		      &ModelHandler::mof__parameter__direction__end);

	mHandlerMap["MOF:Parameter.multiplicity"] =
	    make_pair(&ModelHandler::mof__parameter__multiplicity__start,
		      &ModelHandler::mof__parameter__multiplicity__end);

	mHandlerMap["MOF:Parameter"] =
	    make_pair(&ModelHandler::mof__parameter__start,
		      &ModelHandler::mof__parameter__end);

	mHandlerMap["MOF:BehavioralFeature"] =
	    make_pair(&ModelHandler::mof__behavioral_feature__start,
		      &ModelHandler::mof__behavioral_feature__end);

	mHandlerMap["MOF:Import.visibility"] =
	    make_pair(&ModelHandler::mof__import__visibility__start,
		      &ModelHandler::mof__import__visibility__end);

	mHandlerMap["MOF:Import.importedNamespace"] =
	    make_pair(&ModelHandler::mof__import__imported_namespace__start,
		      &ModelHandler::mof__import__imported_namespace__end);

	mHandlerMap["MOF:Import"] =
	    make_pair(&ModelHandler::mof__import__start,
		      &ModelHandler::mof__import__end);

	mHandlerMap["MOF:GeneralizableElement.visibility"] =
	    make_pair(&ModelHandler::mof__generalizable_element__visibility__start,
		      &ModelHandler::mof__generalizable_element__visibility__end);

	mHandlerMap["MOF:GeneralizableElement.isAbstract"] =
	    make_pair(&ModelHandler::mof__generalizable_element__is_abstract__start,
		      &ModelHandler::mof__generalizable_element__is_abstract__end);

	mHandlerMap["MOF:GeneralizableElement.isRoot"] =
	    make_pair(&ModelHandler::mof__generalizable_element__is_root__start,
		      &ModelHandler::mof__generalizable_element__is_root__end);

	mHandlerMap["MOF:GeneralizableElement.isLeaf"] =
	    make_pair(&ModelHandler::mof__generalizable_element__is_leaf__start,
		      &ModelHandler::mof__generalizable_element__is_leaf__end);

	mHandlerMap["MOF:GeneralizableElement.supertypes"] =
	    make_pair(&ModelHandler::mof__generalizable_element__supertypes__start,
		      &ModelHandler::mof__generalizable_element__supertypes__end);

	mHandlerMap["MOF:GeneralizableElement"] =
	    make_pair(&ModelHandler::mof__generalizable_element__start,
		      &ModelHandler::mof__generalizable_element__end);

	mHandlerMap["MOF:Constant.constValue"] =
	    make_pair(&ModelHandler::mof__constant__const_value__start,
		      &ModelHandler::mof__constant__const_value__end);

	mHandlerMap["MOF:Constant"] =
	    make_pair(&ModelHandler::mof__constant__start,
		      &ModelHandler::mof__constant__end);

	mHandlerMap["MOF:Reference.referencedEnd"] =
	    make_pair(&ModelHandler::mof__reference__referenced_end__start,
		      &ModelHandler::mof__reference__referenced_end__end);

	mHandlerMap["MOF:Reference"] =
	    make_pair(&ModelHandler::mof__reference__start,
		      &ModelHandler::mof__reference__end);

	mHandlerMap["MOF:MofException"] =
	    make_pair(&ModelHandler::mof__mof_exception__start,
		      &ModelHandler::mof__mof_exception__end);

	mHandlerMap["MOF:DataType.typeCode"] =
	    make_pair(&ModelHandler::mof__data_type__type_code__start,
		      &ModelHandler::mof__data_type__type_code__end);

	mHandlerMap["MOF:DataType"] =
	    make_pair(&ModelHandler::mof__data_type__start,
		      &ModelHandler::mof__data_type__end);

	mHandlerMap["MOF:AssociationEnd.multiplicity"] =
	    make_pair(&ModelHandler::mof__association_end__multiplicity__start,
		      &ModelHandler::mof__association_end__multiplicity__end);

	mHandlerMap["MOF:AssociationEnd.aggregation"] =
	    make_pair(&ModelHandler::mof__association_end__aggregation__start,
		      &ModelHandler::mof__association_end__aggregation__end);

	mHandlerMap["MOF:AssociationEnd.isNavigable"] =
	    make_pair(&ModelHandler::mof__association_end__is_navigable__start,
		      &ModelHandler::mof__association_end__is_navigable__end);

	mHandlerMap["MOF:AssociationEnd.isChangeable"] =
	    make_pair(&ModelHandler::mof__association_end__is_changeable__start,
		      &ModelHandler::mof__association_end__is_changeable__end);

	mHandlerMap["MOF:AssociationEnd.otherEnd"] =
	    make_pair(&ModelHandler::mof__association_end__other_end__start,
		      &ModelHandler::mof__association_end__other_end__end);

	mHandlerMap["MOF:AssociationEnd"] =
	    make_pair(&ModelHandler::mof__association_end__start,
		      &ModelHandler::mof__association_end__end);

	mHandlerMap["MOF:TypedElement.type"] =
	    make_pair(&ModelHandler::mof__typed_element__type__start,
		      &ModelHandler::mof__typed_element__type__end);

	mHandlerMap["MOF:TypedElement"] =
	    make_pair(&ModelHandler::mof__typed_element__start,
		      &ModelHandler::mof__typed_element__end);

	mHandlerMap["MOF:Multiplicity.lower"] =
	    make_pair(&ModelHandler::mof__multiplicity__lower__start,
		      &ModelHandler::mof__multiplicity__lower__end);

	mHandlerMap["MOF:Multiplicity.upper"] =
	    make_pair(&ModelHandler::mof__multiplicity__upper__start,
		      &ModelHandler::mof__multiplicity__upper__end);

	mHandlerMap["MOF:Multiplicity.isOrdered"] =
	    make_pair(&ModelHandler::mof__multiplicity__is_ordered__start,
		      &ModelHandler::mof__multiplicity__is_ordered__end);

	mHandlerMap["MOF:Multiplicity.isUnique"] =
	    make_pair(&ModelHandler::mof__multiplicity__is_unique__start,
		      &ModelHandler::mof__multiplicity__is_unique__end);

	mHandlerMap["MOF:Multiplicity"] =
	    make_pair(&ModelHandler::mof__multiplicity__start,
		      &ModelHandler::mof__multiplicity__end);

	mHandlerMap["MOF:Tag.tagId"] =
	    make_pair(&ModelHandler::mof__tag__tag_id__start,
		      &ModelHandler::mof__tag__tag_id__end);

	mHandlerMap["MOF:Tag.values"] =
	    make_pair(&ModelHandler::mof__tag__values__start,
		      &ModelHandler::mof__tag__values__end);

	mHandlerMap["MOF:Tag.elements"] =
	    make_pair(&ModelHandler::mof__tag__elements__start,
		      &ModelHandler::mof__tag__elements__end);

	mHandlerMap["MOF:Tag"] =
	    make_pair(&ModelHandler::mof__tag__start,
		      &ModelHandler::mof__tag__end);

	mHandlerMap["MOF:Mof"] =
	    make_pair(&ModelHandler::mof__mof__start,
		      &ModelHandler::mof__mof__end);

	mHandlerMap["MOF:Model"] =
	    make_pair(&ModelHandler::mof__model__start,
		      &ModelHandler::mof__model__end);

	// data types

	mHandlerMap["MOF:AliasType"] =
	    make_pair(&ModelHandler::mof__alias_type__start,
		      &ModelHandler::mof__alias_type__end);

	mHandlerMap["MOF:PrimitiveType"] =
	    make_pair(&ModelHandler::mof__primitive_type__start,
		      &ModelHandler::mof__primitive_type__end);

	mHandlerMap["MOF:StructureType"] =
	    make_pair(&ModelHandler::mof__structure_type__start,
		      &ModelHandler::mof__structure_type__end);

	mHandlerMap["MOF:StructureField"] =
	    make_pair(&ModelHandler::mof__structure_field__start,
		      &ModelHandler::mof__structure_field__end);

	mHandlerMap["MOF:CollectionType"] =
	    make_pair(&ModelHandler::mof__collection_type__start,
		      &ModelHandler::mof__collection_type__end);

	mHandlerMap["MOF:EnumerationType"] =
	    make_pair(&ModelHandler::mof__enumeration_type__start,
		      &ModelHandler::mof__enumeration_type__end);

	mHandlerMap["MOF:EnumerationType.labels"] =
	    make_pair(&ModelHandler::mof__enumeration_type__labels__start,
		      &ModelHandler::mof__enumeration_type__labels__end);
    }

    ModelHandler::~ModelHandler() {
    }

    void
    ModelHandler::warning(SAXParser * parser, const string & msg) {
	WARNING("Unhandled warning: " << msg);
    }

    void
    ModelHandler::error(SAXParser * parser, const string & msg) {
	WARNING("Unhandled error: " << msg);
    }

    void
    ModelHandler::fatal(SAXParser * parser, const string & msg) {
	WARNING("Unhandled fatal: " << msg);
    }

    void
    ModelHandler::comment(SAXParser * parser, const string & comment) {
    }

    void
    ModelHandler::start_document(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
    }

    void
    ModelHandler::end_document(SAXParser * parser) {
	if(parser->get_pass() == 2) {
	    for(hash_map<string, Object *>::iterator iter = mElementMap.begin();
		iter != mElementMap.end();
		++iter) {
		iter->second->unref();
	    }

	    mElementMap.clear();
	}
    }

    void
    ModelHandler::start_element(SAXParser * parser, const string & name,
				const Properties & properties) {
	TRACE(5, "Start: " << name);

	mCharacters.resize(0);

	hash_map<string, HandlerPair>::iterator iter = mHandlerMap.find(name);

	if(iter == mHandlerMap.end()) {
	    if(mUnhandledElements.insert(name).second) {
		WARNING("Unhandled element: " << name);
	    }

	    // push a dumb handler so that we don't try to parse
	    // the contents of the tag

	    XML::SAXHandler * handler = new Common::DumbHandler();
	    parser->use(handler);
	    handler->unref();
	}
	else {
	    HandlerPair handler_pair = iter->second;
	    (this->*handler_pair.first)(parser, properties);
	}
    }

    void
    ModelHandler::end_element(SAXParser * parser, const string & name) {
	hash_map<string, HandlerPair>::iterator iter = mHandlerMap.find(name);

	if(iter == mHandlerMap.end()) {
	}
	else {
	    HandlerPair handler_pair = iter->second;
	    (this->*handler_pair.second)(parser);
	}

	TRACE(5, "End: " << name);
    }

    void
    ModelHandler::characters(SAXParser * parser, const string & data) {
	mCharacters += data;
    }

    void
    ModelHandler::reference(SAXParser * parser, const string & name) {
	WARNING("Unhandled Reference: " << name);
    }

    void
    ModelHandler::cdata_block(SAXParser * parser, const string & data) {
	WARNING("Unhandled CDATA: " << data);
    }

    //
    // Feature
    //

    void
    ModelHandler::mof__feature__visibility__start(SAXParser * parser,
						  const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Feature * feature = known_cast<Feature *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    VisibilityKind * visibility = new VisibilityKind();
	    visibility->ref_set_value(value);

	    feature->set_visibility(visibility);

	    visibility->unref();
	}
    }

    void
    ModelHandler::mof__feature__visibility__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__feature__scope__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Feature * feature = known_cast<Feature *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    ScopeKind * scope = new ScopeKind();
	    scope->ref_set_value(value);

	    feature->set_scope(scope);

	    scope->unref();
	}
    }

    void
    ModelHandler::mof__feature__scope__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__feature__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__feature__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Classifier
    //

    void
    ModelHandler::mof__classifier__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__classifier__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Namespace
    //

    void
    ModelHandler::mof__namespace__contents__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__namespace__contents__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    list<ModelElement *> contents;

	    transform(objects.begin(), objects.end(), back_inserter(contents),
		      Caster<Object, ModelElement>());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    Namespace * namspace = known_cast<Namespace *>(parser->top().back());

	    namspace->set_contained_elements(contents);

	    for_each(contents.begin(), contents.end(), mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__namespace__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__namespace__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Structural Feature
    //

    void
    ModelHandler::mof__structural_feature__multiplicity__start(SAXParser * parser,
							       const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__structural_feature__multiplicity__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    StructuralFeature * structural_feature =
		known_cast<StructuralFeature *>(parser->top().back());

	    structural_feature->set_multiplicity(multiplicity);

	    multiplicity->unref();
	}
    }

    void
    ModelHandler::mof__structural_feature__is_changeable__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    StructuralFeature * structural_feature =
		known_cast<StructuralFeature *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * changeable = new Boolean();
	    changeable->ref_set_value(value);

	    structural_feature->set_changeable(changeable);

	    changeable->unref();
	}
    }

    void
    ModelHandler::mof__structural_feature__is_changeable__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__structural_feature__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__structural_feature__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Attribute
    //

    void
    ModelHandler::mof__attribute__is_derived__start(SAXParser * parser,
							const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Attribute * attribute = known_cast<Attribute *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * derived = new Boolean();
	    derived->ref_set_value(value);

	    attribute->set_derived(derived);

	    derived->unref();
	}
    }

    void
    ModelHandler::mof__attribute__is_derived__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__attribute__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Attribute());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * attribute;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		attribute = lookup(idref);
		attribute->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		attribute = lookup(id);
		attribute->ref();
	    }
	    else {
		attribute = new Attribute();
	    }

	    ASSERT(attribute);

	    parser->top().push_back(attribute);
	}
    }

    void
    ModelHandler::mof__attribute__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Operation
    //

    void
    ModelHandler::mof__operation__is_query__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Operation * operation = known_cast<Operation *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * query = new Boolean();
	    query->ref_set_value(value);

	    operation->set_query(query);

	    query->unref();
	}
    }

    void
    ModelHandler::mof__operation__is_query__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__operation__exceptions__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__operation__exceptions__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    list<Exception *> exceptions;

	    transform(objects.begin(), objects.end(), back_inserter(exceptions),
		      Caster<Object, Exception>());

	    parser->pop();

	    //

	    ASSERT(parser->top().size() == 1);

	    Operation * operation = known_cast<Operation *>(parser->top().back());

	    operation->set_exceptions(exceptions);

	    for_each(exceptions.begin(), exceptions.end(), mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__operation__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Operation());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * operation;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		operation = lookup(idref);
		operation->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		operation = lookup(id);
		operation->ref();
	    }
	    else {
		operation = new Operation();
	    }

	    ASSERT(operation);

	    parser->top().push_back(operation);
	}
    }

    void
    ModelHandler::mof__operation__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Class
    //

    void
    ModelHandler::mof__class__is_singleton__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Class * klass = known_cast<Class *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * singleton = new Boolean();
	    singleton->ref_set_value(value);

	    klass->set_singleton(singleton);

	    singleton->unref();
	}
    }

    void
    ModelHandler::mof__class__is_singleton__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__class__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Class());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * klass;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		klass = lookup(idref);
		klass->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		klass = lookup(id);
		klass->ref();
	    }
	    else {
		klass = new Class();
	    }

	    ASSERT(klass);

	    parser->top().push_back(klass);
	}
    }

    void
    ModelHandler::mof__class__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // ModelElement
    //

    void
    ModelHandler::mof__model_element__name__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__model_element__name__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    ModelElement * model_element = known_cast<ModelElement *>(parser->top().back());

	    Name * name = new Name();
	    name->set_value(mCharacters);

	    model_element->set_name(name);

	    name->unref();
	}
    }

    void
    ModelHandler::mof__model_element__annotation__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__model_element__annotation__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    ModelElement * model_element = known_cast<ModelElement *>(parser->top().back());

	    Name * annotation = new Name();
	    annotation->set_value(mCharacters);

	    model_element->set_annotation(annotation);

	    annotation->unref();
	}
    }

    void
    ModelHandler::mof__model_element__container__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__model_element__container__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Namespace * container = known_cast<Namespace *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    ModelElement * model_element = known_cast<ModelElement *>(parser->top().back());

	    model_element->set_container(container);

	    container->unref();
	}
    }

    void
    ModelHandler::mof__model_element__constraints__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__model_element__constraints__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    set<Constraint *> constraints;

	    transform(objects.begin(), objects.end(), associative_inserter(constraints),
		      Caster<Object, Constraint>());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    ModelElement * model_element = known_cast<ModelElement *>(parser->top().back());

	    model_element->set_constraints(constraints);

	    for_each(constraints.begin(), constraints.end(), mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__model_element__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__model_element__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Package
    //

    void
    ModelHandler::mof__package__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Package());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * package;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		package = lookup(idref);
		package->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		package = lookup(id);
		package->ref();
	    }
	    else {
		package = new Package();
	    }

	    ASSERT(package);

	    parser->top().push_back(package);
	}
    }

    void
    ModelHandler::mof__package__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Constraint
    //

    void
    ModelHandler::mof__constraint__expression__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__constraint__expression__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->pop();
	}
    }

    void
    ModelHandler::mof__constraint__language__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__constraint__language__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    Constraint * constraint = known_cast<Constraint *>(parser->top().back());

	    Name * language = new Name();
	    language->set_value(mCharacters);

	    constraint->set_language(language);

	    language->unref();
	}
    }

    void
    ModelHandler::mof__constraint__evaluation_policy__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    Constraint * constraint = known_cast<Constraint *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    EvaluationKind * evaluation_policy = new EvaluationKind();
	    evaluation_policy->ref_set_value(value);

	    constraint->set_evaluation_policy(evaluation_policy);

	    evaluation_policy->unref();
	}
    }

    void
    ModelHandler::mof__constraint__evaluation_policy__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__constraint__constrained_elements__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__constraint__constrained_elements__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    set<ModelElement *> constrained_elements;

	    transform(objects.begin(), objects.end(), associative_inserter(constrained_elements),
		      Caster<Object, ModelElement>());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    Constraint * constraint = known_cast<Constraint *>(parser->top().back());

	    constraint->set_constrained_elements(constrained_elements);

	    for_each(constrained_elements.begin(), constrained_elements.end(),
		     mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__constraint__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Constraint());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * constraint;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		constraint = lookup(idref);
		constraint->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		constraint = lookup(id);
		constraint->ref();
	    }
	    else {
		constraint = new Constraint();
	    }

	    ASSERT(constraint);

	    parser->top().push_back(constraint);
	}
    }

    void
    ModelHandler::mof__constraint__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Association
    //

    void
    ModelHandler::mof__association__is_derived__start(SAXParser * parser,
						      const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    Association * association = known_cast<Association *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * derived = new Boolean();
	    derived->ref_set_value(value);

	    association->set_derived(derived);

	    derived->unref();
	}
    }

    void
    ModelHandler::mof__association__is_derived__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__association__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Association());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * association;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		association = lookup(idref);
		association->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		association = lookup(id);
		association->ref();
	    }
	    else {
		association = new Association();
	    }

	    parser->top().push_back(association);
	}
    }

    void
    ModelHandler::mof__association__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Parameter
    //

    void
    ModelHandler::mof__parameter__direction__start(SAXParser * parser,
						   const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    Parameter * parameter = known_cast<Parameter *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    DirectionKind * direction = new DirectionKind();
	    direction->ref_set_value(value);

	    parameter->set_direction(direction);

	    direction->unref();
	}
    }

    void
    ModelHandler::mof__parameter__direction__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__parameter__multiplicity__start(SAXParser * parser,
						      const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__parameter__multiplicity__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    Parameter * parameter = known_cast<Parameter *>(parser->top().back());

	    parameter->set_multiplicity(multiplicity);

	    multiplicity->unref();
	}
    }

    void
    ModelHandler::mof__parameter__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Parameter());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * parameter;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		parameter = lookup(idref);
		parameter->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		parameter = lookup(id);
		parameter->ref();
	    }
	    else {
		parameter = new Parameter();
	    }

	    ASSERT(parameter);

	    parser->top().push_back(parameter);
	}
    }

    void
    ModelHandler::mof__parameter__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Behavioral Feature
    //

    void
    ModelHandler::mof__behavioral_feature__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__behavioral_feature__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Import
    //

    void
    ModelHandler::mof__import__visibility__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    Import * import = known_cast<Import *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    VisibilityKind * visibility = new VisibilityKind();
	    visibility->ref_set_value(value);

	    import->set_visibility(visibility);

	    visibility->unref();
	}
    }

    void
    ModelHandler::mof__import__visibility__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__import__imported_namespace__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__import__imported_namespace__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Namespace * namspace = known_cast<Namespace *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    Import * import = known_cast<Import *>(parser->top().back());

	    import->set_imported(namspace);

	    namspace->unref();
	}
    }

    void
    ModelHandler::mof__import__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Import());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * import;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		import = lookup(idref);
		import->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		import = lookup(id);
		import->ref();
	    }
	    else {
		import = new Import();
	    }

	    ASSERT(import);

	    parser->top().push_back(import);
	}
    }

    void
    ModelHandler::mof__import__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Generalizable Element
    //

    void
    ModelHandler::mof__generalizable_element__visibility__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    GeneralizableElement * generalizable_element =
		known_cast<GeneralizableElement *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    VisibilityKind * visibility = new VisibilityKind();
	    visibility->ref_set_value(value);

	    generalizable_element->set_visibility(visibility);

	    visibility->unref();
	}
    }

    void
    ModelHandler::mof__generalizable_element__visibility__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_abstract__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    GeneralizableElement * generalizable_element =
		known_cast<GeneralizableElement *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * abstract = new Boolean();
	    abstract->ref_set_value(value);

	    generalizable_element->set_abstract(abstract);

	    abstract->unref();
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_abstract__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_root__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    GeneralizableElement * generalizable_element =
		known_cast<GeneralizableElement *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * root = new Boolean();
	    root->ref_set_value(value);

	    generalizable_element->set_root(root);

	    root->unref();
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_root__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_leaf__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    GeneralizableElement * generalizable_element =
		known_cast<GeneralizableElement *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * leaf = new Boolean();
	    leaf->ref_set_value(value);

	    generalizable_element->set_leaf(leaf);

	    leaf->unref();
	}
    }

    void
    ModelHandler::mof__generalizable_element__is_leaf__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__generalizable_element__supertypes__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__generalizable_element__supertypes__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    list<GeneralizableElement *> supertypes;

	    transform(objects.begin(), objects.end(), back_inserter(supertypes),
		      Caster<Object, GeneralizableElement>());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    GeneralizableElement * generalizable_element =
		known_cast<GeneralizableElement *>(parser->top().back());

	    generalizable_element->set_supertypes(supertypes);

	    for_each(supertypes.begin(), supertypes.end(), mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__generalizable_element__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__generalizable_element__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Constant
    //

    void
    ModelHandler::mof__constant__const_value__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__constant__const_value__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->pop();
	}
    }

    void
    ModelHandler::mof__constant__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Constant());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * constant;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		constant = lookup(idref);
		constant->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		constant = lookup(id);
		constant->ref();
	    }
	    else {
		constant = new Constant();
	    }

	    ASSERT(constant);

	    parser->top().push_back(constant);
	}
    }

    void
    ModelHandler::mof__constant__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Reference
    //

    void
    ModelHandler::mof__reference__referenced_end__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__reference__referenced_end__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    AssociationEnd * association_end = known_cast<AssociationEnd *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    Reference * reference = known_cast<Reference *>(parser->top().back());

	    reference->set_referenced_end(association_end);

	    association_end->unref();
	}
    }

    void
    ModelHandler::mof__reference__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Reference());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * reference;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		reference = lookup(idref);
		reference->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		reference = lookup(id);
		reference->ref();
	    }
	    else {
		reference = new Reference();
	    }

	    ASSERT(reference);

	    parser->top().push_back(reference);
	}
    }

    void
    ModelHandler::mof__reference__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Exception
    //

    void
    ModelHandler::mof__mof_exception__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Exception());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * exception;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		exception = lookup(idref);
		exception->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		exception = lookup(id);
		exception->ref();
	    }
	    else {
		exception = new Exception();
	    }

	    ASSERT(exception);

	    parser->top().push_back(exception);
	}
    }

    void
    ModelHandler::mof__mof_exception__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Data Type
    //

    void
    ModelHandler::mof__data_type__type_code__start(SAXParser * parser,
						   const Properties & properties) {
	if(parser->get_pass() == 1) {
	    XML::SAXHandler * handler = new Common::DumbHandler();
	    parser->use(handler);
	    handler->unref();
	}
	else {
	    XML::SAXHandler * handler = new Common::TypeCodeHandler();
	    parser->use(handler);
	    parser->push();
	}
    }

    void
    ModelHandler::mof__data_type__type_code__end(SAXParser * parser) {
	// TODO somehow get the result back from the SAX handler
	// that we pushed on...

	if(parser->get_pass() == 1) {
	}
	else {
	    parser->pop();
	}
    }

    void
    ModelHandler::mof__data_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new DataType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * data_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		data_type = lookup(idref);
		data_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		data_type = lookup(id);
		data_type->ref();
	    }
	    else {
		data_type = new DataType();
	    }

	    ASSERT(data_type);

	    parser->top().push_back(data_type);
	}
    }

    void
    ModelHandler::mof__data_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Association
    //

    void
    ModelHandler::mof__association_end__multiplicity__start(SAXParser * parser,
							    const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__association_end__multiplicity__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    AssociationEnd * association_end =
		known_cast<AssociationEnd *>(parser->top().back());

	    association_end->set_multiplicity(multiplicity);

	    multiplicity->unref();
	}
    }

    void
    ModelHandler::mof__association_end__aggregation__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    AssociationEnd * association_end = known_cast<AssociationEnd *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    AggregationKind * aggregation = new AggregationKind();
	    aggregation->ref_set_value(value);

	    association_end->set_aggregation(aggregation);

	    aggregation->unref();
	}
    }

    void
    ModelHandler::mof__association_end__aggregation__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__association_end__is_navigable__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    AssociationEnd * association_end = known_cast<AssociationEnd *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * navigable = new Boolean();
	    navigable->ref_set_value(value);

	    association_end->set_navigable(navigable);

	    navigable->unref();
	}
    }

    void
    ModelHandler::mof__association_end__is_navigable__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__association_end__is_changeable__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    AssociationEnd * association_end = known_cast<AssociationEnd *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");

	    //

	    Boolean * changeable = new Boolean();
	    changeable->ref_set_value(value);

	    association_end->set_changeable(changeable);

	    changeable->unref();
	}
    }

    void
    ModelHandler::mof__association_end__is_changeable__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__association_end__other_end__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__association_end__other_end__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    AssociationEnd * other_end = known_cast<AssociationEnd *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    AssociationEnd * association_end = known_cast<AssociationEnd *>(parser->top().back());

	    association_end->set_other_end(other_end);

	    other_end->unref();
	}
    }

    void
    ModelHandler::mof__association_end__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new AssociationEnd());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * association_end;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		association_end = lookup(idref);
		association_end->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		association_end = lookup(id);
		association_end->ref();
	    }
	    else {
		association_end = new AssociationEnd();
	    }

	    ASSERT(association_end);

	    parser->top().push_back(association_end);
	}
    }

    void
    ModelHandler::mof__association_end__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Typed Element
    //

    void
    ModelHandler::mof__typed_element__type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__typed_element__type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Classifier * type = known_cast<Classifier *>(parser->top().back());

	    parser->pop();

	    //

	    ASSERT(!parser->top().empty());

	    TypedElement * typed_element = known_cast<TypedElement *>(parser->top().back());

	    typed_element->set_type(type);

	    type->unref();
	}
    }

    void
    ModelHandler::mof__typed_element__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__typed_element__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Multiplicity << struct >>
    //

    void
    ModelHandler::mof__multiplicity__lower__start(SAXParser * parser,
						  const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__multiplicity__lower__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    istringstream ist(mCharacters + "\n");

	    long lower;

	    ist >> lower;

	    if(ist.good()) {
		Integer * l = new Integer();
		l->set_value(lower);
		multiplicity->set_lower(l);
		l->unref();
	    }
	    else {
		WARNING("Couldn't parse integer: " << mCharacters);
	    }
	}
    }

    void
    ModelHandler::mof__multiplicity__upper__start(SAXParser * parser,
						  const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__multiplicity__upper__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    istringstream ist(mCharacters + "\n");

	    int upper;
	    ist >> upper;

	    if(ist.good()) {
		Integer * u = new Integer();
		u->set_value(upper);
		multiplicity->set_upper(u);
		u->unref();
	    }
	    else {
		WARNING("Couldn't parse integer: " << mCharacters);
	    }
	}
    }

    void
    ModelHandler::mof__multiplicity__is_ordered__start(SAXParser * parser,
						       const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");
	    bool ordered = value == "true" ? true : false;

	    Boolean * o = new Boolean();
	    o->set_value(ordered);
	    multiplicity->set_ordered(o);
	    o->unref();
	}
    }

    void
    ModelHandler::mof__multiplicity__is_ordered__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__multiplicity__is_unique__start(SAXParser * parser,
						      const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    MultiplicityType * multiplicity =
		known_cast<MultiplicityType *>(parser->top().back());

	    string value = properties.get_required_property("xmi.value");
	    bool unique = value == "true" ? true : false;

	    Boolean * u = new Boolean();
	    u->set_value(unique);
	    multiplicity->set_unique(u);
	    u->unref();
	}
    }

    void
    ModelHandler::mof__multiplicity__is_unique__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__multiplicity__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->top().push_back(new MultiplicityType());
	}
    }

    void
    ModelHandler::mof__multiplicity__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Alias Type
    //

    void
    ModelHandler::mof__alias_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new AliasType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * alias_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		alias_type = lookup(idref);
		alias_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		alias_type = lookup(id);
		alias_type->ref();
	    }
	    else {
		alias_type = new AliasType();
	    }

	    ASSERT(alias_type);

	    parser->top().push_back(alias_type);
	}
    }

    void
    ModelHandler::mof__alias_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Primitive Type
    //

    void
    ModelHandler::mof__primitive_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new PrimitiveType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * primitive_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		primitive_type = lookup(idref);
		primitive_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		primitive_type = lookup(id);
		primitive_type->ref();
	    }
	    else {
		primitive_type = new PrimitiveType();
	    }

	    ASSERT(primitive_type);

	    parser->top().push_back(primitive_type);
	}
    }

    void
    ModelHandler::mof__primitive_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Structure Type
    //

    void
    ModelHandler::mof__structure_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new StructureType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * structure_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		structure_type = lookup(idref);
		structure_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		structure_type = lookup(id);
		structure_type->ref();
	    }
	    else {
		structure_type = new StructureType();
	    }

	    ASSERT(structure_type);

	    parser->top().push_back(structure_type);
	}
    }

    void
    ModelHandler::mof__structure_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Collection Type
    //

    void
    ModelHandler::mof__collection_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new CollectionType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * collection_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		collection_type = lookup(idref);
		collection_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		collection_type = lookup(id);
		collection_type->ref();
	    }
	    else {
		collection_type = new CollectionType();
	    }

	    ASSERT(collection_type);

	    parser->top().push_back(collection_type);
	}
    }

    void
    ModelHandler::mof__collection_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Enumeration Type
    //

    void
    ModelHandler::mof__enumeration_type__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new EnumerationType());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * enumeration_type;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		enumeration_type = lookup(idref);
		enumeration_type->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		enumeration_type = lookup(id);
		enumeration_type->ref();
	    }
	    else {
		enumeration_type = new EnumerationType();
	    }

	    ASSERT(enumeration_type);

	    parser->top().push_back(enumeration_type);
	}
    }

    void
    ModelHandler::mof__enumeration_type__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__enumeration_type__labels__start(SAXParser * parser,
						       const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__enumeration_type__labels__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(!parser->top().empty());

	    EnumerationType * et = known_cast<EnumerationType *>(parser->top().back());

	    Name * name = new Name();
	    name->set_value(mCharacters);

	    et->set_labels(name);

	    name->unref();
	}
    }

    //
    // StructureField
    //

    void
    ModelHandler::mof__structure_field__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new StructureField());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * structure_field;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		structure_field = lookup(idref);
		structure_field->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		structure_field = lookup(id);
		structure_field->ref();
	    }
	    else {
		structure_field = new StructureField();
	    }

	    ASSERT(structure_field);

	    parser->top().push_back(structure_field);
	}
    }

    void
    ModelHandler::mof__structure_field__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // Tag
    //

    void
    ModelHandler::mof__tag__tag_id__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__tag__tag_id__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    ASSERT(parser->top().size() == 1);

	    Tag * tag = known_cast<Tag *>(parser->top().back());

	    Name * tag_id = new Name();
	    tag_id->set_value(mCharacters);

	    tag->set_tag_id(tag_id);

	    tag_id->unref();
	}
    }

    void
    ModelHandler::mof__tag__values__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__tag__values__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    for_each(parser->top().begin(), parser->top().end(),
		     mem_fun(&Object::unref));

	    parser->pop();		// NYI
	}
    }

    void
    ModelHandler::mof__tag__elements__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	    parser->push();
	}
    }

    void
    ModelHandler::mof__tag__elements__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	    const vector<Object *> & objects = parser->top();;
	    set<ModelElement *> elements;

	    transform(objects.begin(), objects.end(), associative_inserter(elements),
		      Caster<Object, ModelElement>());

	    parser->pop();

	    //

	    ASSERT(parser->top().size() == 1);

	    Tag * tag = known_cast<Tag *>(parser->top().back());

	    tag->set_models(elements);

	    for_each(elements.begin(), elements.end(), mem_fun(&Object::unref));
	}
    }

    void
    ModelHandler::mof__tag__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    string id;

	    if(properties.get_implied_property("xmi.id", id)) {
		store(id, new Tag());
	    }
	}
	else {
	    string idref;
	    string id;

	    Object * tag;

	    if(properties.get_implied_property("xmi.idref", idref)) {
		tag = lookup(idref);
		tag->ref();
	    }
	    else if(properties.get_implied_property("xmi.id", id)) {
		tag = lookup(id);
		tag->ref();
	    }
	    else {
		tag = new Tag();
	    }

	    ASSERT(tag);

	    parser->top().push_back(tag);
	}
    }

    void
    ModelHandler::mof__tag__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    // MOF
    //

    void
    ModelHandler::mof__mof__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	    ASSERT_NOT_REACHED();
	}
	else {
	}
    }

    void
    ModelHandler::mof__mof__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    //
    //

    void
    ModelHandler::mof__model__start(SAXParser * parser, const Properties & properties) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    void
    ModelHandler::mof__model__end(SAXParser * parser) {
	if(parser->get_pass() == 1) {
	}
	else {
	}
    }

    //
    //
    //

    void
    ModelHandler::store(const string & id, Object * object) {
	if(!mElementMap.insert(make_pair(id, object)).second) {
	    ERROR("Duplicate id: " << id);
	}
    }

    Object *
    ModelHandler::lookup(const string & id) {
	hash_map<string, Object *>::iterator iter = mElementMap.find(id);

	if(iter == mElementMap.end()) {
	    ERROR("Couldn't locate element with id: " << id);
	    throw;	// FIXME
	}
	else {
	    return iter->second;
	}
    }

} // namespace MOF
