// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // CONFIG_H

#include "operation.h"

#include "exception.h"
#include "parameter.h"
#include "constraint.h"
#include "tag.h"
#include "model.h"

#include "foundation/macros.h"
#include "foundation/commands.h"
#include "reflective/changes.h"

#include <algorithm>
#include <vector>
#include <iterator>
#include <typeinfo>
using namespace std;

namespace MOF {

    const string Operation::TYPE = "Operation";

    //

    Operation::Operation() : BehavioralFeature(),
        mQuery(false) {
    }

    Operation::~Operation() {
    }

    void
    Operation::add_exception(Exception * exception, int index) {
	RETURN_IF_FAIL(index <= static_cast<int>(mExceptions.size()));

	if(index < 0) {
	    index = mExceptions.size();
	}

	list<Exception *>::iterator iter = mExceptions.begin();

	for(int i = 0; i < static_cast<int>(mExceptions.size()); ++i) {
	    RETURN_IF_FAIL(*iter != exception);

	    if(i < index) {
		++iter;
	    }
	}

	//

	begin_transaction();

	Command * command = command_flip_method(this,
						&Operation::add_exception,
						&Operation::remove_exception,
						exception, index);
	log_command(command);
	command->unref();

	// make the changes

	exception->ref();

	mExceptions.insert(iter, exception);
	exception->internal_add_operation(this);

	// changes made

	ChangeEvent * change = change_role_added(this, "exception", exception, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Operation::remove_exception(Exception * exception) {
	int index = 0;

	list<Exception *>::iterator iter = mExceptions.begin();

	for(; iter != mExceptions.end(); ++iter, ++index) {
	    if(*iter == exception) {
		break;
	    }
	}

	RETURN_IF_FAIL(iter != mExceptions.end());

	//

	begin_transaction();

	Command * command = command_flip_method(this,
						&Operation::remove_exception,
						&Operation::add_exception,
						exception, index);
	log_command(command);
	command->unref();

	// make the changes


	exception->internal_remove_operation(this);
	mExceptions.erase(iter);

	exception->unref();

	// changes made

	ChangeEvent * change = change_role_removed(this, "exception", exception, index);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
    Operation::set_exceptions(const list<Exception *> & exceptions) {
	set<Exception *> before;
	set<Exception *> after;

	before.insert(mExceptions.begin(), mExceptions.end());
	after.insert(exceptions.begin(), exceptions.end());

	// check for duplicates
	RETURN_IF_FAIL(before.size() == mExceptions.size());
	RETURN_IF_FAIL(after.size() == exceptions.size());

	// work out what is being added and what is being removed

	vector<Exception *> added;
	vector<Exception *> removed;

	set_difference(after.begin(), after.end(), 
		       before.begin(), before.end(), back_inserter(added));

	set_difference(before.begin(), before.end(), 
		       after.begin(), after.end(), back_inserter(removed));

	// transaction time

	begin_transaction();

	Command * command = command_flip_argument(this,
						  &Operation::set_exceptions,
						  exceptions, mExceptions);
	log_command(command);
	command->unref();

	// make the changes

	for_each(added.begin(), added.end(),
		 bind2nd(mem_fun(&Exception::internal_add_operation), this));
	for_each(removed.begin(), removed.end(),
		 bind2nd(mem_fun(&Exception::internal_remove_operation), this));

	for_each(added.begin(), added.end(), mem_fun(&Object::ref));
	for_each(removed.begin(), removed.end(), mem_fun(&Object::unref));

	mExceptions = exceptions;

	// changes made

	for(vector<Exception *>::iterator iter = added.begin();
	    iter != added.end();
	    ++iter) {
	    Exception * exception = *iter;
	    ChangeEvent * change = change_role_added(this, "exception", exception, false);
	    log_change(change);
	    change->unref();
	}

	for(vector<Exception *>::iterator iter = removed.begin();
	    iter != removed.end();
	    ++iter) {
	    Exception * exception = *iter;
	    ChangeEvent * change = change_role_removed(this, "exception", exception, false);
	    log_change(change);
	    change->unref();
	}

	//

	end_transaction();
    }

    void
    Operation::set_query(const Boolean * query) {
	begin_transaction();

	//

	Boolean * old_value = mQuery ? new Boolean(*mQuery) : 0;
	Boolean * new_value = query ? new Boolean(*query) : 0;

	//

	Command * command = command_flip_argument(this, &Operation::set_query,
						  new_value, old_value);
	log_command(command);
	command->unref();

	// make the changes

	if(mQuery) {
	    mQuery->unref();
	}

	mQuery = new_value ? new Boolean(*new_value) : 0;

	// changes made

	ChangeEvent * change = change_attribute_changed(this, "query", new_value, old_value);
	log_change(change);
	change->unref();

	//

	if(new_value) {
	    new_value->unref();
	}

	if(old_value) {
	    old_value->unref();
	}

	//

	end_transaction();
    }

    //

    void
    Operation::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    list<Exception *> empty_list;
	    set_exceptions(empty_list);

	    BehavioralFeature::ref_delete_impl(visited);
	}
    }

    bool
    Operation::ref_is_a_type_of_impl(set<string> & visited,
				const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || BehavioralFeature::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Operation::ref_get_type() const {
	return TYPE;
    }

    void
    Operation::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_operation(this);

	    BehavioralFeature::ref_implant_impl(visited);
	}
    }

    void
    Operation::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_operation(this);

	    BehavioralFeature::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
