// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // CONFIG_H

#include "reference.h"

#include "association_end.h"
#include "model.h"

#include "foundation/commands.h"
#include "reflective/changes.h"

namespace MOF {

    const string Reference::TYPE = "Reference";

    //

    Reference::Reference() : StructuralFeature(),
	mReferencedEnd(0),
        mExposedEnd(0) {
    }

    Reference::~Reference() {
    }

    void
    Reference::set_referenced_end(AssociationEnd * referenced_end) {
	begin_transaction();

	Command * command = command_flip_argument(this, &Reference::set_referenced_end,
						  referenced_end, mReferencedEnd);
	log_command(command);
	command->unref();

	// make the changes

	if(mReferencedEnd) {
	    mReferencedEnd->internal_remove_referent(this);
	    mReferencedEnd->unref();
	}

	mReferencedEnd = referenced_end;

	if(mReferencedEnd) {
	    mReferencedEnd->ref();
	    mReferencedEnd->internal_add_referent(this);
	}

	// changes made

	ChangeEvent * change = change_role_changed(this, "referenced_end", referenced_end,
						   mReferencedEnd, false);
	log_change(change);
	change->unref();

	//

	end_transaction();
    }

    void
	Reference::set_exposed_end(AssociationEnd * exposed_end) {
	    begin_transaction();

	    Command * command = command_flip_argument(this, &Reference::set_exposed_end,
						      exposed_end, mExposedEnd);
	    log_command(command);
	    command->unref();

	    // make the changes

	    if(mExposedEnd) {
		mExposedEnd->internal_remove_referrer(this);
		mExposedEnd->unref();
	    }

	    mExposedEnd = exposed_end;

	    if(mExposedEnd) {
		mExposedEnd->ref();
		mExposedEnd->internal_add_referrer(this);
	    }

	    // changes made

	    ChangeEvent * change = change_role_changed(this, "exposed_end", exposed_end,
						       mExposedEnd, false);
	    log_change(change);
	    change->unref();

	    //

	    end_transaction();
    }

    //

    void
    Reference::ref_delete_impl(set<string> & visited) {
	if(visited.insert(TYPE).second) {
	    set_referenced_end(static_cast<AssociationEnd *>(0));
	    set_exposed_end(static_cast<AssociationEnd *>(0));

	    StructuralFeature::ref_delete_impl(visited);
	}
    }

    //

    bool
    Reference::ref_is_a_type_of_impl(set<string> & visited, const string & class_name) const {
	if(visited.insert(TYPE).second) {
	    return class_name == TYPE || StructuralFeature::ref_is_a_type_of(class_name);
	}
	else {
	    return false;
	}
    }

    const string &
    Reference::ref_get_type() const {
	return TYPE;
    }

    //

    void
    Reference::ref_implant_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_add_reference(this);

	    StructuralFeature::ref_implant_impl(visited);
	}
    }

    void
    Reference::ref_uproot_impl(set<string> & visited) throw() {
	if(visited.insert(TYPE).second) {
	    REF::Context * context = ref_get_context();

	    ASSERT(context);

	    Model * model = known_cast<Model *>(context);

	    model->internal_remove_reference(this);

	    StructuralFeature::ref_uproot_impl(visited);
	}
    }

} // namespace MOF
