#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "lexer.h"

#include <iostream>
using namespace std;

#include "foundation/macros.h"

#include "common/characters.h"

namespace OCL {

    using namespace Common;

    LexicalException::LexicalException(const string & details) : Exception(details) {
    }

    LexicalException::~LexicalException() {
    }

    //
    //
    //

    ExpectedToken::ExpectedToken(const string & details) : LexicalException(details) {
    }

    ExpectedToken::~ExpectedToken() {
    }

    //
    //
    //

    BadToken::BadToken(const string & details) : LexicalException(details) {
    }

    BadToken::~BadToken() {
    }

    //
    //
    //

    Lexer::Lexer(istream & input) : mInput(input), mOffset(-1) {
    }

    Lexer::~Lexer() {
	//ASSERT(mMarks.empty());

	if(!mMarks.empty()) {
	    cerr << "Warning, " << mMarks.size() << " marks left" << endl;
	}
    }

    TokenKind
    Lexer::lookahead(int amount) {
	for(int i = 0; i < 1 + mOffset + amount - static_cast<int>(mDeque.size()); ++i) {
	    fetch();
	}

	return mDeque[mOffset + amount].first;
    }

    void
    Lexer::match(TokenKind token) {
	shift();

	if(mDeque[mOffset].first != token) {
	    //cerr << "Expected: " << token << endl;
	    throw ExpectedToken("Expected a different token");
	}
    }

    const string &
    Lexer::value() {
	ASSERT(!mDeque.empty());
	return mDeque[mOffset].second;
    }

    //
    //
    //

    void
    Lexer::push() {
	mMarks.push(mOffset);
    }

    void
    Lexer::pop() {
	ASSERT(!mMarks.empty());
	mMarks.pop();

	if(mMarks.empty()) {
	    for(; mOffset > 0; --mOffset) {
		mDeque.pop_front();
	    }
	}
    }

    void
    Lexer::restore() {
	mOffset = mMarks.top();
	mMarks.pop();
    }

    //
    //
    //

    void
    Lexer::shift() {
	++mOffset;

	if(mOffset == static_cast<int>(mDeque.size())) {
	    fetch();
	}
	else if(mMarks.empty()) {
	    mDeque.pop_front();
	    --mOffset;
	}

	//cerr << "LEXER GOT: " << mDeque[mOffset].first << endl;
    }

    void
    Lexer::fetch() {
	if(mInput.eof()) {
	    mDeque.push_back(make_pair(TOKEN_EOF, string()));
	    return;
	}

TRYAGAIN:
	char c;

	// skip the whitespace

	do {
	    mInput.get(c);
	} while(!mInput.eof() && is_space(c));

	// start checking for tokens

	if(mInput.eof()) {
	    mDeque.push_back(make_pair(TOKEN_EOF, string()));
	    return;
	}
	else if(c == '(') {
	    mDeque.push_back(make_pair(TOKEN_LEFT_ROUND_BRACKET, string()));
	    return;
	}
	else if(c == ')') {
	    mDeque.push_back(make_pair(TOKEN_RIGHT_ROUND_BRACKET, string()));
	    return;
	}
	else if(c == '[') {
	    mDeque.push_back(make_pair(TOKEN_LEFT_SQUARE_BRACKET, string()));
	    return;
	}
	else if(c == ']') {
	    mDeque.push_back(make_pair(TOKEN_RIGHT_SQUARE_BRACKET, string()));
	    return;
	}
	else if(c == '{') {
	    mDeque.push_back(make_pair(TOKEN_LEFT_CURLY_BRACKET, string()));
	    return;
	}
	else if(c == '}') {
	    mDeque.push_back(make_pair(TOKEN_RIGHT_CURLY_BRACKET, string()));
	    return;
	}
	else if(c == ':') {
	    if(mInput.peek() == ':') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_DOUBLE_COLON, string()));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_COLON, string()));
		return;
	    }
	}
	else if(c == ',') {
	    mDeque.push_back(make_pair(TOKEN_COMMA, string()));
	    return;
	}
	else if(c == '=') {
	    mDeque.push_back(make_pair(TOKEN_EQUAL, string()));
	    return;
	}
	else if(c == '<') {
	    if(mInput.peek() == '>') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_NOT_EQUAL, string()));
		return;
	    }
	    else if(mInput.peek() == '=') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_LESS_EQUAL, string()));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_LESS_THAN, string()));
		return;
	    }
	}
	else if(c == '>') {
	    if(mInput.peek() == '=') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_GREATER_EQUAL, string()));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_GREATER_THAN, string()));
		return;
	    }
	}
	else if(c == '|') {
	    mDeque.push_back(make_pair(TOKEN_BAR, string()));
	    return;
	}
	else if(c == '.') {
	    if(mInput.peek() == '.') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_DOT_DOT, string()));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_DOT, string()));
		return;
	    }
	}
	else if(c == '*') {
	    mDeque.push_back(make_pair(TOKEN_MULTIPLY, string()));
	    return;
	}
	else if(c == '/') {
	    mDeque.push_back(make_pair(TOKEN_DIVIDE, string()));
	    return;
	}
	else if(c == '+') {
	    mDeque.push_back(make_pair(TOKEN_PLUS, string()));
	    return;
	}
	else if(c == '-') {
	    if(mInput.peek() == '-') {
		do {
		    mInput.get(c);
		} while(!mInput.eof() && c != '\n');

		goto TRYAGAIN;
	    }
	    else if(mInput.peek() == '>') {
		mInput.get(c);
		mDeque.push_back(make_pair(TOKEN_RIGHT_ARROW, string()));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_MINUS, string()));
		return;
	    }
	}
	else if(c == ';') {
	    mDeque.push_back(make_pair(TOKEN_SEMICOLON, string()));
	    return;
	}
	else if(c == '@') {
	    mDeque.push_back(make_pair(TOKEN_AT, string()));
	    return;
	}
	else if(c == '\'') {
	    bool had_escape = false;

	    string value;

	    for(;;) {
		mInput.get(c);
		value.push_back(c);

		if(mInput.eof()) {
		    cerr << "Unexpected EOF" << endl;
		    ASSERT_NOT_REACHED();
		}
		else {
		    if(!had_escape && c == '\'') {
			mDeque.push_back(make_pair(TOKEN_STRING, value));
			return;
		    }
		}
	    }
	}
	else if(is_digit(c)) {
	    string number;

	    number.push_back(c);

	    while(is_digit(mInput.peek())) {
		mInput.get(c);
		number.push_back(c);
	    }

	    if(mInput.peek() == '.') {
		// it is range...
		mInput.get(c);

		if(mInput.peek() == '.') {
		    mInput.putback(c);
		}
		else {
		    number.push_back(c);

		    while(is_digit(mInput.peek())) {
			mInput.get(c);
			number.push_back(c);
		    }
		}
	    }

	    if(mInput.peek() == 'E' || mInput.peek() == 'e') {
		mInput.get(c);
		number.push_back(c);

		if(mInput.peek() == '+' || mInput.peek() == '-') {
		    mInput.get(c);
		    number.push_back(c);
		}

		while(is_digit(mInput.peek())) {
		    mInput.get(c);
		    number.push_back(c);
		}
	    }

	    mDeque.push_back(make_pair(TOKEN_NUMBER, number));
	    return;
	}
	else if(is_alpha(c)) {
	    string name;
	    
	    name.push_back(c);

	    while(is_ident(mInput.peek())) {
		mInput.get(c);
		name.push_back(c);
	    }

	    if(name == "if") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_IF, name));
		return;
	    }
	    else if(name == "then") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_THEN, name));
		return;
	    }
	    else if(name == "else") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_ELSE, name));
		return;
	    }
	    else if(name == "endif") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_ENDIF, name));
		return;
	    }
	    else if(name == "context") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_CONTEXT, name));
		return;
	    }
	    else if(name == "inv") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_INV, name));
		return;
	    }
	    else if(name == "pre") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_PRE, name));
		return;
	    }
	    else if(name == "def") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_DEF, name));
		return;
	    }
	    else if(name == "post") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_POST, name));
		return;
	    }
	    else if(name == "let") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_LET, name));
		return;
	    }
	    else if(name == "in") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_IN, name));
		return;
	    }
	    else if(name == "enum") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_ENUM, name));
		return;
	    }
	    else if(name == "and") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_AND, name));
		return;
	    }
	    else if(name == "or") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_OR, name));
		return;
	    }
	    else if(name == "xor") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_XOR, name));
		return;
	    }
	    else if(name == "not") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_NOT, name));
		return;
	    }
	    else if(name == "implies") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_IMPLIES, name));
		return;
	    }
	    else if(name == "Set") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_SET, name));
		return;
	    }
	    else if(name == "Bag") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_BAG, name));
		return;
	    }
	    else if(name == "Sequence") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_SEQUENCE, name));
		return;
	    }
	    else if(name == "Collection") {
		mDeque.push_back(make_pair(TOKEN_LITERAL_COLLECTION, name));
		return;
	    }
	    else {
		mDeque.push_back(make_pair(TOKEN_NAME, name));
		return;
	    }
	}
	else {
	    cerr << "Unrecognised token" << endl;
	    throw BadToken("Bad token");
	}
    }

    bool
    Lexer::is_ident(char c) {
	return is_alpha(c) || is_digit(c) || c == '_';
    }

    ostream &
    operator << (ostream & ost, TokenKind token_kind) {
	static const char * const tokens[] = {
	    "TOKEN_EOF",
	    "TOKEN_LEFT_ROUND_BRACKET",
	    "TOKEN_RIGHT_ROUND_BRACKET",
	    "TOKEN_LEFT_SQUARE_BRACKET",
	    "TOKEN_RIGHT_SQUARE_BRACKET",
	    "TOKEN_LEFT_CURLY_BRACKET",
	    "TOKEN_RIGHT_CURLY_BRACKET",
	    "TOKEN_COLON",
	    "TOKEN_DOUBLE_COLON",
	    "TOKEN_COMMA",
	    "TOKEN_EQUAL",
	    "TOKEN_NOT_EQUAL",
	    "TOKEN_LESS_THAN",
	    "TOKEN_GREATER_THAN",
	    "TOKEN_LESS_EQUAL",
	    "TOKEN_GREATER_EQUAL",
	    "TOKEN_RIGHT_ARROW",
	    "TOKEN_DOT_DOT",
	    "TOKEN_DOT",
	    "TOKEN_HASH",
	    "TOKEN_SEMICOLON",
	    "TOKEN_BAR",
	    "TOKEN_AT",
	    "TOKEN_PLUS",
	    "TOKEN_MINUS",
	    "TOKEN_MULTIPLY",
	    "TOKEN_DIVIDE",
	    "TOKEN_NAME",
	    "TOKEN_NUMBER",
	    "TOKEN_STRING",
	    "TOKEN_LITERAL_CONTEXT",
	    "TOKEN_LITERAL_INV",
	    "TOKEN_LITERAL_PRE",
	    "TOKEN_LITERAL_POST",
	    "TOKEN_LITERAL_DEF",
	    "TOKEN_LITERAL_LET",
	    "TOKEN_LITERAL_IN",
	    "TOKEN_LITERAL_ENUM",
	    "TOKEN_LITERAL_AND",
	    "TOKEN_LITERAL_OR",
	    "TOKEN_LITERAL_XOR",
	    "TOKEN_LITERAL_IMPLIES",
	    "TOKEN_LITERAL_SET",
	    "TOKEN_LITERAL_BAG",
	    "TOKEN_LITERAL_SEQUENCE",
	    "TOKEN_LITERAL_COLLECTION",
	    "TOKEN_LITERAL_NOT",
	    "TOKEN_LITERAL_IF",
	    "TOKEN_LITERAL_THEN",
	    "TOKEN_LITERAL_ELSE",
	    "TOKEN_LITERAL_ENDIF"
	};

	ost << tokens[token_kind];

	return ost;
    }

} // namespace OCL
