// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/macros.h"
#include "foundation/object.h"
#include "foundation/utility.h"

#include "toolkit/widget.h"
#include "toolkit/factory.h"
#include "toolkit/box.h"
#include "toolkit/palette.h"
#include "toolkit/canvas.h"

#include "diagram/editor.h"
#include "diagram/layer_perspective.h"
#include "diagram/palette_manager.h"

#include "uml/diagram/diagram.h"

#include "uml-ext/class_diagram_model.h"
#include "uml-ext/class_diagram_renderer.h"
#include "uml-ext/class_diagram_palette_builder.h"

// Python Interfaces

#include <Python.h>

#include "python/common/convert.h"
#include "python/foundation/object.h"
#include "python/toolkit/widget.h"
#include "python/diagram/module.h"

namespace {

    using namespace Python;

    class ProxyDiagram : public Object {
    private:
        Diagram::Editor * mEditor;
        Diagram::LayerPerspective * mLayer;
        Diagram::PaletteManager * mPaletteManager;
        Toolkit::Box * mBox;

    public:
        Toolkit::Widget * get_widget() { return mBox; }

        explicit ProxyDiagram(UML::Diagram * diagram) {
            Toolkit::Factory * factory = Toolkit::Factory::instance();

            Toolkit::Palette * palette = factory->create_palette();
            Toolkit::Canvas * canvas = factory->create_canvas(1025, 1025);

            mEditor = new Diagram::Editor(canvas);

            Graph::Model * graph_model = new UMLExt::ClassDiagramModel(diagram);
            Diagram::Renderer * renderer = new UMLExt::ClassDiagramRenderer();

            mLayer = new Diagram::LayerPerspective(graph_model, renderer);

            renderer->unref();
            graph_model->unref();

            mPaletteManager = new Diagram::PaletteManager(mEditor, palette);

            UMLExt::build_class_diagram_palette(palette);

            mBox = factory->create_box(Toolkit::OK_HORIZONTAL);
            mBox->attach_object(this); // unref this object on destroy

            mBox->add(palette, false);
            mBox->add(canvas, true);

            palette->unref();
            canvas->unref();

            //

            mEditor->add(mLayer);
            mEditor->set_active(true);
        }

        ~ProxyDiagram() {
            mPaletteManager->unref();
            mLayer->unref();
            mEditor->unref();
            mBox->unref();
        }
    };


    PyObject *
    class_diagram_window_reference_new(PyObject * self, PyObject * args) {
        MESSAGE("Create a new Diagram Window");
        try {
            PyObject * ref_diagram = 0;
            if (!PyArg_ParseTuple(args, "O", & ref_diagram))
                return NULL;

            // Convert the arguments
            Object * diagram = 0;
            if (! convert_to_object(diagram, ref_diagram))
                return NULL;

            // Create the window in which to put the diagram

            // Create the proxy to manage the diagram workings and attach to window
            ProxyDiagram * proxy = new ProxyDiagram(known_cast<UML::Diagram *>(diagram));

            PyObject * result = widget_reference_new(proxy->get_widget());

            return result;
        }
        catch (...) {
            PyErr_SetString(PyExc_AttributeError, "Unknown C++ exception");
            return NULL;
        }
    }

} // namespace {anonymous}

namespace Python {

    void diagram_initialise() {
        static struct PyMethodDef diagram_methods[] = {
            { "Class", class_diagram_window_reference_new, METH_VARARGS },
            { NULL, NULL }
        };

        TRACE(1, "Initialise Diagram interface...");
        Py_InitModule("Diagram", diagram_methods);
    }
}
