// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

// C++ Implementations

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"
#include "foundation/undo_manager.h"
#include "foundation/change_listener.h"

#include "reflective/class.h"
#include "reflective/context.h"
#include "reflective/changes.h"

// Python Interfaces

#include "python/foundation/object.h"
#include "python/reflective/class.h"
#include "python/reflective/context.h"
#include "python/reflective/changes.h"

#include "python/reflective/change_listener.h"

//================================================================================

namespace {
    using namespace Python;

  // FIXUP - proxy is installed for every python client,
  // it would be more efficient to install a single proxy
  // and send python change list to all listeners....


  class ProxyChangeListener : public ChangeListener {
    PyObject * mObject;

  public:
    virtual void changes_occurred(const list<ChangeEvent *> & changes) throw() {
      // Build a list of changes to propogate to python
      PyObject * ref_change_list = PyList_New(0);

      // Transfer the changes to the python representation
      for(list<ChangeEvent *>::const_iterator iter = changes.begin();
	  iter != changes.end(); ++iter) {
	PyObject * ref_change = change_reference_new(*iter);
	PyList_Append(ref_change_list, ref_change);
      }
      // FIXUP object handling leaks...

      // Issue the changes to python
      //TRACE(2, "Calling changes_occurred on " << (void *) mObject);
      PyObject * ref_result = PyObject_CallMethod(mObject, "changes_occurred", 
						  "(O)", ref_change_list);
      Py_DECREF(ref_change_list);
      if (ref_result == NULL) {
	ERROR("Error returned by 'changes_occurred'");
	if (PyErr_Occurred())
	  PyErr_Print();
      }
      else {
	Py_DECREF(ref_result);
      }
    }

    PyObject * get() { return mObject; }
    ProxyChangeListener(PyObject * object) : mObject(object) { 
      Py_INCREF(mObject);
    }
    virtual ~ProxyChangeListener() { 
      Py_DECREF(mObject); 
    }
  };

} // namespace {anonymous}

//================================================================================

namespace Python {

    PyObject *
    set_change_listener(PyObject * self, PyObject * args) {
        try {
            // Extract the arguments
            PyObject * object = 0;
            if (!PyArg_ParseTuple(args, "O:set_change_listener", & object))
                return NULL;
            // Error checking?

            // Create the change listener proxy
            TRACE(1, "Adding a proxy change listener");
            ProxyChangeListener * proxy = new ProxyChangeListener(object);
            UndoManager * undo_manager = UndoManager::instance();
            undo_manager->add_change_listener(proxy);
        }
        catch (Exception & e) {
            string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
        // Return None
        return Py_BuildValue("");
    }

} // namespace Python
