// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/change_listener.h"
#include "foundation/undo_manager.h"

#include "uml/foundation/extension_mechanisms/tagged_value.h"

#include "reflective/changes.h"
#include "reflective/context.h"
#include "reflective/attribute.h"
#include "reflective/iterator.h"

// Python interface

#include "python/common/convert.h"
#include "python/reflective/class.h"
#include "python/reflective/context.h"
#include "python/reflective/changes.h"
#include "python/foundation/object.h"


//================================================================================


namespace {
  using namespace Python;

  // The following functions are defined by a table for the ModelElement python type

  // Exception objects used by this module
  //static PyObject * AttributeError;
  //static PyObject * TypeError;


  // --------------------------------------------------------------------------------

  PyObject *
  wrapped_change_get_type(PyObject * self, PyObject * args) {

    if (!PyArg_ParseTuple(args, ":Change.get_type"))
      return NULL;
    try {
      assert(PyObject_TypeCheck(self, & reflectiveChangeType));
      ChangeReference * rp = (ChangeReference *) self;
      const ChangeEvent * change = rp->get();

      const char * type = "unknown";
      if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	type = "AttributeChanged";
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	type = "RoleChanged";
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	type = "RoleAdded";
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	type = "RoleRemoved";
      }
      else {
	WARNING("Unknown change type");
      }
      return PyString_FromString(type);
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_change_get_object(PyObject * self, PyObject * args) {
    if (!PyArg_ParseTuple(args, ":Change.get_object"))
      return NULL;
    try {
      REF::Class * object = 0;
      assert(PyObject_TypeCheck(self, & reflectiveChangeType));
      ChangeReference * rp = (ChangeReference *) self;
      ChangeEvent * change = rp->get();
      if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	REF::ChangeAttributeChanged * attr_change = 
	  known_cast<REF::ChangeAttributeChanged *>(change);
	object = attr_change->get_object();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	REF::ChangeRoleChanged * role_change = 
	  known_cast<REF::ChangeRoleChanged *>(change);
	object = role_change->get_object();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	REF::ChangeRoleAdded * role_added = 
	  known_cast<REF::ChangeRoleAdded *>(change);
	object = role_added->get_object();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	REF::ChangeRoleRemoved * role_removed = 
	  known_cast<REF::ChangeRoleRemoved *>(change);
	object = role_removed->get_object();
      }
      else {
	ERROR("Unknown change type");
	PyErr_SetString(PyExc_TypeError, "Unknown change type encountered");
	return NULL;
      }
      return class_reference_new(object);
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_change_get_name(PyObject * self, PyObject * args) {
    if (!PyArg_ParseTuple(args, ":Change.get_name"))
      return NULL;
    try {
      const char * name = 0;
      assert(PyObject_TypeCheck(self, & reflectiveChangeType));
      ChangeReference * rp = (ChangeReference *) self;
      ChangeEvent * change = rp->get();
      if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	REF::ChangeAttributeChanged * attr_change = 
	  known_cast<REF::ChangeAttributeChanged *>(change);
	name = attr_change->get_attribute_name().c_str();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	REF::ChangeRoleChanged * role_change = 
	  known_cast<REF::ChangeRoleChanged *>(change);
	name = role_change->get_role_name().c_str();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	REF::ChangeRoleAdded * role_added = 
	  known_cast<REF::ChangeRoleAdded *>(change);
	name = role_added->get_role_name().c_str();
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	REF::ChangeRoleRemoved * role_removed = 
	  known_cast<REF::ChangeRoleRemoved *>(change);
	name = role_removed->get_role_name().c_str();
      }
      else {
	ERROR("Unknown change type");
	PyErr_SetString(PyExc_TypeError, "Unknown change type encountered");
	return NULL;
      }
      return PyString_FromString(name);
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_change_get_value(PyObject * self, PyObject * args) {
    if (!PyArg_ParseTuple(args, ":Change.get_value"))
      return NULL;
    try {
      assert(PyObject_TypeCheck(self, & reflectiveChangeType));
      ChangeReference * rp = (ChangeReference *) self;
      ChangeEvent * change = rp->get();
      if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	REF::ChangeAttributeChanged * attr_change = 
	  known_cast<REF::ChangeAttributeChanged *>(change);
	
	// Return the attribute string
	string value;
	if (attr_change->get_new_value())
	  attr_change->get_new_value()->ref_get_value(value);
	return PyString_FromString(value.c_str());
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	REF::ChangeRoleChanged * role_changed = 
	  known_cast<REF::ChangeRoleChanged *>(change);

	// Return the new value of the role
	return class_reference_new(role_changed->get_new_value());
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	REF::ChangeRoleAdded * role_added = 
	  known_cast<REF::ChangeRoleAdded *>(change);

	// Return the new value of the role
	return class_reference_new(role_added->get_value());
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	REF::ChangeRoleRemoved * role_removed = 
	  known_cast<REF::ChangeRoleRemoved *>(change);

	// Return the new value of the role
	return class_reference_new(role_removed->get_value());
      }

      ERROR("Unknown change type");
      PyErr_SetString(PyExc_TypeError, "Unknown change type encountered");
      return NULL;
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }


  PyObject *
  wrapped_change_get_old_value(PyObject * self, PyObject * args) {
    if (!PyArg_ParseTuple(args, ":Change.get_old_value"))
      return NULL;
    try {
      assert(PyObject_TypeCheck(self, & reflectiveChangeType));
      ChangeReference * rp = (ChangeReference *) self;
      ChangeEvent * change = rp->get();
      if(typeid(*change) == typeid(REF::ChangeAttributeChanged)) {
	REF::ChangeAttributeChanged * attr_change = 
	  known_cast<REF::ChangeAttributeChanged *>(change);
	
	// Return the attribute string
	string value;
	attr_change->get_old_value()->ref_get_value(value);
	return PyString_FromString(value.c_str());
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleChanged)) {
	REF::ChangeRoleChanged * role_change = 
	  known_cast<REF::ChangeRoleChanged *>(change);

	// Return the new value of the role
	return class_reference_new(role_change->get_old_value());
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleAdded)) {
	//REF::ChangeRoleAdded * role_added = 
	//known_cast<REF::ChangeRoleAdded *>(change);

	PyErr_SetString(PyExc_TypeError, "No old value for role added change");
	return NULL;
      }
      else if(typeid(*change) == typeid(REF::ChangeRoleRemoved)) {
	REF::ChangeRoleRemoved * role_removed = 
	  known_cast<REF::ChangeRoleRemoved *>(change);

	// Return the old value of the role
	return class_reference_new(role_removed->get_value());
      }

      ERROR("Unknown change type");
      PyErr_SetString(PyExc_TypeError, "Unknown change type encountered");
      return NULL;
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }



  //--------------------------------------------------------------------------------



  struct PyMethodDef change_methods[ ] = {
    { "get_type",           wrapped_change_get_type,	  METH_VARARGS },
    { "get_object",         wrapped_change_get_object,    METH_VARARGS },
    { "get_name",           wrapped_change_get_name,	  METH_VARARGS },
    { "get_attr_name",      wrapped_change_get_name,	  METH_VARARGS },
    { "get_role_name",      wrapped_change_get_name,	  METH_VARARGS },
    { "get_attribute_name", wrapped_change_get_name,	  METH_VARARGS },
    { "get_value",          wrapped_change_get_value,     METH_VARARGS },
    { "get_new_value",      wrapped_change_get_value,     METH_VARARGS },
    { "get_old_value",      wrapped_change_get_old_value, METH_VARARGS },
//    { "get_index",          wrapped_change_get_index,     METH_VARARGS },
//    { "is_aggregate",       wrapped_change_is_aggregate,  METH_VARARGS },
    { NULL, NULL }
  };


  //--------------------------------------------------------------------------------


  PyObject *
  change_getattr(PyObject * self, char * name) {
    return Py_FindMethod(change_methods, self, name);
  }


  int
  change_setattr(PyObject * self, char * name, PyObject * value) {
    PyErr_SetString(PyExc_NameError, "Changes have no settable attributes");
    return -1;
  }


  PyObject *
  change_repr(PyObject * self) {

    assert(PyObject_TypeCheck(self, & reflectiveChangeType));
    ChangeReference * rp = (ChangeReference *) self;
    ChangeEvent * change_event = rp->get();
    string result;

    REF::ChangeAttributeChanged * ac = dynamic_cast<REF::ChangeAttributeChanged *>(change_event);
    if(ac) {
      //const Attribute * old_value = ac->get_old_value();
      //const Attribute * new_value = ac->get_new_value();

      result = "Attribute Changed";
    }

    REF::ChangeRoleChanged * rc = dynamic_cast<REF::ChangeRoleChanged *>(change_event);
    if(rc) {
      result = "Role Changed";
    }

    REF::ChangeRoleAdded * ra = dynamic_cast<REF::ChangeRoleAdded *>(change_event);
    if(ra) {
      result = "Role Added";
    }

    REF::ChangeRoleRemoved * rr = dynamic_cast<REF::ChangeRoleRemoved *>(change_event);
    if(rr) {
      result = "Role Removed";
    }
    return PyString_FromString(result.c_str());
  }

  void
  change_dealloc(PyObject * self) {
    // Release our ownership of the object
    if (PyObject_TypeCheck(self, & reflectiveChangeType)) {
      ChangeReference * rp = (ChangeReference *) self;
      if (rp->mChange)
	rp->mChange->unref();
    }
    PyMem_DEL(self);
  }

} // namespace

//================================================================================


namespace Python {

  PyTypeObject reflectiveChangeType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "Reflective.Change",			// tp_name	: Type Name
    sizeof(ChangeReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0

    // Standard methods
    (destructor) change_dealloc,		// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) change_getattr,		// tp_getattr	: x.attr
    (setattrfunc) change_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) change_repr,			// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto

    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc

  };

  //--------------------------------------------------------------------------------


  PyObject *
  change_reference_new(ChangeEvent * change) {
    ChangeReference * self = PyObject_NEW(ChangeReference, & reflectiveChangeType);
    if (! self)
      return NULL;

    change->ref();		 // take joint ownership of the object
    self->mChange = change;
    return (PyObject *) self;
  }


}
