// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/change_listener.h"
#include "foundation/undo_manager.h"

#include "uml/foundation/extension_mechanisms/tagged_value.h"

#include "reflective/class.h"
#include "reflective/context.h"
#include "reflective/attribute.h"
#include "reflective/iterator.h"

// Python interface

#include "python/common/convert.h"
#include "python/reflective/class.h"
#include "python/reflective/context.h"
#include "python/foundation/object.h"

#include <map>
#include <sstream>
using namespace std;

//================================================================================


namespace {

    using namespace Python;

    // The following functions are defined by a table for the ModelElement python type

  // Exception objects used by this module
  //static PyObject * AttributeError;
  //static PyObject * TypeError;

  
  typedef map<Object *, PyObject *> ReferenceMap;
  ReferenceMap reference_table;

  // --------------------------------------------------------------------------------

  PyObject *
  wrapped_class_get_type(PyObject * self, PyObject * args) {

    if (!PyArg_ParseTuple(args, ":Class.get_type"))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the type as a string
      const string & type = rp->mClass->ref_get_type();
      return PyString_FromString(type.c_str());
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_class_is_a_type_of(PyObject * self, PyObject * args) {

    const char * type = 0;
    if (!PyArg_ParseTuple(args, "s:Class.is_a_type_of", & type))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Test if we are a class or supper class of given type
      int result = rp->mClass->ref_is_a_type_of(type);
      return Py_BuildValue("i", result);
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_class_get_attribute(PyObject * self, PyObject * args) {

    const char * name = 0;
    if (!PyArg_ParseTuple(args, "s", & name))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the attribute as a string value
      const REF::Attribute * ap = rp->mClass->ref_get_attribute(string(name));
      if (ap) {
	// Get the attribute string value	
	string s;
	ap->ref_get_value(s);
	return PyString_FromString(s.c_str());
      }
      return PyString_FromString("");
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_class_set_attribute(PyObject * self, PyObject * args) {

    const char * name = 0;
    const char * value = 0;
    if (!PyArg_ParseTuple(args, "ss", & name, & value))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the attribute as a string value
      REF::Attribute * attr = rp->mClass->ref_create_attribute(name);
      attr->ref_set_value(value);
      rp->mClass->ref_set_attribute(name, attr);
      attr->unref();
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_class_get_role(PyObject * self, PyObject * args) {

    const char * name = 0;
    if (!PyArg_ParseTuple(args, "s", & name))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    REF::Class * role = 0;
    try {
      // Get the role, returns NULL if not set
      role = rp->mClass->ref_get_role(string(name));
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }

    // Return an object of type ClassReference
    if (role) {
      // Return the role argument as a new object
      return class_reference_new(role);
    }
    else {
      // Return None
      return Py_BuildValue("");
    }
  }


  PyObject *
  wrapped_class_get_role_list(PyObject * self, PyObject * args) {

    const char * name = 0;
    if (!PyArg_ParseTuple(args, "s", & name))
      return NULL;

    PyObject * list = PyList_New(0);
    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the role, returns NULL if not set
      REF::Iterator * iter = rp->mClass->ref_get_role_iterator(string(name));

      while (iter->is_valid()) {
	PyList_Append(list, class_reference_new(iter->get()));
	iter->increment();
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }

    // Return the list
    return list;
  }

  PyObject *
  wrapped_class_set_role(PyObject * self, PyObject * args) {
    //virtual void ref_set_role(const string & role_name, Class * role) = 0;
    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Extract the arguments
      const char * role_name = 0;
      PyObject * ref_role = 0;
      if (!PyArg_ParseTuple(args, "sO:Class.set_role", role_name, & ref_role))
	return NULL;

      // Convert the arguments to C++ type
      REF::Class * role = 0;
      if (!convert_to_class(role, ref_role))
	return NULL;

      // Finally call the method
      rp->mClass->ref_set_role(role_name, role);
      return Py_BuildValue("");
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  PyObject *
  wrapped_class_add_role(PyObject * self, PyObject * args) {
    //virtual void ref_add_role(const string & role_name, Class * role,
    //                          const Iterator * iter = 0) = 0;
    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Extract the arguments
      const char * role_name = 0;
      PyObject * ref_role = 0;
      // FIXUP don't know how to deal with the iterator
      if (!PyArg_ParseTuple(args, "sO:Class.set_role", & role_name, & ref_role))
	return NULL;

      // Convert the arguments to C++ type
      REF::Class * role = 0;
      if (!convert_to_class(role, ref_role))
	return NULL;

      // Finally call the method
      rp->mClass->ref_add_role(role_name, role);
      return Py_BuildValue("");
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_class_get_context(PyObject * self, PyObject * args) {
    //Context * ref_get_context() { return mContext; }
    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Class.get_context"))
	return NULL;

      // Finally call the method
      REF::Context * context = rp->mClass->ref_get_context();
      //TRACE(1, "Got context " << context);
      PyObject * result = Py_BuildValue("O", context_reference_new(context));
      context->unref();
      return result;
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }

  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_class_has_tagged_value(PyObject * self, PyObject * args) {

    const char * name = 0;
    if (!PyArg_ParseTuple(args, "s", & name))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    int has_value = 0;
    try {
      // Get the attribute as a string value
      REF::Iterator * iter = rp->mClass->ref_get_role_iterator("tagged_value");
      while (iter->is_valid()) {
	if (const REF::Attribute * ap = iter->get()->ref_get_attribute("name")) {
	  string attr_name;
	  ap->ref_get_value(attr_name);
	  if (attr_name == name) {
	    if (iter->get()->ref_get_attribute(string("data_value")))
	      has_value = 1;
	    break;
	  }
	}
	iter->increment();
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    return Py_BuildValue("i", & has_value);
  }


  PyObject *
  wrapped_class_get_tagged_value(PyObject * self, PyObject * args) {

    const char * name = 0;
    if (!PyArg_ParseTuple(args, "s", & name))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the attribute as a string value
      REF::Iterator * iter = rp->mClass->ref_get_role_iterator("tagged_value");
      while (iter->is_valid()) {
	if (const REF::Attribute * nap = iter->get()->ref_get_attribute("name")) {
	  string attr_name;
	  nap->ref_get_value(attr_name);
	  if (attr_name == name) {
	    if (const REF::Attribute * vap = iter->get()->ref_get_attribute("data_value")) {
	      string attr_value;
	      vap->ref_get_value(attr_value);
	      return PyString_FromString(attr_value.c_str());
	    }
	  }
	}
	iter->increment();
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    return PyString_FromString("");
  }


  PyObject *
  wrapped_class_set_tagged_value(PyObject * self, PyObject * args) {

    const char * name = 0;
    const char * value = 0;
    if (!PyArg_ParseTuple(args, "ss", & name, & value))
      return NULL;

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    bool value_set = false;
    try {
      // Get the attribute as a string value
      REF::Iterator * iter = rp->mClass->ref_get_role_iterator("tagged_value");
      while (iter->is_valid()) {
	REF::Class * tag = iter->get();
	if (const REF::Attribute * attr_name = tag->ref_get_attribute("name")) {
	  string attr_name_string;
	  attr_name->ref_get_value(attr_name_string);
	  if (attr_name_string == name) {
	    REF::Attribute * attr_value = tag->ref_create_attribute("data_value");
	    attr_value->ref_set_value(value);
	    tag->ref_set_attribute("data_value", attr_value);
	    attr_value->unref();
	    value_set = true;
	    break;
	  }
	}
	iter->increment();
      }

      // If we did not update an existing tagged value, add a new one
      if (! value_set) {
	MESSAGE("Creating a new tagged value for " << name << " = " << value);
	UML::TaggedValue * new_tagged_value = new UML::TaggedValue();
	REF::Attribute * attr_name  = new_tagged_value->ref_create_attribute("name");
	REF::Attribute * attr_value = new_tagged_value->ref_create_attribute("data_value");
	attr_name->ref_set_value(name);
	attr_value->ref_set_value(value);
	new_tagged_value->ref_set_attribute("name", attr_name);
	new_tagged_value->ref_set_attribute("data_value", attr_value);
	rp->mClass->ref_add_role("tagged_value", new_tagged_value);
	attr_name->unref();
	attr_value->unref();
	new_tagged_value->unref();
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    return Py_BuildValue("");
  }

  //--------------------------------------------------------------------------------

  struct PyMethodDef class_methods[ ] = {
    { "get_type",         wrapped_class_get_type,	  METH_VARARGS },
    { "is_a_type_of",     wrapped_class_is_a_type_of,	  METH_VARARGS },
    { "get_context",      wrapped_class_get_context,	  METH_VARARGS },

    { "get_attribute",    wrapped_class_get_attribute,    METH_VARARGS },
    { "set_attribute",    wrapped_class_set_attribute,    METH_VARARGS },

    { "get_role",         wrapped_class_get_role,         METH_VARARGS },
    { "get_role_list",    wrapped_class_get_role_list,    METH_VARARGS },
    { "set_role",         wrapped_class_set_role,         METH_VARARGS },
    { "add_role",         wrapped_class_add_role,         METH_VARARGS },

    { "has_tagged_value", wrapped_class_has_tagged_value, METH_VARARGS },
    { "get_tagged_value", wrapped_class_get_tagged_value, METH_VARARGS },
    { "set_tagged_value", wrapped_class_set_tagged_value, METH_VARARGS },

    { NULL, NULL }
  };


  //--------------------------------------------------------------------------------


  PyObject *
  class_getattr(PyObject * self, char * name) {
    //MESSAGE("Get attribute " << name);

    // Create a few shortcuts for the basic attributes
    try {
      if (strcmp(name, "type") == 0) {

	// Get the type as a string
	assert(PyObject_TypeCheck(self, & reflectiveClassType));
	ClassReference * rp = (ClassReference *) self;
	const string & type = rp->mClass->ref_get_type();
	return PyString_FromString(type.c_str());
      }
      if (strcmp(name, "name") == 0) {
	assert(PyObject_TypeCheck(self, & reflectiveClassType));
	ClassReference * rp = (ClassReference *) self;
	const REF::Attribute * ap = rp->mClass->ref_get_attribute("name");
	string attr_value;
	if (ap)
	  ap->ref_get_value(attr_value);
	return PyString_FromString(attr_value.c_str());
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }

    // Look for a method instead
    return Py_FindMethod(class_methods, self, name);
  }


  int
  class_setattr(PyObject * self, char * name, PyObject * value) {
    //MESSAGE("Set attribute " << name);

    // Create a few shortcuts for the basic attributes
    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // type is immutable, right?
      if (strcmp(name, "name") == 0) {
	// Get the attribute as a string value
	REF::Attribute * attr = rp->mClass->ref_create_attribute(name);
	attr->ref_set_value(PyString_AsString(value));
	rp->mClass->ref_set_attribute(name, attr);
	attr->unref();
	return 0;
      }
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_AttributeError, error.c_str());
      return -1;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return -1;      
    }
    // Unknown attribute name
    string error = string(name) + " not an attribute of element " + rp->mClass->ref_get_type();
    PyErr_SetString(PyExc_NameError, error.c_str());
    return -1;
  }

/*
  PyObject *
  class_repr(PyObject * self) {

    assert(PyObject_TypeCheck(self, & reflectiveClassType));
    ClassReference * rp = (ClassReference *) self;
    try {
      // Get the type as a string
      const string & type = rp->mClass->ref_get_type();
      const REF::Attribute * ap = rp->mClass->ref_get_attribute(string("name"));
      string name;
      if (ap) {
	// Get the attribute string value
	ap->ref_get_value(name);
      }
      return PyString_FromString(form("ClassReference of type %s, name '%s'", 
				      type.c_str(), name.c_str()));
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }
*/
    PyObject *
    class_repr(PyObject * self) {

        assert(PyObject_TypeCheck(self, & reflectiveClassType));
        ClassReference * rp = (ClassReference *) self;
        try {
            // Get the type as a string
            const string & type = rp->mClass->ref_get_type();
            const REF::Attribute * ap = rp->mClass->ref_get_attribute(string("name"));
            string name;
            if (ap) {
                // Get the attribute string value
                ap->ref_get_value(name);
            }

            ostringstream ost;
            ost << "ClassReference of type " << type << ", name '" << name << "'";

            return PyString_FromString(ost.str().c_str());
        }
        catch (Exception & e) {
            string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
            PyErr_SetString(PyExc_RuntimeError, error.c_str());
            return NULL;
        }
        catch (...) {
            PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
            return NULL;      
        }
    }
  

  void
  class_dealloc(PyObject * self) {
    // Release our ownership of the object
    if (PyObject_TypeCheck(self, & reflectiveClassType)) {
      ClassReference * rp = (ClassReference *) self;
      if (rp->mClass) {
	reference_table.erase(rp->mClass);
	rp->mClass->unref();
      }
    }
  }

} // namespace {anonymous}

//================================================================================


namespace Python {

  PyTypeObject reflectiveClassType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "Reflective.Class",				// tp_name	: Type Name
    sizeof(ClassReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
    
    // Standard methods
    (destructor) class_dealloc,			// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) class_getattr,		// tp_getattr	: x.attr
    (setattrfunc) class_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,				// tp_compare	: x > y
    (reprfunc) class_repr,			// tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto

    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc

  };

  //--------------------------------------------------------------------------------


  PyObject *
  class_reference_new(REF::Class * klass) {

    // Check for NULL
    if (klass == 0)
      return Py_BuildValue("");

    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(klass);
    if (iter == reference_table.end()) {

      // Create a new reference
      ClassReference * self = PyObject_NEW(ClassReference, & reflectiveClassType);
      if (! self)
	return NULL;
    
      klass->ref();		 // take joint ownership of the object
      self->mClass = klass;
      result = (PyObject *) self;
      reference_table[klass] = result;
    }
    else {
      //TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }


  PyObject *
  class_reference_new(PyObject * self, PyObject * args) {
    MESSAGE("Create a new ClassReference.Class");
    try {
      PyObject * ref_object = 0;
      if (!PyArg_ParseTuple(args, "O:Class", & ref_object))
	return NULL;

      REF::Class * klass = 0;
      if (!convert_to_class(klass, ref_object))
	return NULL;

      PyObject * result = class_reference_new(klass);
      klass->unref();  // careful of the order here
      return result;
    }
    catch (Exception & e) {
      string error = "Reflective Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_TypeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
  }
    
} // namespace Python
