// Dachshund
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <map>

#include "foundation/object.h"
#include "foundation/macros.h"
#include "foundation/exception.h"
#include "foundation/undo_manager.h"
#include "foundation/change_listener.h"

#include "toolkit/widget.h"
#include "toolkit/tree.h"
#include "toolkit/tree_maintainer.h"
#include "toolkit/mouse_listener.h"
#include "toolkit/mouse_event.h"

#include "drawing/point.h"
#include "drawing/image.h"

#include "python/foundation/object.h"
#include "python/toolkit/widget.h"


//================================================================================

namespace {
  using namespace Python;


  typedef map<Object *, PyObject *> ReferenceMap;
  ReferenceMap reference_table;


  //================================================================================

/*
  PyObject *
  wrapped_widget_begin_transaction(PyObject * self, PyObject * args) {
    //virtual void begin_transaction()
    assert(PyObject_TypeCheck(self, & toolkitWidgetType));
    WidgetReference * wrp = (WidgetReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Widget.begin_transaction"))
	return NULL;

      // Finally call the method
      wrp->mWidget->begin_transaction();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_widget_end_transaction(PyObject * self, PyObject * args) {
    //virtual void end_transaction()
    assert(PyObject_TypeCheck(self, & toolkitWidgetType));
    WidgetReference * wrp = (WidgetReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Widget.end_transaction"))
	return NULL;

      // Finally call the method
      wrp->mWidget->end_transaction();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }
*/
  //--------------------------------------------------------------------------------

  PyObject *
  wrapped_widget_show(PyObject * self, PyObject * args) {
    //virtual void show()
    assert(PyObject_TypeCheck(self, & toolkitWidgetType));
    WidgetReference * wrp = (WidgetReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Widget.show"))
	return NULL;

      // Finally call the method
      wrp->mWidget->show();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }

  PyObject *
  wrapped_widget_hide(PyObject * self, PyObject * args) {
    //virtual void hide()
    assert(PyObject_TypeCheck(self, & toolkitWidgetType));
    WidgetReference * wrp = (WidgetReference *) self;
    try {
      // Extract the arguments
      if (!PyArg_ParseTuple(args, ":Widget.hide"))
	return NULL;

      // Finally call the method
      wrp->mWidget->hide();
    }
    catch (Exception & e) {
      string error = "Toolkit Exception " + e.get_name() + ": " + e.get_details();
      PyErr_SetString(PyExc_RuntimeError, error.c_str());
      return NULL;
    }
    catch (...) {
      PyErr_SetString(PyExc_RuntimeError, "Undefined C++ exception caught");
      return NULL;      
    }
    // Return None
    return Py_BuildValue("");
  }


  //--------------------------------------------------------------------------------

  struct PyMethodDef toolkit_widget_methods[ ] = {
    { "show",			wrapped_widget_show,	      METH_VARARGS },
    { "hide",			wrapped_widget_hide,          METH_VARARGS },
/*    
    { "begin_transaction",	wrapped_widget_begin_transaction, METH_VARARGS },
    { "end_transaction",	wrapped_widget_end_transaction, METH_VARARGS },
*/
    { NULL, NULL }
  };


  PyObject *
  widget_getattr(PyObject * self, char * name) {

    // Look for a method
    return Py_FindMethod(toolkit_widget_methods, self, name);
  }


  int
  widget_setattr(PyObject * self, char * name, PyObject * value) {
    PyErr_SetString(PyExc_NameError, "No attributes to be set");
    return -1;
  }


  void
  widget_dealloc(PyObject * self) {
    if (PyObject_TypeCheck(self, & toolkitWidgetType)) {
      WidgetReference * rp = (WidgetReference *) self;
      if (rp->mWidget) {
	reference_table.erase(rp->mWidget);
	rp->mWidget->unref();
      }
    }
    PyMem_DEL(self);
  }


}


namespace Python {


  //--------------------------------------------------------------------------------

  PyTypeObject toolkitWidgetType = {
    PyObject_HEAD_INIT(&PyType_Type)		// Initialisation
    0,						// ob_size	: Usually 0
    "toolkit.Widget",				// tp_name	: Type Name
    sizeof(WidgetReference),			// tp_basicsize : Object size
    0,						// tp_itemsize  : Usually 0
    
    // Standard methods
    (destructor) widget_dealloc,		// tp_dealloc	: refcount = 0
    (printfunc) 0,				// tp_print	: print x
    (getattrfunc) widget_getattr,		// tp_getattr	: x.attr
    (setattrfunc) widget_setattr,		// tp_setattr	: a.attr = v
    (cmpfunc) 0,			        // tp_compare	: x > y
    (reprfunc) 0,			        // tp_repr	: repr(x)

    // Type categories
    0,						// tp_as_number
    0,						// tp_as_sequence
    0,						// tp_as_mapping

    (hashfunc) 0,				// tp_hash
    (ternaryfunc) 0,				// tp_call
    (reprfunc) 0,				// tp_str
    (getattrofunc) 0,				// tp_getatto
    (setattrofunc) 0,				// tp_setatto

    0,						// tp_as_buffer
    0,						// tp_flags
    0,						// tp_doc
  };


  PyObject *
  widget_reference_new(PyObject * self, PyObject * args) {
    TRACE(1, "Create a new Toolkit.Widget");
    PyErr_SetString(PyExc_TypeError, "don't know how to construct");
    return NULL;
  }

  PyObject *
  widget_reference_new(Toolkit::Widget * widget) {

    PyObject * result = 0;
    ReferenceMap::iterator iter = reference_table.find(widget);
    if (iter == reference_table.end()) {

      // Create a new reference
      WidgetReference * self = PyObject_NEW(WidgetReference, & toolkitWidgetType);
      if (! self)
	return NULL;
    
      widget->ref();		 // take joint ownership of the object
      self->mWidget = widget;
      result = (PyObject *) self;
      reference_table[widget] = result;
    }
    else {
      TRACE(2, "Reusing an existing reference");
      // Use an existing reference
      result = iter->second;
      Py_INCREF(result);
    }
    return result;
  }


}

