// Dachshund
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef REFLECTIVE_CHANGES_H
#define REFLECTIVE_CHANGES_H

#include "foundation/change_event.h"
#include "foundation/utility.h"

#include "class.h"
#include "attribute.h"

namespace REF {


    //
    // This means that an attribute of a class has change its value.
    //
    // For example, a class might have an attribute called "name"
    // of type String and its value might have changed from "Observer"
    // to "Listener".
    //

    class ChangeAttributeChanged : public ChangeEvent {
    public:
	ChangeAttributeChanged(Class * object,
			       const string & attribute_name,
			       Attribute * new_value,
			       Attribute * old_value) : ChangeEvent(),
	    mObject(object),
	    mAttributeName(attribute_name),
	    mNewValue(new_value),
	    mOldValue(old_value) {
	    ASSERT(mObject);
	    //ASSERT(mNewValue || mOldValue);

	    mObject->ref();

	    if(mNewValue) {
		mNewValue->ref();
	    }

	    if(mOldValue) {
		mOldValue->ref();
	    }
	}

	Class * get_object() {
	    return mObject;
	}

	const string & get_attribute_name() {
	    return mAttributeName;
	}

	Attribute * get_new_value() {
	    return mNewValue;
	}

	Attribute * get_old_value() {
	    return mOldValue;
	}

	// helpers

	template<class T>
	T * get_object_cast() {
	    return known_cast<T *>(mObject);
	}

	template<class T>
	T * get_new_value_cast() const {
	    return known_cast<T *>(mNewValue);
	}

	template<class T>
	T * get_old_value_cast() const {
	    return known_cast<T *>(mOldValue);
	}

    protected:
	~ChangeAttributeChanged() {
	    if(mOldValue) {
		mOldValue->unref();
	    }

	    if(mNewValue) {
		mNewValue->unref();
	    }

	    mObject->unref();
	}

    private:
	Class * mObject;
	string mAttributeName;
	Attribute * mNewValue;
	Attribute * mOldValue;
    };

    inline ChangeAttributeChanged * change_attribute_changed(Class * object,
							     const string & attribute_name,
							     Attribute * new_value,
							     Attribute * old_value) {
	return new ChangeAttributeChanged(object, attribute_name, new_value, old_value);
    }

    //
    // This means that a class that is associated with another class has either
    // gone from null to a new class, new class to null or one class to another
    //
    // For example, ModelElements belong in Namespaces, if we move a ModelElement
    // from one namespace to another then its "container" role has changed.
    //

    class ChangeRoleChanged : public ChangeEvent {
    public:
	ChangeRoleChanged(Class * object, const string & role_name,
			  Class * new_role, Class * old_role,
			  bool aggregate) : ChangeEvent(),
	    mObject(object),
	    mRoleName(role_name),
	    mNewValue(new_role),
	    mOldValue(old_role),
	    mAggregate(aggregate) {
	    ASSERT(mObject);
	    //ASSERT(mOldValue || mNewValue);

	    mObject->ref();

	    if(mNewValue) {
		mNewValue->ref();
	    }

	    if(mOldValue) {
		mOldValue->ref();
	    }
	}

	Class * get_object() {
	    return mObject;
	}

	const string & get_role_name() const {
	    return mRoleName;
	}

	Class *
	get_new_value() {
	    return mNewValue;
	}

	Class *
	get_old_value() {
	    return mOldValue;
	}

	bool is_aggregate() const { return mAggregate; }

	// helpers

	template<class T>
	T * get_object_cast() {
	    return known_cast<T *>(mObject);
	}

	template<class T>
	T * get_new_value_cast() {
	    return known_cast<T *>(mNewValue);
	}

	template<class T>
	T * get_old_value_cast() {
	    return known_cast<T *>(mOldValue);
	}

    protected:
	~ChangeRoleChanged() {
	    if(mOldValue) {
		mOldValue->unref();
	    }

	    if(mNewValue) {
		mNewValue->unref();
	    }

	    mObject->unref();
	}

    private:
	Class * mObject;
	string mRoleName;
	Class * mNewValue;
	Class * mOldValue;
	bool mAggregate;
    };

    inline ChangeRoleChanged * change_role_changed(Class * object, const string & role_name,
						   Class * new_role, Class * old_role,
						   bool aggregate) {
	return new ChangeRoleChanged(object, role_name, new_role, old_role, aggregate);
    }

    //
    // A new association has been added to a container of associations.
    //
    // For example, a new ModelElement has been added to a Package's existing
    // list of "contained_elements"
    //

    class ChangeRoleAdded : public ChangeEvent {
    public:
	ChangeRoleAdded(Class * object, const string & role_name,
			Class * role, bool aggregate, int index = -1) : ChangeEvent(),
	    mObject(object),
	    mRoleName(role_name),
	    mValue(role),
	    mAggregate(aggregate),
	    mIndex(index) {
	    ASSERT(mObject);
	    ASSERT(mValue);

	    mObject->ref();
	    mValue->ref();
	}

	Class * get_object() { return mObject; }

	const string & get_role_name() const {
	    return mRoleName;
	}

	Class * get_value() {
	    return mValue;
	}

	bool is_aggregate() const { return mAggregate; }

	int get_index() const {
	    return mIndex;
	}

	// helpers

	template<class T>
	T * get_object_cast() {
	    return known_cast<T *>(mObject);
	}

	template<class T>
	T * get_value_cast() {
	    return known_cast<T *>(mValue);
	}

    protected:
	~ChangeRoleAdded() {
	    mValue->unref();
	    mObject->unref();
	}

    private:
	Class * mObject;
	string mRoleName;
	Class * mValue;
	bool mAggregate;
	int mIndex;
    };

    inline ChangeRoleAdded * change_role_added(Class * object, const string & role_name,
					       Class * role, bool aggregate, int index = -1) {
	return new ChangeRoleAdded(object, role_name, role, aggregate, index);
    }

    //
    // An association has been removed from a container of associations
    //
    // For example, a ModelElement has been removed from a Package's list
    // of "contained_elements"
    //

    class ChangeRoleRemoved : public ChangeEvent {
    public:
	ChangeRoleRemoved(Class * object, const string & role_name,
			  Class * role, bool aggregate, int index = -1) : ChangeEvent(),
	    mObject(object),
	    mRoleName(role_name),
	    mValue(role),
	    mAggregate(aggregate),
	    mIndex(index) {
	    ASSERT(mObject);
	    ASSERT(mValue);

	    mObject->ref();
	    mValue->ref();
	}

	Class * get_object() {
	    return mObject;
	}

	Class * get_value() {
	    return mValue;
	}

	const string & get_role_name() const {
	    return mRoleName;
	}

	bool is_aggregate() const { return mAggregate; }

	int get_index() const {
	    return mIndex;
	}

	// helpers

	template<class T>
	T * get_object_cast() {
	    return known_cast<T *>(mObject);
	}

	template<class T>
	T * get_value_cast() {
	    return known_cast<T *>(mValue);
	}

    protected:
	~ChangeRoleRemoved() {
	    mValue->unref();
	    mObject->unref();
	}

    private:
	Class * mObject;
	string mRoleName;
	Class * mValue;
	bool mAggregate;
	int mIndex;
    };

    inline ChangeRoleRemoved * change_role_removed(Class * object, const string & role_name,
						   Class * role, bool aggregate, int index = -1) {
	return new ChangeRoleRemoved(object, role_name, role, aggregate, index);
    }

} // namespace REF

#endif // REFLECTIVE_CHANGES_H
