// Dachshund,
// Copyright (C) 2001 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "class.h"

#include "exceptions.h"

#include "foundation/macros.h"

namespace REF {

    Class::Class(const Common::UUID & uuid) : Object(), mContext(0), mUUID(uuid) {
    }

    // if a copy is made of the class then the context is cleared and a new UUID is created
    Class::Class(const Class & klass) : Object(klass), mContext(0), mUUID(Common::UUID::create()) {
    }

    Class::~Class() {
    }

    //
    // operations that traverse the hierarchy using the _impl operations below
    //

    bool
    Class::ref_is_a_type_of(const string & class_name) const {
	set<string> visited;
	return ref_is_a_type_of_impl(visited, class_name);
    }

    void
    Class::ref_delete() {
	begin_transaction();

	set<string> visited;
	ref_delete_impl(visited);

	end_transaction();
    }

    //

    void
    Class::ref_set_role(const string & role_name, Class * role) {
	RETURN_IF_FAIL(role);

	set<string> visited;

	if(!ref_set_role_impl(visited, role_name, role)) {
	    throw NotFound("Can't find role: " + role_name + " in class: " + ref_get_type());
	}
    }

    Class *
    Class::ref_get_role(const string & role_name) {
	set<string> visited;

	Class * role = 0;

	if(!ref_get_role_impl(visited, role, role_name)) {
	    throw NotFound("Can't find role: " + role_name + " in class: " + ref_get_type());
	}
	else {
	    // role can be null
	    return role;
	}
    }

    //

    void
    Class::ref_add_role(const string & role_name, Class * role, Class * position) {
	RETURN_IF_FAIL(role);

	set<string> visited;

	if(!ref_add_role_impl(visited, role_name, role, position)) {
	    throw NotFound("Can't find role: " + role_name + " in class: " + ref_get_type());
	}
    }

    void
    Class::ref_remove_role(const string & role_name, Class * role) {
	RETURN_IF_FAIL(role);

	set<string> visited;

	if(!ref_remove_role_impl(visited, role_name, role)) {
	    throw NotFound("Can't find role: " + role_name + " in class: " + ref_get_type());
	}
    }

    Iterator *
    Class::ref_get_role_iterator(const string & role_name) {
	set<string> visited;

	Iterator * iterator = 0;

	if(!ref_get_role_iterator_impl(visited, iterator, role_name)) {
	    throw NotFound("Can't find role: " + role_name + " in class: " + ref_get_type());
	}
	else {
	    ASSERT(iterator);
	    return iterator;
	}
    }

    Class *
    Class::ref_create_class(const string & role_name) const {
	set<string> visited;

	Class * klass = 0;

	if(!ref_create_class_impl(visited, klass, role_name)) {
	    throw NotFound("Can't find role: " + role_name +
			   " in class: " + ref_get_type());
	}
	else {
	    ASSERT(klass);
	    return klass;
	}
    }

    //

    const Attribute *
    Class::ref_get_attribute(const string & attribute_name) const {
	set<string> visited;

	const Attribute * attribute = 0;

	if(!ref_get_attribute_impl(visited, attribute, attribute_name)) {
	    throw NotFound("Can't find attribute: " + attribute_name +
			   " in class: " + ref_get_type());
	}
	else {
	    return attribute;
	}
    }

    void
    Class::ref_set_attribute(const string & attribute_name, const Attribute * attribute) {
	RETURN_IF_FAIL(attribute);

	set<string> visited;

	if(!ref_set_attribute_impl(visited, attribute_name, attribute)) {
	    throw NotFound("Can't find attribute: " + attribute_name +
			   " in class: " + ref_get_type());
	}
    }

    const vector<const Attribute *> &
    Class::ref_get_attribute_set(const string & attribute_name) const {
	set<string> visited;

	const vector<const Attribute *> * attribute_set = 0;

	if(!ref_get_attribute_set_impl(visited, attribute_set, attribute_name)) {
	    throw NotFound("Can't find attribute set: " + attribute_name +
			   " in class: " + ref_get_type());
	}
	else {
	    ASSERT(attribute_set);
	    return *attribute_set;
	}
    }

    void
    Class::ref_set_attribute_set(const string & attribute_name,
				 const vector<const Attribute *> & attribute_set) {
	set<string> visited;

	if(!ref_set_attribute_set_impl(visited, attribute_name, &attribute_set)) {
	    throw NotFound("Can't find attribute set: " + attribute_name +
			   " in class: " + ref_get_type());
	}
    }

    Attribute *
    Class::ref_create_attribute(const string & attribute_name) const {
	set<string> visited;

	Attribute * attribute = 0;

	if(!ref_create_attribute_impl(visited, attribute, attribute_name)) {
	    throw NotFound("Can't find attribute: " + attribute_name +
			   " in class: " + ref_get_type());
	}
	else {
	    ASSERT(attribute);
	    return attribute;
	}
    }

    //
    // internal functions
    //

    void
    Class::internal_ref_set_context(Context * context) {
	if(mContext != context) {
	    if(mContext) {
		internal_ref_uproot();
	    }

	    mContext = context;

	    set<string> visited;
	    ref_set_context_impl(visited, mContext);

	    if(mContext) {
		internal_ref_implant();
	    }
	}
    }

    void
    Class::internal_ref_implant() {
	set<string> visited;
	ref_implant_impl(visited);
    }

    void
    Class::internal_ref_uproot() {
	set<string> visited;
	ref_uproot_impl(visited);
    }

} // namespace REF
