// Dachshund,
// Copyright (C) 2000 David Bryant
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "stream.h"

#include "foundation/macros.h"

#include <exception>
#include <iterator>
#include <algorithm>
using namespace std;

namespace XML {

    Stream::Stream(ostream & ost, const string & doctype,
		   DTDKind dtd_kind, const string & dtd) : mOs(ost), mDocType(doctype) {
        mState = STATE_START;
	mHadRootElement = false;

	mIndentLevel = 0;
	mIndentWidth = DEFAULT_INDENT_WIDTH;

	ASSERT(ost);

	ost << "<?xml version=\"1.0\"?>" << endl;
	ost << "<!DOCTYPE " << doctype;

	if(dtd_kind == DTDK_SYSTEM) {
	    ost << " SYSTEM ";
	}
	else {
	    ASSERT(dtd_kind == DTDK_PUBLIC);

	    ost << " PUBLIC ";
	}

	ost << "\"" << dtd << "\">" << endl;
    }

    Stream::~Stream() {
	mOs << endl;

	if(!uncaught_exception()) {
	    ASSERT(mHadRootElement);
	    ASSERT(mStack.empty());
	}
    }

    void
    Stream::comment(const string & s) {
	mOs << endl << "<!-- " << s << " -->";
    }

    void
    Stream::begin_open(const string & name) {
	ASSERT(mState != STATE_OUT_OF_CDATA_TAG);

	if(mState == STATE_OUT_OF_EMPTY_TAG) {
	    mOs << ">";
	    mState = STATE_OUT_OF_NESTED_TAG;
	}

	if(!mHadRootElement) {
	    mHadRootElement = true;
	    ASSERT(mState == STATE_START);

	    if(name != mDocType) {
		WARNING("Name doesn't match doctype: " << name << " != " << mDocType);
	    }
	}
	else {
	    ASSERT(!mStack.empty());
	}

	mOs << endl;

	fill_n(ostream_iterator<char>(mOs), (mIndentLevel * mIndentWidth) / 8, '\t');
	fill_n(ostream_iterator<char>(mOs), (mIndentLevel * mIndentWidth) % 8, ' ');

	mOs << "<" << name;

	mStack.push(name);

	mState = STATE_IN_TAG;
    }

    void
    Stream::attribute(const string & name, const string & value) {
	ASSERT(mState == STATE_IN_TAG);
	mOs << " " << name << "=\"" << value << "\"";
    }

    void
    Stream::begin_close() {
	ASSERT(mState == STATE_IN_TAG);

	++mIndentLevel;

	mState = STATE_OUT_OF_EMPTY_TAG;
    }

    void
    Stream::write(const string & s) {
	if(mState == STATE_OUT_OF_EMPTY_TAG) {
	    mOs << ">";
	    mState = STATE_OUT_OF_CDATA_TAG;
	}
	else {
	    ASSERT(mState == STATE_OUT_OF_CDATA_TAG);
	}

	for(string::const_iterator iter = s.begin(); iter != s.end(); ++iter) {
	    switch(*iter) {
	    case '<':
		mOs << "&lt;";
		break;
	    case '>':
		mOs << "&gt;";
		break;
	    case '&':
		mOs << "&amp;";
		break;
	    default:
		mOs.put(*iter);
		break;
	    }
	}
    }

    void
    Stream::end() {
	--mIndentLevel;

	if(mState == STATE_OUT_OF_EMPTY_TAG) {
	    mOs << "/>";
	}
	else if(mState == STATE_OUT_OF_NESTED_TAG) {
	    mOs << endl;
	    fill_n(ostream_iterator<char>(mOs), (mIndentLevel * mIndentWidth) / 8, '\t');
	    fill_n(ostream_iterator<char>(mOs), (mIndentLevel * mIndentWidth) % 8, ' ');
	    mOs << "</" << mStack.top() << ">";
	}
	else if(mState == STATE_OUT_OF_CDATA_TAG) {
	    mOs << "</" << mStack.top() << ">";
	}
	else {
	    ASSERT_NOT_REACHED();
	}

	ASSERT(!mStack.empty());
	mStack.pop();

	mState = STATE_OUT_OF_NESTED_TAG;
    }

} // namespace XML
